# Download all MODIS data

source("init.R")

# Parameters ------------------------------------------------------------------

# Datasets to download
# Scene lists downloaded from http://search.earthdata.nasa.gov/
datasets <- list(
  list(
    name = "MOD11A1.006",
    description = "MODIS/Terra Land Surface Temperature and Emissivity Daily L3 Global 1 km Grid SIN V006",
    scenes = "0.raw/modis/MOD11A1.006_scenes.txt"
  ),
  list(
    name = "MYD11A1.006",
    description = "MODIS/Aqua Land Surface Temperature and Emissivity Daily L3 Global 1 km Grid SIN V006",
    scenes = "0.raw/modis/MYD11A1.006_scenes.txt"
  ),
  list(
    name = "MOD13A3.006",
    description = "MODIS/Terra Vegetation Indices Monthly L3 Global 1km Grid SIN V006",
    scenes = "0.raw/modis/MOD13A3.006_scenes.txt"
  ),
  list(
    name = "MYD13A3.006",
    description = "MODIS/Aqua Vegetation Indices Monthly L3 Global 1km Grid SIN V006",
    scenes = "0.raw/modis/MYD13A3.006_scenes.txt"
  )
)

# Functions -------------------------------------------------------------------

# Function to download file via cURL
download <- function(url, path, cookies) {

  if (length(url) != 1) {
    stop("download() can only process one URL at a time")
  }
  dir.create(dirname(path), recursive = TRUE, showWarnings = FALSE)
  # args <- paste("-b", cookies, "-c", cookies, "-Lnf# -o", path, url) # progress bar
  args <- paste("-b", cookies, "-c", cookies, "-Lnfs -o", path, url) # silent
  system2("curl", args)

}

# Get the size of a file on the server
get_server_size <- function(urls, cookies) {

  # Split the list of URLs if it is long (get error if try to run too many at once)
  chunks <- ceiling(seq_along(urls) / 100) %>% split(urls, .)

  # Get the size of each url in each chunk
  regex <- "^Content-Length: "
  lapply(chunks, function(chunk) {

    args <- paste(chunk, collapse = " ") %>% paste("-b", cookies, "-c", cookies, "-sI", .)
    headers <- system2("curl", args, stdout = TRUE)
    grepl(regex, headers) %>% magrittr::extract(headers, .) %>% sub(regex, "", .) %>% as.integer

  }) %>% unlist(use.names = FALSE)

}

# Setup -----------------------------------------------------------------------

# Ensure ~/.netrc contains EarthData username and password
# See https://wiki.earthdata.nasa.gov/display/EL/How+To+Access+Data+With+cURL+And+Wget
if (!file.exists("~/.netrc") || !any(grepl("urs.earthdata.nasa.gov", readLines("~/.netrc")))) {

  msg <- paste(
    "You must add your EarthData username and password to ~/.netrc",
    "See https://wiki.earthdata.nasa.gov/display/EL/How+To+Access+Data+With+cURL+And+Wget",
    sep = "\n"
  )
  stop(msg)

}

# Create a cookie file
# See https://wiki.earthdata.nasa.gov/display/EL/How+To+Access+Data+With+cURL+And+Wget
cookies <- "urs_cookies"
file.create(cookies)

# Download the data -----------------------------------------------------------

for (dataset in datasets) {

  paste("Acquiring", dataset$name) %>% message
  message(dataset$description)

  # Load the list of scenes to download
  scenes <- fread(dataset$scenes, header = FALSE, col.names = "url")

  # Determine the destination for each URL (sort into folders by year)
  data_dir <- dirname(dataset$scenes) %>% file.path(dataset$name)
  scenes[, scene := basename(url)]
  scenes[, date := strsplit(scene, "\\.") %>%
                   sapply("[", 2) %>%
                   sub("A", "", .) %>%
                   as.Date(format = "%Y%j")]
  scenes[, path := file.path(data_dir, year(date), scene)]

  # Check which scenes have already been downloaded
  paste("Checking disk for", nrow(scenes), "scenes") %>% message
  scenes[, downloaded := file.exists(path)]
  scenes[downloaded == TRUE, size_on_disk := file.size(path)]

  # Check the size of downloaded scenes against a saved list of scene sizes on the server
  sizes_path <- file.path(data_dir, "scene_sizes.txt")
  if (file.exists(sizes_path)) {
    scenes[fread(sizes_path), size_on_server := size_on_server, on = "scene"]
  } else {
    scenes[, size_on_server := as.integer(NA)]
  }
  scenes[, size_ok := (size_on_disk == size_on_server)]
  scenes[is.na(size_ok), size_ok := FALSE]

  # Download missing or partially downloaded scenes
  dl_idx <- which(scenes$downloaded == FALSE | scenes$size_ok == FALSE)
  if (length(dl_idx) > 0) {

    paste("Downloading", length(dl_idx), "scenes to", data_dir) %>% message
    for (i in dl_idx) {

      paste0("  ", scenes[i, scene]) %>% message

      # Get the size of the scene on the server
      scenes[i, size_on_server := get_server_size(url, cookies)]

      # Download the scene
      download(scenes[i, url], scenes[i, path], cookies)
      scenes[i, downloaded := TRUE]

      # Get the size of the file on disk
      scenes[i, size_on_disk := file.size(path)]
      scenes[i, size_ok := size_on_disk == size_on_server]

      # Save scene sizes in case the process crashes
      scenes[!is.na(size_on_server), .(scene, size_on_server)] %>% fwrite(sizes_path)

    }

  } else {

    message("All scenes on disk")

  }

  message("Saving scene sizes")
  scenes[order(scene), .(scene, size_on_server)] %>% fwrite(sizes_path)
  paste("->", sizes_path) %>% message
  message("")

}

# Cleanup
file.remove(cookies)

## Python code to move files from dirs by date to dirs by year
# data_dir <- dirname(dataset$scenes) %>% file.path(dataset$name)
# years <- list.files(data_dir, pattern = "\\d{4}-") %>% as.Date %>% year %>% unique
# for (year in years) {
#   new_dir <- file.path(data_dir, year)
#   dir.create(new_dir, showWarnings = FALSE)
#   for (d in dir(data_dir, pattern = paste0(year, "-"), full.names = T)) {
#     message(d)
#     for (f in list.files(d, full.names = T)) {
#       if (grepl("\\.hdf$", f)) {
#         new_path <- file.path(new_dir, basename(f))
#         file.rename(f, new_path)
#       } else {
#         file.remove(f)
#       }
#     }
#     # file.remove(d)
#   }
# }
# next
# stop()
