source("init.R")

report("*** Reformatting .rds data extracts ***")
rds_paths <- list.files(constants$extracts_dir, pattern = "*.rds", full.names = TRUE,
                        recursive = TRUE)
mclapply(rds_paths, function(rds_path) {

  paste("Loading", basename(rds_path)) %>% report
  dat <- readRDS(rds_path)

  # Confirm data.table
  if (!("data.table" %in% class(dat))) {
    report("  converting to data.table")
    dat <- as.data.table(dat)
  }

  # Confirm grd_1km_id col
  if ("lst_1km_id" %in% names(dat)) {
    report("  lst_1km_id -> grd_1km_id")
    setnames(dat, "lst_1km_id", "grd_1km_id")
  }

  # Confirm no month col
  if ("month" %in% names(dat)) {
    report("  dropping month column")
    dat[, month := NULL]
  }

  # Confirm date col is Date
  if ("date" %in% names(dat) && class(dat$date) != "Date") {
    report("  casting date to Date")
    dat[, date := as.Date(date)]
  }

  # Confirm keys
  if ("date" %in% names(dat)) {
    report("  keying on date and grd_1km_id")
    setkeyv(dat, c("date", "grd_1km_id"))
  } else if (key(dat) != "grd_1km_id") {
    report("  keying on grd_1km_id")
    setkey(dat, "grd_1km_id")
  }

  fst_path <- sub("\\.rds$", ".fst", rds_path) %>% sub("grid_lst_1km", "grd_1km", .)
  paste("  ->", basename(fst_path)) %>% report
  write_fst(dat, fst_path, compress = 100)

  paste("  removing", basename(rds_path)) %>% report
  file.remove(rds_path)

}, mc.cores = ncores * 0.75, mc.preschedule = FALSE)


report("*** Renaming tif files ***")
tif_paths <- list.files(constants$extracts_dir, pattern = "^grid_lst_1km.+\\.tif",
                        full.names = TRUE, recursive = TRUE)
for (tif_path in tif_paths) {

  new_path <- sub("grid_lst_1km", "grd_1km", tif_path)
  paste(basename(tif_path), "->", basename(new_path)) %>% report
  file.rename(tif_path, new_path)

}
rm(tif_path, new_path)


report("*** Checking sim data extracts ***")
sim_paths <- list.files(constants$extracts_dir, pattern = "^grid_lst_1km_sim.+\\.fst",
                        full.names = TRUE, recursive = TRUE)
mclapply(sim_paths, function(sim_path) {

  paste("Loading", basename(sim_path)) %>% report
  dat <- read_fst(sim_path, as.data.table = TRUE)

  # Confirm grd_1km_id col
  if ("lst_1km_id" %in% names(dat)) {
    report("  lst_1km_id -> grd_1km_id")
    setnames(dat, "lst_1km_id", "grd_1km_id")
  }

  # Confirm date col is Date
  if (class(dat$date) != "Date") {
    report("  casting date to Date")
    dat[, date := as.Date(date)]
  }

  # Confirm keys
  if (is.null(key(dat)) || key(dat) != c("date", "grd_1km_id")) {
    report("  keying on date and grd_1km_id")
    setkeyv(dat, c("date", "grd_1km_id"))
  }

  new_path <- sub("grid_lst_1km", "grd_1km", sim_path)
  paste("  ->", basename(new_path)) %>% report
  write_fst(dat, new_path, compress = 100)

  paste("  removing", basename(sim_path)) %>% report
  file.remove(sim_path)

}, mc.cores = ncores * 0.75, mc.preschedule = FALSE)


report("*** Checking for missing NDVI ***")
for (sat in c("aqua", "terra")) {
  for (year in constants$model_years) {
    for (month in 1:12) {
      month <- paste0(year, "-%02d") %>% sprintf(month)
      path <-
        paste0("grd_1km_", sat, "_ndvi_", month, ".fst") %>%
        file.path(constants$extracts_dir, "modis_ndvi", .)
      if (!file.exists(path)) {
        paste("Missing", basename(path)) %>% print
      }
    }
  }
}


report("*** Checking for missing LST ***")
for (sat in c("aqua", "terra")) {
  for (year in constants$model_years) {
    for (month in 1:12) {
      month <- paste0(year, "-%02d") %>% sprintf(month)
      path <-
        paste0("grd_1km_", sat, "_lst_", month, ".fst") %>%
        file.path(constants$extracts_dir, "modis_lst", .)
      if (!file.exists(path)) {
        paste("Missing", basename(path)) %>% print
      }
    }
  }
}


report("*** Checking for missing SIM ***")
for (year in constants$model_years) {
  for (month in 1:12) {
    month <- paste0(year, "-%02d") %>% sprintf(month)
    path <-
      paste0("grd_1km_sim_", month, ".fst") %>%
      file.path(constants$extracts_dir, "sim", .)
    if (!file.exists(path)) {
      paste("Missing", basename(path)) %>% print
    }
  }
}


report("Done")
