# What is wrong with 2004 Tmean predictions?

library(magrittr)
library(data.table)
library(fst)
library(glue)
library(sf)        # for nngeo
library(nngeo)     # nearest neighbor calculation (used to fill any water / snow gaps)
library(sp)        # for visualisation
library(raster)    # for visualisation
library(rasterVis) # for visualisation
library(mapview)   # for visualisation

source("2.scripts/helpers/constants.R")
source("2.scripts/helpers/get_ncores.R")
source("2.scripts/helpers/report.R")
source("2.scripts/helpers/summarize_as_raster.R")


ncores <- get_ncores()
threads_fst(ncores)


# 1. Check predictions ----------

# Load predictions
report("Loading predictions")
mod3 <- read_fst("1.work/predictions/mod3_2004_pred_m2.fst", as.data.table = TRUE,
                 columns = c("date", "grd_1km_id", "grd_1km_x", "grd_1km_y", "idw_tmean",
                             "m1_tmean", "m2_tmean"))
preds <- read_fst("1.work/predictions/mod3_2004_pred_final.fst", as.data.table = TRUE,
                  columns = c("date", "grd_1km_id", "tmean", "tmean_src"))
if (identical(mod3[, .(date, grd_1km_id)], preds[, .(date, grd_1km_id)])) {
  mod3[, c("final_tmean", "final_tmean_src") := preds[, .(tmean, tmean_src)]]
  rm(preds)
} else {
  stop("MISMATCH mod3 vs preds")
}


# Load meteo observations
report("Loading meteo data")
meteo <- read_fst("1.work/meteo_obs/clean/meteo_data_2004-clean.fst", as.data.table = TRUE,
                  columns = c("date", "insee_id", "latitude", "longitude", "temperature_mean")) %>%
  setnames(c("latitude", "longitude", "temperature_mean"), c("lat", "lon", "tmean"))

# Drop stations that do not record tmean
meteo <- meteo[!is.na(tmean), ]


# Load the grid
grd <- mod3[date == first(date), .(grd_1km_id, grd_1km_x, grd_1km_y)]


# Associate meteo observations with grid cells
report("Matching observed Ta to predictions")
lookup <- meteo[, .(insee_id, lat, lon)] %>%
          unique
xy <- st_as_sf(lookup, coords = c("lon", "lat"), crs = 4326) %>%
      st_transform(2154) %>%
      st_coordinates %>%
      as.data.table
lookup[, c("x", "y") := xy]
rm(xy)

matches <- FNN::get.knnx(grd[, .(grd_1km_x, grd_1km_y)], lookup[, .(x, y)], k = 1) %>%
           as.data.table %>%
           setnames(c("grd_idx", "dist"))
lookup[, c("grd_idx", "dist") := matches]
lookup[, "grd_1km_id" := grd[lookup$grd_idx, grd_1km_id]]
lookup <- lookup[dist < 1500, ] # Drop stations >= 1500 m from a grid point

meteo <- meteo[lookup, on = c("insee_id", "lat", "lon")] %>%
         .[, .(date, insee_id, lat, lon, x, y, tmean, grd_1km_id)]
rm(lookup, matches)


# Match meteo observations to predictions
dat <- meteo[, .(date, grd_1km_id, insee_id, tmean)] %>%
       .[mod3, .(date, grd_1km_id, grd_1km_x, grd_1km_y, insee_id, tmean, idw_tmean, m1_tmean,
                 m2_tmean, final_tmean, final_tmean_src),
         on = c("date", "grd_1km_id")] %>%
       setkeyv(c("date", "grd_1km_id"))
rm(mod3)


# Check performance metrics
report("Assessing performance")
check <- function(dt, obs_col, pred_col) {
  c(
    mae = dt[, get(obs_col) - get(pred_col)] %>% abs %>% mean(na.rm = TRUE),
    rmse = dt[, get(obs_col) - get(pred_col)]^2 %>% mean(na.rm = TRUE) %>% sqrt
  )
}

stn_cells <- dat[!is.na(tmean), ]
check(stn_cells, "tmean", "idw_tmean")
check(stn_cells, "tmean", "m1_tmean")
check(stn_cells, "tmean", "m2_tmean")
check(stn_cells, "tmean", "final_tmean")


# Summarize predictions
report("Summarizing predictions")
count_na <- function(x) {
  is.na(x) %>% sum
}
sdcols <- c("tmean", "idw_tmean", "m1_tmean", "m2_tmean", "final_tmean")

means <- dat[, lapply(.SD, mean, na.rm = TRUE), by = grd_1km_id, .SDcols = sdcols] %>%
         setnames(sdcols, glue("avg_{sdcols}"))
nas <- dat[, lapply(.SD, count_na), by = grd_1km_id, .SDcols = sdcols] %>%
       setnames(sdcols, glue("na_{sdcols}"))
smry <- means[nas, on = "grd_1km_id"]


# Save the summary
report("Saving summary")
dir.create("1.work/debug", showWarnings = FALSE)
smry_path <- "1.work/debug/tmean_2004_m2_smry.fst"
write_fst(smry, smry_path, compress = 100)
report(glue("-> {smry_path}"))


# Map the predictions
report("Mapping predictions")
rast <- smry[grd, on = "grd_1km_id"]
rast[, grd_1km_id := as.factor(grd_1km_id) %>% as.integer]
coordinates(rast) <- ~ grd_1km_x + grd_1km_y
proj4string(rast) <- constants$epsg_2154
rast <- spTransform(rast, constants$proj_sinu)
gridded(rast) <- TRUE
rast <- stack(rast)

rast_path <- "1.work/debug/tmean_2004_m2_smry.tif"
bn_path <- paste0(rast_path, ".bandnames")
names(rast) %>% t %>% as.data.frame %>% fwrite(bn_path, col.names = FALSE)
save_geotiff(rast, rast_path)
report(glue("-> {rast_path}"))


for (i in 1:length(sdcols)) {

  avg <- rast[[glue("avg_{x}")]]
  missing <- rast[[glue("na_{x}")]]

  lims <- quantile(avg, c(0.001, 0.999))
  lims[1] <- floor(lims[1])
  lims[2] <- ceiling(lims[2])
  breaks <- seq(lims[1], lims[2], by = 0.25)

  p1 <- levelplot(avg, main = x, at = breaks, margin = FALSE, xlab = NULL, ylab = NULL,
                  par.settings = infernoTheme, scales = list(draw = FALSE))
  p2 <- levelplot(missing, main = "missing", at = seq(0, 365, by = 10), margin = FALSE,
                  xlab = NULL, ylab = NULL, scales = list(draw = FALSE))

  map_path <- glue("1.work/debug/tmean_2004_m2_map{i}_{x}.png")
  png(map_path, width = 1920, height = 920)
  print(p1, split = c(1, 1, 2, 1), more = TRUE)
  print(p2, split = c(2, 1, 2, 1))
  dev.off()
  report(glue("-> {map_path}"))

}

# mapview(rast$na_tmean, native.crs = TRUE)
# mapview(rast$na_idw_tmean, native.crs = TRUE)
# mapview(rast$na_m1_tmean, native.crs = TRUE)
# mapview(rast$na_m2_tmean, native.crs = TRUE)
# mapview(rast$na_final_tmean, native.crs = TRUE)
#
# mapview(rast$avg_tmean, native.crs = TRUE)
# mapview(rast$avg_idw_tmean, native.crs = TRUE)
# mapview(rast$avg_m1_tmean, native.crs = TRUE)
# mapview(rast$avg_m2_tmean, native.crs = TRUE)
# mapview(rast$avg_final_tmean, native.crs = TRUE)


# 2. Check monthly m2 predictions ----------

dat[, month := month(date)]
setkeyv(dat, c("date", "grd_1km_id", "month"))

mo_smrys <- dat[, .(date, grd_1km_id, month, m2_tmean)] %>%
  split(by = "month") %>%
  lapply(function(dt) {
    dt[, .(month = mean(month), m2_tmean = mean(m2_tmean, na.rm = TRUE),
           missing = sum(is.na(m2_tmean))),
       by = c("grd_1km_id", "month")] %>%
      setkey("grd_1km_id")
  })

mo_rasts <- mo_smrys %>%
  mclapply(function(dt) {
    rast <- dt[grd, on = "grd_1km_id"]
    rast[, grd_1km_id := as.factor(grd_1km_id) %>% as.integer]
    coordinates(rast) <- ~ grd_1km_x + grd_1km_y
    proj4string(rast) <- constants$epsg_2154
    rast <- spTransform(rast, constants$proj_sinu)
    gridded(rast) <- TRUE
    rast <- stack(rast)
  }, mc.cores = ncores)

for (i in 1:length(mo_rasts)) {

  r <- mo_rasts[[i]]

  lims <- quantile(r$m2_tmean, c(0.001, 0.999))
  lims[1] <- floor(lims[1])
  lims[2] <- ceiling(lims[2])
  breaks <- seq(lims[1], lims[2], by = 0.25)

  title <- glue("{i}: m2_tmean")
  p1 <- levelplot(r$m2_tmean, main = title, at = breaks, margin = FALSE, xlab = NULL, ylab = NULL,
                  par.settings = infernoTheme, scales = list(draw = FALSE))
  p2 <- levelplot(r$missing, main = "missing", at = seq(0, 365, by = 10), margin = FALSE,
                  xlab = NULL, ylab = NULL, scales = list(draw = FALSE))

  map_path <- glue("1.work/debug/tmean_2004_m2_{sprintf('%02d', i)}.png")
  png(map_path, width = 1920, height = 920)
  print(p1, split = c(1, 1, 2, 1), more = TRUE)
  print(p2, split = c(2, 1, 2, 1))
  dev.off()
  report(glue("-> {map_path}"))

}


# 3. Check LST data ----------

# Load Terra night LST
report("Loading Terra night LST")
lst <- list.files("1.work/extracts/modis_lst", pattern = "grd_1km_terra_lst_2004",
                  full.names = T) %>%
  lapply(function(path) {
    report(glue("Loading {basename(path)}"))
    read_fst(path, as.data.table = TRUE, columns = c("date", "grd_1km_id", "night_lst",
                                                     "night_lst_error"))
  }) %>%
  rbindlist %>%
  setnames(c("night_lst", "night_lst_error"), c("lst", "lst_error")) %>%
  setkeyv(c("date", "grd_1km_id"))

# Remove LST with error > 2K
lst[lst_error == "<= 3K", lst := NA]

# Summarize LST by cell
report("Summarizing LST")
lst_smry <- lst[, .(avg_lst = mean(lst, na.rm = TRUE), na_lst = is.na(lst) %>% sum),
                by = grd_1km_id]

# Save the summary
report("Saving summary")
smry_path <- "1.work/debug/tmean_2004_lst_smry.fst"
write_fst(lst_smry, smry_path, compress = 100)
report(glue("-> {smry_path}"))

# Map LST
report("Mapping LST")
lst_rast <- lst_smry[grd, on = "grd_1km_id"]
lst_rast[, grd_1km_id := as.factor(grd_1km_id) %>% as.integer]
coordinates(lst_rast) <- ~ grd_1km_x + grd_1km_y
proj4string(lst_rast) <- constants$epsg_2154
lst_rast <- spTransform(lst_rast, constants$proj_sinu)
gridded(lst_rast) <- TRUE
lst_rast <- stack(lst_rast)

# mapview(lst_rast$na_lst, native.crs = TRUE)
# mapview(lst_rast$avg_lst, native.crs = TRUE)

rast_path <- glue("1.work/debug/tmean_2004_lst_smry.tif")
bn_path <- paste0(rast_path, ".bandnames")
names(lst_rast) %>% t %>% as.data.frame %>% fwrite(bn_path, col.names = FALSE)
save_geotiff(lst_rast, rast_path)
report(glue("-> {rast_path}"))


lims <- quantile(lst_rast$avg_lst, c(0.001, 0.999))
lims[1] <- floor(lims[1])
lims[2] <- ceiling(lims[2])
breaks <- seq(lims[1], lims[2], by = 0.25)

p1 <- levelplot(lst_rast$avg_lst, main = "avg_lst", at = breaks, margin = FALSE, xlab = NULL,
                ylab = NULL, par.settings = infernoTheme, scales = list(draw = FALSE))
p2 <- levelplot(lst_rast$na_lst, main = "missing", at = seq(0, 365, by = 10), margin = FALSE,
                xlab = NULL, ylab = NULL, scales = list(draw = FALSE))

map_path <- glue("1.work/debug/tmean_2004/lst_smry.png")
png(map_path, width = 1920, height = 920)
print(p1, split = c(1, 1, 2, 1), more = TRUE)
print(p2, split = c(2, 1, 2, 1))
dev.off()
report(glue("-> {map_path}"))
