# Regenerate 2004 tmean predictions

library(magrittr)
library(data.table)
library(fst)
library(glue)
library(sf)        # for nngeo
library(nngeo)     # nearest neighbor calculation (used to fill any water / snow gaps)
library(sp)        # for visualisation
library(raster)    # for visualisation
library(rasterVis) # for visualisation
library(mapview)   # for visualisation

source("2.scripts/helpers/constants.R")
source("2.scripts/helpers/get_ncores.R")
source("2.scripts/helpers/report.R")
source("2.scripts/helpers/summarize_as_raster.R")


ncores <- get_ncores()
threads_fst(ncores)



# Load old m2 predictions (15 min, 13 GB)
report("Loading 1.work/predictions/mod3_2004_pred_m2.fst.bak")
threads_fst(ncores)
mod3 <- read_fst("1.work/predictions/mod3_2004_pred_m2.fst.bak", as.data.table = TRUE,
                 columns = c("date", "grd_1km_id", "grd_1km_x", "grd_1km_y", "idw_tmean", "m2_tmean"))

# Load old and new m2 (25 min, 18.6 GB for both)
report("Loading 1.work/models/m2_tmean_2004.rds.bak (old model)")
m2_old <- readRDS("1.work/models/m2_tmean_2004.rds.bak")
report("Loading 1.work/models/m2_tmean_2004.rds (new model)")
m2_new <- readRDS("1.work/models/m2_tmean_2004.rds")

# HACK set the fixed part of the model's formula
m2_old$call$fixed <- m1_tmean ~ idw_tmean
m2_new$call$fixed <- m1_tmean ~ idw_tmean



report("CHECK MODELS")
stop("CHECK MODELS")



# Split mod3 to speed up predictions (12.9 GB)
report("Splitting mod3 into 10 groups")
mod3[, split_idx := ceiling(1:nrow(mod3) / (nrow(mod3) / 10))]
splits <- split(mod3, by = "split_idx")
mod3[, split_idx := NULL]

# Predict with old model
report("Predicting tmean with old model")
preds_old <- mclapply(splits, function(splt) {
               predict(m2_old, newdata = splt)
             }, mc.cores = ncores) %>%
             unlist(use.names = FALSE)
mod3[, m2_tmean_old := preds_old]
rm(preds_old)

# Predict with new model
report("Predicting tmean with new model")
preds_new <- mclapply(splits, function(splt) {
               predict(m2_new, newdata = splt)
             }, mc.cores = ncores) %>%
             unlist(use.names = FALSE)
mod3[, m2_tmean_new := preds_new]
rm(preds_new)

# Summarizing predictions
report("Summarizing predictions")
smry <- mod3[, .(
                 idw = mean(idw_tmean, na.rm = TRUE),
                 orig = mean(m2_tmean, na.rm = TRUE),
                 old = mean(m2_tmean_old, na.rm = TRUE),
                 new = mean(m2_tmean_new, na.rm = TRUE)
               ),
             by = grd_1km_id] %>%
        setkey("grd_1km_id")
rast <- smry[grd, on = "grd_1km_id"] %>%
        .[, grd_1km_id := NULL] %>%
        st_as_sf(coords = c("grd_1km_x", "grd_1km_y"), crs = 2154) %>%
        st_transform(constants$proj_sinu) %>%
        as_Spatial
gridded(rast) <- TRUE
rast <- stack(rast)

par(mar = c(1, 1, 1, 1))
plot(rast, axes = FALSE, box = FALSE, legend = FALSE)



report("CHECK PREDICTIONS")
stop("CHECK PREDICTIONS")



# Store the good predictions
preds <- mod3[, m2_tmean_old]
rm(m2_old, m2_new, mod3)

# Load the bad predictions
report("Loading 1.work/predictions/mod3_2004_pred_m2.fst.bak")
mod3 <- read_fst("1.work/predictions/mod3_2004_pred_m2.fst.bak", as.data.table = TRUE)

# Replace the bad predictions with the good predictions
mod3[, m2_tmean := preds]

# Save
report("Saving updated predictions")
write_fst(mod3, "1.work/predictions/mod3_2004_pred_m2.fst", compress = 100)
report("-> 1.work/predictions/mod3_2004_pred_m2.fst")

report("Summarizing predictions")
smry_path <- "1.work/predictions/mod3_2004_pred_m2_tmean_smry.tif"
smry <- summarize_as_raster(mod3, "m2_tmean")
save_geotiff(smry, smry_path, overwrite = TRUE)
paste(year, ": ->", basename(smry_path)) %>% report



report("ok")
