#
# Create mod3 by joining extracted data to the 1 km grid
#

# Load packages, helper functions, and constants
source("init.R")

report("---")
report("Create mod3 (db to predict 1 km Ta from LST)")


# Define functions ------------------------------------------------------------

# Add columns to a table by reference (does not copy the table)
add_by_reference <- function(main_table, new_data) {
  if (safe_to_add(main_table, new_data)) {
    # Drop the key columns from the new data, then add the remaining columns to the main table
    new_data[, key(main_table) := NULL]
    main_table[, names(new_data) := new_data]
    invisible(main_table)
  } else {
    stop("Not safe to add by reference")
  }
}

# List all dates in the month of the passed date
dates_in_month <- function(d) {
  month_start <- format(d, "%Y-%m-01") %>% as.Date
  seq.Date(
    from = month_start,
    to = seq.Date(month_start, by = "month", length.out = 2)[2] - 1,
    by = 1
  )
}

# Join data to a table by reference (does not copy the table)
join_by_reference <- function(main_table, new_data) {
  if (safe_to_join(main_table, new_data)) {
    # Find the table rows that match the new data and update them by reference
    cols_to_add <- not_key(new_data)
    main_table[new_data, (cols_to_add) := mget(paste0("i.", cols_to_add)), on = key(new_data)]
    invisible(main_table)
  } else {
    stop("Not safe to join by reference")
  }
}

# Return the path to mod3 data for the specified date
mod3_path_for_date <- function(d) {
  format(d, "mod3_%Y-%m.fst") %>% file.path(constants$joined_dir, .)
}

# List the columns of a data.table that are not keys
not_key <- function(data_table) {
  setdiff(names(data_table), key(data_table))
}

# Add a prefix to the name of every non-key column in a data.table
prefix_non_key_cols <- function(data_table, prefix) {
  cols <- not_key(data_table)
  setnames(data_table, old = cols, new = paste(prefix, cols, sep = "_"))
  invisible(data_table)
}

# Check whether it's safe to add new data to a table by reference
safe_to_add <- function(main_table, new_data) {
  all(
    # Confirm the data is safe to join (it won't overwrite existing columns)
    safe_to_join(main_table, new_data),

    # Confirm the key columns are identical i.e. the data is in the same grid id - date order
    identical(
      main_table[, key(main_table), with = FALSE],
      new_data[, key(new_data), with = FALSE]
    )
  )
}

# Check whether it's safe to join new data to table by reference
safe_to_join <- function(main_table, new_data) {
  # Confirm the data will not overwrite existing columns in the main table
  shared_cols <- intersect(names(main_table), names(new_data))
  identical(key(new_data), shared_cols)
}

# Join data for each year -----------------------------------------------------

threads_fst(ncores)

for (year in constants$model_years) {

  # mod3 = all cell-days ------------------------------------------------------
  # Process data by month (entire year uses too much memory)
  for (month in 1:12) {

    month_start <- paste0(year, "-%02d-01") %>% sprintf(month) %>% as.Date

    # Skip the month if mod3 already exists
    mod3_path <- mod3_path_for_date(month_start)
    if (file.exists(mod3_path) && !OVERWRITE) {
      next
    }

    # Load the 1km grid
    format(month_start, "%Y-%m: loading 1 km grid") %>% report
    grid_table <- file.path(constants$grids_dir, "grd_1km.fst") %>%
                  read_fst(as.data.table = TRUE)

    # List all cell-days for the month (10 sec, 1.2 GB)
    # (Entire year would take 30 seconds and be 11.9 GB)
    format(month_start, "%Y-%m: listing cell-days in month") %>% report
    mod3 <- CJ("date" = dates_in_month(month_start), "id" = grid_table$id) %>%
            grid_table[., on = "id"] %>%
            setnames(old = names(grid_table), new = paste0("grd_1km_", names(grid_table))) %>%
            setkeyv(cols = c("date", "grd_1km_id"))
    c(key(mod3), not_key(mod3)) %>% setcolorder(mod3, .)

    # Join Elevation (10 seconds, 0.1 GB)
    format(month_start, "%Y-%m: joining EU DEM v1.1 elevation") %>% report
    for (prefix in "eu_dem_v11") {

      # Load the extracted elevation data
      tmp <- file.path(constants$extracts_dir, "grd_1km_elev.fst") %>%
             read_fst(as.data.table = TRUE)

      # Prefix the data column with "eu_dem_v11"
      prefix_non_key_cols(tmp, prefix)

      # Join the data to mod3
      join_by_reference(mod3, tmp)

      rm(tmp, prefix)

    }

    # Join Population (10 seconds, 0.1 GB)
    format(month_start, "%Y-%m: joining INSEE population") %>% report
    for (prefix in "insee") {

      # Load the extracted population data
      tmp <- file.path(constants$extracts_dir, "grd_1km_pop.fst") %>%
             read_fst(as.data.table = TRUE)

      # Prefix the data column with "insee"
      prefix_non_key_cols(tmp, prefix)

      # Join the data to mod3
      join_by_reference(mod3, tmp)

      rm(tmp, prefix)

    }

    # Join Land Cover (10 seconds, 1.4 GB)
    format(month_start, "%Y-%m: joining Corine Land Cover") %>% report
    for (prefix in "clc") {

      # Load the extracted CLC data for the data year closest to the current year
      diffs <- abs(year - constants$clc_years)
      closest_year <- constants$clc_years[diffs == min(diffs)][1]
      tmp <- paste0("grd_1km_clc_", closest_year, ".fst") %>%
             file.path(constants$extracts_dir, .) %>%
             read_fst(as.data.table = TRUE)
      rm(diffs, closest_year)

      # HACK ensure the CLC data includes a column for category "241"
      # In 2000 this column is missing b/c no areas in France had land cover "241"
      if (!"241" %in% names(tmp)) {
        tmp[, "241" := NA]
      }

      # Aggregate land use data into categories
      categories <- list(
        artificial = c(
          "111", # Continuous urban fabric
          "112", # Discontinuous urban fabric
          "121", # Industrial or commercial units
          "122", # Road and rail networks and associated land
          "123", # Port areas
          "124", # Airports
          "131", # Mineral extraction sites
          "132", # Dump sites
          "133", # Construction sites
          "141", # Green urban areas
          "142"  # Sport and leisure facilities
        ),
        vegetation = c(
          "211", # Non-irrigated arable land
          "212", # Permanently irrigated land
          "213", # Rice fields
          "221", # Vineyards
          "222", # Fruit trees and berry plantations
          "223", # Olive groves
          "231", # Pastures
          "241", # Annual crops associated with permanent crops
          "242", # Complex cultivation patterns
          "243", # Land principally occupied by agriculture, with significant areas of natural vegetation
          "244", # Agro-forestry areas
          "311", # Broad-leaved forest
          "312", # Coniferous forest
          "313", # Mixed forest
          "321", # Natural grasslands
          "322", # Moors and heathland
          "323", # Sclerophyllous vegetation
          "324"  # Transitional woodland-shrub
        ),
        bare = c(
          "331", # Beaches, dunes, sands
          "332", # Bare rocks
          "333", # Sparsely vegetated areas
          "334", # Burnt areas
          "335"  # Glaciers and perpetual snow
        ),
        water = c(
          "411", # Inland marshes
          "412", # Peat bogs
          "421", # Salt marshes
          "422", # Salines
          "423", # Intertidal flats
          "511", # Water courses
          "512", # Water bodies
          "521", # Coastal lagoons
          "522", # Estuaries
          "523"  # Sea and ocean
        )
      )
      for (category in names(categories)) {
        clc_codes <- categories[[category]]
        tmp[, (category) := rowSums(.SD, na.rm = TRUE), .SDcols = clc_codes]
      }
      tmp <- tmp[, c(key(tmp), names(categories)), with = FALSE]
      rm(category, categories, clc_codes)

      # Prefix the data columns with "clc"
      prefix_non_key_cols(tmp, prefix)

      # Join the data to mod3
      join_by_reference(mod3, tmp)

      rm(tmp, prefix)

    }

    # Join climate type (10 seconds, 0.1 GB)
    format(month_start, "%Y-%m: joining Joly (2010) climate type") %>% report
    for (prefix in "joly_2010") {

      # Load the extracted data
      tmp <- file.path(constants$extracts_dir, "grd_1km_climate.fst") %>%
             read_fst(as.data.table = TRUE)

      # Prefix the data column with "insee"
      prefix_non_key_cols(tmp, prefix)

      # Join the data to mod3
      join_by_reference(mod3, tmp)

      rm(tmp, prefix)

    }

    # # Join Meteo France SIM model output (15 seconds, 1.4 GB)
    # format(month_start, "%Y-%m: joining Meteo France SIM surface model output") %>% report
    # for (prefix in "sim") {
    #
    #   # Load the extracted SIM data
    #   tmp <- paste0("grd_1km_sim_", format(month_start, "%Y-%m"), ".fst") %>%
    #          file.path(constants$extracts_dir, "sim", .) %>%
    #          read_fst(as.data.table = TRUE)
    #
    #   # Drop variables we don't need
    #   sim_vars <- c("temperature_min", "temperature_mean", "temperature_max")
    #   tmp <- tmp[, c(key(tmp), sim_vars), with = FALSE]
    #   rm(sim_vars)
    #
    #   # Prefix the data columns with "sim"
    #   prefix_non_key_cols(tmp, prefix)
    #
    #   # Add the data to mod3 by reference
    #   add_by_reference(mod3, tmp)
    #
    #   rm(tmp, prefix)
    #
    # }

    # Join MODIS NDVI (25 seconds, 0.4 GB)
    for (sat in constants$modis_satellites) {

      # Find the extracted data for the month
      path <- paste0("grd_1km_", sat, "_ndvi_", format(month_start, "%Y-%m"), ".fst") %>%
              file.path(constants$extracts_dir, "modis_ndvi", .)

      if (file.exists(path)) {

        format(month_start, "%Y-%m: joining MODIS") %>% paste(sat, "NDVI") %>% report
        tmp <- read_fst(path, as.data.table = TRUE)

      } else {

        format(month_start, "%Y-%m: no MODIS") %>% paste(sat, "NDVI; filling with NA") %>% report

        # Get the names of the columns that would have been added had data existed
        ndvi_cols <- paste0("grd_1km_", sat, "_ndvi") %>%
                     list.files(constants$extracts_dir, full.names = TRUE, pattern = .,
                                recursive = TRUE) %>%
                     magrittr::extract(1) %>%
                     read_fst(as.data.table = TRUE, to = 1) %>%
                     names

        # Make a dummy data.table with grid ids but no data
        # Don't include dates b/c real data doesn't have dates - NDVI data is monthly
        tmp <- mod3[date == mod3$date[1], "grd_1km_id", with = FALSE]
        tmp %>% setkey("grd_1km_id")
        tmp[, setdiff(ndvi_cols, names(tmp)) := NA]
        rm(ndvi_cols)

      }
      rm(path)

      # Prefix the data columns with the satellite
      prefix_non_key_cols(tmp, sat)

      # Join the data to mod3 by reference
      join_by_reference(mod3, tmp)
      rm(tmp, sat)

    }

    # Join MODIS LST (40 seconds, 2.2 GB)
    for (sat in constants$modis_satellites) {

      # Find the extracted data for the month
      data_path <- paste0("grd_1km_", sat, "_lst_", format(month_start, "%Y-%m"), ".fst") %>%
                   file.path(constants$extracts_dir, "modis_lst", .)

      if (file.exists(data_path)) {

        format(month_start, "%Y-%m: joining MODIS") %>% paste(sat, "LST") %>% report

        # Load the extracted LST data
        tmp <- read_fst(data_path, as.data.table = TRUE)

      } else {

        format(month_start, "%Y-%m: no MODIS") %>% paste(sat, "LST; filling with NA") %>% report

        # Get the names of the columns that would have been added had data existed
        lst_cols <- paste0("grd_1km_", sat, "_lst") %>%
                    list.files(constants$extracts_dir, full.names = TRUE, pattern = .,
                               recursive = TRUE) %>%
                    magrittr::extract(1) %>%
                    read_fst(as.data.table = TRUE, to = 1) %>%
                    names

        # Make a table listing every cell-day with NA-filled columns for LST data
        tmp <- mod3[, key(mod3), with = FALSE]
        tmp[, setdiff(lst_cols, names(tmp)) := NA]
        rm(lst_cols)

      }
      rm(data_path)

      # Prefix the non-key data columns with the satellite name
      prefix_non_key_cols(tmp, sat)

      # Add the data to mod3 by reference
      add_by_reference(mod3, tmp)
      rm(tmp, sat)

    }

    # Remove LST for cells that are > 33% water or have NDVI < 0 (5 seconds)
    format(month_start, "%Y-%m: removing LST for cells with > 33% water or NDVI < 0") %>% report
    for (sat in constants$modis_satellites) {

      for (overpass in constants$modis_overpasses) {

        lst_col <- paste(sat, overpass, "lst", sep = "_")
        ndvi_col <- paste(sat, "ndvi", sep = "_")
        mod3[clc_water > 0.33, (lst_col) := NA]
        mod3[get(ndvi_col) < 0, (lst_col) := NA]

      }

      rm(lst_col, ndvi_col, sat, overpass)

    }

    # Join IDW interpolated temperatures (120 seconds, 0.5 GB)
    for (prefix in "meteo") {

      # Load cleaned meteorological data for the current month
      format(month_start, "%Y-%m: loading meteo observations") %>% report
      {

        # Load data for the month
        pattern <- paste0("meteo_data_", year, "-clean\\.fst")
        meteo <- file.path(constants$work_dir, "meteo_obs", "clean") %>%
                 list.files(full.names = TRUE, pattern = pattern) %>%
                 read_fst(as.data.table = TRUE) %>%
                 .[date %in% dates_in_month(month_start), ]
        rm(pattern)

        # Convert the observation locations to EPSG:2154 (Lambert-93)
        xy_lamb93 <- meteo[, .(longitude, latitude)] %>%
                     SpatialPoints(proj4string = CRS(constants$epsg_4326)) %>%
                     spTransform(constants$epsg_2154) %>%
                     as.data.table
        meteo[, c("x", "y") := xy_lamb93]
        rm(xy_lamb93)

        # Drop unneeded columns
        meteo <- meteo[, .(date, insee_id, latitude, longitude, x, y, station_type,
                           station_elevation, temperature_min, temperature_mean, temperature_max)]

      }

      # For each cell-day, predict each temperature using inverse distance weighted interpolation
      # 75 seconds, +0.4 GB -> 8.1 GB
      for (t_col in c("temperature_min", "temperature_mean", "temperature_max")) {

        format(month_start, "%Y-%m: interpolating") %>% paste(t_col, "by IDW") %>% report

        # Drop observations where the variable is missing and split by date
        splits <- meteo[!is.na(get(t_col)), c("date", "x", "y", t_col), with = FALSE] %>%
                  split(by = "date")

        # Run the IDW interpolation for all dates in parallel
        # 25 seconds
        idw <- mclapply(splits, function(splt, t_col, grid_table) {

          # Predict for each grid point and round
          predictions <- paste0(t_col, " ~ 1") %>%
                         as.formula %>%
                         gstat(formula = ., locations = ~ x + y, data = splt) %>%
                         predict(newdata = grid_table) %>%
                         extract2("var1.pred")

          # Return a data table with the date, lst 1 km id, and predictions
          result <- data.table(date = splt$date[1], grd_1km_id = grid_table$id) %>%
                    .[, paste0("idw_", t_col) := predictions]

        }, mc.cores = ncores, t_col = t_col, grid_table = grid_table) %>% rbindlist

        # Sort predictions by date and lst 1 km id and add to mod3 by reference
        idw %>% setkeyv(cols = c("date", "grd_1km_id"))
        add_by_reference(mod3, idw)

        # Cleanup
        rm(idw, splits, t_col)

      }
      rm(meteo, prefix)

    }

    # Save mod3
    #  90 seconds ( 35 to read), 1063 MB for fst
    # 280 seconds (100 to read),  915 MB for RDS
    format(month_start, "%Y-%m: saving mod3") %>% report
    threads_fst(ncores)
    tmp_path <- paste0(mod3_path, ".tmp")
    write_fst(mod3, tmp_path, compress = 100)
    file.rename(tmp_path, mod3_path)
    format(month_start, "%Y-%m: ->") %>% paste(basename(mod3_path)) %>% report

    # Cleanup
    rm(mod3)

  }
  rm(mod3_path, month, month_start, year)

}

# Cleanup
rm(add_by_reference, dates_in_month, join_by_reference, mod3_path_for_date, not_key,
   prefix_non_key_cols, safe_to_add, safe_to_join)

report(glue("Done {paste(range(constants$model_years), collapse = '-')}"))
