#
# Create mod4 by joining meteo observations to mod5
#

# Load packages, helper functions, and constants
source("init.R")

report("---")
report("Create mod4 (db to fit model that will predict 200 m residual of 1 km Ta)")


# Create mod4 -----

# ~ 7 min
for (year in constants$model_years) {

  # Skip if data for year has already been joined
  mod4_path <- glue("1.work/joined/mod4_{year}.fst")
  if (file.exists(mod4_path) && !OVERWRITE) {

    next

  }

  glue("Creating mod4 for {year}") %>% report

  # Raise if mod5 has not yet been created
  mod5_path <- glue("1.work/joined/mod5_{year}.fst")
  if (!file.exists(mod5_path)) {

    msg <- glue("No mod5 data for {year} [{mod5_path}]")
    stop(msg)

  }

  # Load mod5
  # ~ 2 min, 11.2 GB
  report("  Loading mod5")
  mod5 <- read_fst(mod5_path, as.data.table = TRUE)

  # Load meteo data
  report("  Loading meteo observations")
  meteo <- glue("1.work/meteo_obs/clean/meteo_data_{year}-clean.fst") %>%
           read_fst(as.data.table = TRUE)
  names(meteo) %>% shorten %>% setnames(meteo, .)
  meteo <- meteo[, .(date, stn_id, lat, lon, tmin, tmean, tmax)]

  report("  Joining meteo observations to mod5")

  # Find unique meteo station locations and convert coordinates to EPSG:3035
  lookup <- meteo[, .(stn_id, lat, lon)] %>%
            unique
  xy <- st_as_sf(lookup, coords = c("lon", "lat"), crs = 4326) %>%
        st_transform(3035) %>%
        st_coordinates %>%
        as.data.table
  lookup[, c("x", "y") := xy]
  rm(xy)

  # Find the grid cell closest to each station location
  if (!exists("grd")) {
    grd <- read_fst("1.work/grids/grd_200m.fst", as.data.table = TRUE)
  }
  grd_pts <- grd[, .(x, y)]
  stn_pts <- lookup[, .(x, y)]
  lookup[, c("grd_idx", "dist") := FNN::get.knnx(grd_pts, stn_pts, k = 1)]
  lookup[, "grd_200m_id" := grd[lookup$grd_idx, id]]
  rm(grd_pts, stn_pts)

  # Drop stations >= 142 m from a grid point (i.e. not contained by a grid cell)
  lookup <- lookup[dist < 142, ]

  # Keep only observations from stations associated with a grid point
  meteo <- meteo[lookup, on = c("stn_id", "lat", "lon")] %>%
           .[, .(date, grd_200m_id, stn_id, stn_lat = lat, stn_lon = lon, tmin, tmean, tmax)]
  rm(lookup)

  # Join grid data to the meteo observations
  mod4 <- mod5[meteo, on = c("date", "grd_200m_id")] %>%
          setkeyv(c("date", "stn_id", "grd_200m_id"))
  rm(meteo, mod5)

  # Calculate the residual of the 1 km temperature predictions
  for (tcol in c("tmin", "tmean", "tmax")) {

    mod4[, glue("res_{tcol}") := get(tcol) - get(glue("pred_{tcol}"))]
    rm(tcol)

  }

  # Save
  # ~ 3 min,
  report("  Saving")
  tmp_path <- glue(mod4_path, ".tmp")
  write_fst(mod4, tmp_path, compress = 100)
  file.rename(tmp_path, mod4_path)
  glue("  -> {mod4_path}") %>% report
  rm(tmp_path)

  # Cleanup
  rm(mod4, mod5_path)

}

if (exists("grd")) {
  rm(grd)
}
rm(mod4_path, year)

report(glue("Done {paste(range(constants$model_years), collapse = '-')}"))
