#
# Create mod5 by joining extracted data and 1 km predictions to the 200 m grid
#

# Load packages, helper functions, and constants
source("init.R")
source("2.scripts/helpers/categorize_clc.R")

report("---")
report("Create mod5 (db to predict 200 m residual of 1 km Ta)")


# Functions -----

join_safely <- function(main_table, new_data) {

  if (!identical(main_table$grd_200m_id, new_data$grd_200m_id)) {

    stop("Cannot join: grd_200m_id does not match")

  }

  cols_to_add <- setdiff(names(new_data), key(main_table))
  dup_cols <- intersect(names(main_table), cols_to_add)

  if (length(dup_cols) > 0) {

    msg <- glue("Cannot join: duplicate columns [{glue_collapse(dup_cols, sep = ' ')}]")
    stop(msg)

  }

  for (col in cols_to_add) {

    main_table[, (col) := new_data[, col, with = FALSE]]

  }

}

# Create mod5 -----

for (year in constants$model_years) {

  # Skip if data for year has already been joined
  mod5_path <- glue("1.work/joined/mod5_{year}.fst")
  if (file.exists(mod5_path) && !OVERWRITE) {

    next

  }

  # Load the 200 km grid / 1 km grid crossref and join spatial data
  if (!exists("grd")) {

    report("  Loading 200 m / 1 km grid crossref")
    grd <- read_fst("1.work/grids/grd_crossref.fst", as.data.table = TRUE)

    report("  Adding 200 m grid x and y coordinates")
    grd <- read_fst("1.work/grids/grd_200m.fst", as.data.table = TRUE,
                     columns = c("id", "x", "y")) %>%
           grd[., on = c(grd_200m_id = "id")]

    report("  Joining climate type")
    climate <- read_fst("1.work/extracts/grd_200m_climate.fst", as.data.table = TRUE)
    join_safely(grd, climate)
    rm(climate)

    report("  Joining elevation")
    elev <- read_fst("1.work/extracts/grd_200m_elev.fst", as.data.table = TRUE)
    join_safely(grd, elev)
    rm(elev)

    report("  Joining population")
    pop <- read_fst("1.work/extracts/grd_200m_pop.fst", as.data.table = TRUE)
    join_safely(grd, pop)
    rm(pop)

  }

  # Create a copy before joining spatiotemporal data
  # This avoids having to re-join the temporally constant data for each year
  mod5 <- copy(grd)

  report("  Joining land cover")
  diffs <- abs(year - constants$clc_years)
  clc_year <- constants$clc_years[diffs == min(diffs)][1]
  clc <- glue("1.work/extracts/grd_200m_clc_{clc_year}.fst") %>%
         read_fst(as.data.table = TRUE) %>%
         categorize_clc
  join_safely(mod5, clc)
  rm(clc, diffs, clc_year)

  # ~ 3.5 min, 7.6 GB
  report("  Joining 1 km predicted temperatures")
  {

    # Raise if mod3 final predictions have not yet been created
    mod3_pred_path <- glue("1.work/predictions/mod3_{year}_pred_1km.fst")
    if (!file.exists(mod3_pred_path)) {

      msg <- glue("No mod3 final predictions for {year} [{mod3_pred_path}]")
      stop(msg)

    }

    # ~ 3 min, 18.4 GB
    report("    Loading...")
    load_cols <- c("date", "grd_1km_id", "tmin", "tmean", "tmax")
    pred <- read_fst(mod3_pred_path, as.data.table = TRUE, columns = load_cols)
    pred_cols <- setdiff(load_cols, key(pred))
    setnames(pred, pred_cols, glue("pred_{pred_cols}"))
    rm(load_cols, pred_cols)

    # ~ 30 sec, 7.6 GB
    report("    Joining...")
    mod5 <- pred[mod5, on = "grd_1km_id"] %>%
            .[, grd_1km_id := NULL]
    rm(pred, mod3_pred_path)

  }

  # Add month to mod5 (used for joining Landsat NDVI and brightness temperature)
  mod5[, month := month(date)]

  # ~ 30 sec, 8.7 GB
  report("  Joining Landsat NDVI")
  {

    if (!exists("ndvi")) {

      # Load Landsat NDVI for all months (constant across years)
      ndvi <- list.files("1.work/extracts/landsat/", full.names = TRUE,
                         pattern = "grd_200m_landsat_ndvi_2000-2018_[0-9]{2}\\.csv") %>%
              lapply(fread, select = c("month", "grd_200m_id", "ndvi", "cover")) %>%
              rbindlist %>%
              setkeyv(c("month", "grd_200m_id"))

      # Drop cell-months with < 50% cover and round NDVI to three decimal places
      ndvi <- ndvi[cover >= 50, .(month, grd_200m_id, ndvi)] %>%
              .[, ndvi := round(ndvi, 3)]

    }

    # Join to mod5 by month
    mod5 <- ndvi[mod5, on = c("month", "grd_200m_id")]

  }

  # ~ 30 sec, 9.4 GB
  report("  Joining Landsat at-satellite brightness temperature")
  {

    if (!exists("bt")) {

      # Load Landsat NDVI for all months (constant across years)
      bt <- list.files("1.work/extracts/landsat/", full.names = TRUE,
                       pattern = "grd_200m_landsat_bt_2000-2018_[0-9]{2}\\.csv") %>%
            lapply(fread, select = c("month", "grd_200m_id", "bt", "cover")) %>%
            rbindlist %>%
            setkeyv(c("month", "grd_200m_id"))

      # Drop cell-months with < 50% cover
      bt <- bt[cover >= 50, .(month, grd_200m_id, bt)]

    }

    # Join to mod5 by month
    mod5 <- bt[mod5, on = c("month", "grd_200m_id")]

  }

  # ~ 20 sec, 9.0 GB
  report("  Ordering columns and setting keys")
  mod5[, month := NULL]
  setcolorder(mod5, c("date", "grd_200m_id", "x", "y", "pred_tmin", "pred_tmean", "pred_tmax",
                      "climate", "elev", "pop", "artificial", "vegetation", "bare", "water", "ndvi",
                      "bt"))
  setkeyv(mod5, c("date", "grd_200m_id"))

  # ~ 2 min, 1.5 GB on disk
  report("  Saving")
  tmp_path <- glue(mod5_path, ".tmp")
  write_fst(mod5, tmp_path, compress = 100)
  file.rename(tmp_path, mod5_path)
  glue("  -> {mod5_path}") %>% report
  rm(tmp_path)

  rm(mod5)

}

# Cleanup
if (exists("grd")) {
  rm(grd, bt, ndvi)
}
rm(mod5_path, year, categorize_clc, join_safely)

report(glue("Done {paste(range(constants$model_years), collapse = '-')}"))
