# Extract data to the 1 km grid

# Load packages, helper functions, and constants
source("init.R")
source("2.scripts/extract/post_process_clc.R")

report("---")
report("Extracting all datasets to the 1 km grid")


# Functions --------------------------------------------------------------------

# Convert extracted data to a raster for visualisation
save_as_raster <- function(dtable, grd_pxs, path, overwrite = FALSE) {

  if (file.exists(path) && overwrite != TRUE) {
    stop("Path exists; use overwrite = TRUE")
  }

  # Paranoia: confirm the data id matches the grid id
  if (!identical(grd_pxs$id, dtable$grd_1km_id)) {
    stop("Cannot create raster: extracted$id != grd_pts$id")
  }

  # Replace the grid id with an index (raster layers must be numeric)
  rast_data <- copy(dtable) %>% setnames("grd_1km_id", "index")
  rast_data[, index := grd_pxs$idx]

  # Exclude any character-type data columns (rasters layers must be numeric)
  char_cols <- sapply(rast_data, class) %>% equals("character") %>% which
  if (length(char_cols) > 0) {

    msg <- names(char_cols) %>%
           paste(collapse = " ") %>%
           paste("Raster cannot contain strings; excluding [", ., "]")
    warning(msg)
    rast_data[, (char_cols) := NULL]

  }

  # Create and save the raster
  rast <- as(grd_pxs, "SpatialPoints") %>% SpatialPixelsDataFrame(data = rast_data) %>% brick
  save_geotiff(rast, path, overwrite = overwrite)

  # Save the layer names
  meta_path <- paste0(path, ".bandnames")
  names(rast) %>% t %>% as.data.frame %>% fwrite(meta_path, col.names = FALSE)

}

# Setup ------------------------------------------------------------------------

# Define the datasets to extract
datasets <- list(
  list(
    name = "climate",
    description = "Joly (2010) climate types",
    data_path = "1.work/climate_types/joly_2010_climate_types_50m.tif",
    out_path = "1.work/extracts/grd_1km_climate.fst",
    extract_fun = function(x) { modal(x, na.rm = TRUE) },  # predominant climate type
    post_process = function(x) { as.factor(x) } # convert to factor
  ),
  list(
    name = "elev",
    description = "EU-DEM v1.1",
    data_path = "1.work/elevation/eu-dem_v1-1/eu_dem_v11_france.tif",
    out_path = "1.work/extracts/grd_1km_elev.fst",
    extract_fun = function(x) { mean(x) %>% round %>% as.integer } # mean elevation
  ),
  list(
    name = "clc_2000",
    description = "Corine Land Cover 2000",
    data_path = "1.work/land_cover/clc_2000_france.tif",
    out_path = "1.work/extracts/grd_1km_clc_2000.fst",
    post_process = post_process_clc # percent of cell occupied by each clc class
  ),
  list(
    name = "clc_2006",
    description = "Corine Land Cover 2006",
    data_path = "1.work/land_cover/clc_2006_france.tif",
    out_path = "1.work/extracts/grd_1km_clc_2006.fst",
    post_process = post_process_clc # percent of cell occupied by each clc class
  ),
  list(
    name = "clc_2012",
    description = "Corine Land Cover 2012",
    data_path = "1.work/land_cover/clc_2012_france.tif",
    out_path = "1.work/extracts/grd_1km_clc_2012.fst",
    post_process = post_process_clc # percent of cell occupied by each clc class
  ),
  list(
    name = "pop",
    description = "INSEE 200 m gridded 2010 population",
    data_path = "1.work/population/insee_population_200m.tif",
    out_path = "1.work/extracts/grd_1km_pop.fst",
    extract_fun = function(x) { sum(x, na.rm = TRUE) %>% round %>% as.integer } # total population
  )
)

# Extract ----------------------------------------------------------------------

# Extract the datasets in sequence
for (params in datasets) {

  report(params$description)

  # Skip the dataset if it has already been extracted
  if (file.exists(params$out_path) && !OVERWRITE) {

    next

  }

  # Load the grid as points and a raster; these will be used to check the dataset's alignment with
  # the grid and create a raster of the extracted data for visualization
  if (!exists("grd_pts")) {

    report("  Loading grid points")
    grd_pts <- readRDS("1.work/grids/grd_1km_pts.rds")
    grd_pxs <- readRDS("1.work/grids/grd_1km_pxs_sinu.rds")
    grd_rast <- raster("1.work/grids/grd_1km_pxs_sinu.tif")

  }

  # Check the dataset's alignment with the grid
  dat <- brick(params$data_path)
  align_check <- compareRaster(grd_rast, dat, extent = FALSE, rowcol = FALSE, crs = TRUE,
                               res = TRUE, orig = TRUE, stopiffalse = FALSE)
  if (align_check) {

    # Data is aligned with grid; extract by points (much faster)

    # Extract the data
    report("  Extracting by points")
    extracted <- raster::extract(dat, grd_pts)

  } else {

    # Data is not aligned with grid; extract by polygons (much slower)

    # Load the grid polygons and split into ncores groups
    if (!exists("grd_pys_splits")) {

      report("  Loading grid polygons")
      grd_pys <- readRDS("1.work/grids/grd_1km_pys.rds")
      grd_pys_splits <- split_into_n(grd_pys["id"], ncores)
      rm(grd_pys)

    }

    # Load the data as a velox object for faster extraction
    report("  Loading data with velox")
    vx <- velox(params$data_path)

    report("  Extracting by polygons")
    extracted <- parallel_velox_extract(vx, grd_pys_splits, fun = params$extract_fun,
                                        ncores = ncores) %>%
                 unlist(recursive = FALSE, use.names = FALSE)

  }

  # Perform post-processing of the extracted values (e.g. replace NA, convert to factor)
  if (!is.null(params$post_process)) {

    report("  Post-processing extracted data")
    extracted <- params$post_process(extracted)

  }

  # Create a data table with each grid cell's id and extracted data
  dtable <- extracted
  if (!is.data.table(dtable)) {
    dtable <- as.data.table(dtable) %>% setnames(params$name)
  }
  dtable[, "grd_1km_id" := grd_pts$id]
  names(dtable) %>% setdiff("grd_1km_id") %>% c("grd_1km_id", .) %>% setcolorder(dtable, .)

  # Save as a raster for visualization
  report("  Saving as raster for visualisation")
  tif_path <- sub("\\.fst$", ".tif", params$out_path)
  if (identical(tif_path, params$out_path)) {
    stop(paste("tif_path == out_path:", params$out_path))
  }
  save_as_raster(dtable, grd_pxs, tif_path, OVERWRITE)
  paste("  ->", tif_path) %>% report

  # Key the data table on grid id and save
  report("  Saving as table")
  setkey(dtable, "grd_1km_id")
  write_fst(dtable, params$out_path, compress = 100)
  paste("  ->", params$out_path) %>% report

  # Cleanup
  if (exists("vx")) { rm(vx) }
  rm(align_check, dat, extracted, dtable, tif_path)

}

# Cleanup
rm(datasets, params, save_as_raster, post_process_clc)

report("Done")
