# Extract MODIS land surface temperature data for the 1 km grid cells
#
# For each satellite (aqua / terra)
# * For each month with data
#   - For each time of day (day / night)
#     + For each date in the month
#         Create a template with grid1km id and date to hold extracted data
#         Extract LST and QA info to the template if data exists for the date
#     + Combine the templates for all dates in the month and save

# Load packages, helper functions, and constants
source("init.R")
source("2.scripts/helpers/modis_lst_qa_code_table.R")

report("---")
report("Extracting MODIS land surface temperature data")


# List the scientific datasets we want to extract
to_extract <- c(
  "LST_Day_1km",
  "LST_Night_1km",
  "QC_Day",
  "QC_Night",
  "Emis_31",
  "Emis_32"
)

# Create a directory to hold the extracted data
lst_out_dir <- "1.work/extracts/modis_lst"
dir.create(lst_out_dir, showWarnings = FALSE)

# List the first day of each month in the model timeperiod
month_starts <- seq.Date(
  from = constants$model_years %>% first %>% paste0("-01-01") %>% as.Date,
  to = constants$model_years %>% last %>% paste0("-12-01") %>% as.Date,
  by = "month"
) %>% format("%Y-%m-%d") # Format as string b/c `for` will convert a date to an integer

# Identify the aqua and terra datasets
datasets <- c(aqua = "MYD11A1.006", terra = "MOD11A1.006")

# Extract data by month
# 240 seconds per year
for (month_start in month_starts) {

  # Convert back to date (passed as character b/c `for` converts dates to integers)
  month_start <- as.Date(month_start)

  # Extract Aqua and Terra data in sequence
  # 120 seconds per satellite
  for (dataset in datasets) {

    satellite <- names(datasets)[datasets == dataset]

    # Define the file to extract to
    out_path <- format(month_start, "%Y-%m") %>%
                paste0("grd_1km_", satellite, "_lst_", ., ".fst") %>%
                file.path(lst_out_dir, .)

    # List all days in the month in the format used in MODIS filenames
    month_days <- seq.Date(from = month_start, length.out = 2, by = "month") %>%
                  last %>%
                  subtract(1) %>%
                  seq.Date(from = month_start, to = ., by = 1) %>%
                  format("A%Y%j")

    # List the data files corresponding to those days
    data_paths <- file.path(constants$data_dir, "lst", dataset, "raw", format(month_start, "%Y")) %>%
                  list.files(full.names = T)
    data_paths <- lapply(month_days, function(d) {
                    data_paths[grepl(d, basename(data_paths))]
                  }) %>% magrittr::set_names(month_days)

    # Skip if no data or already extracted
    if (sapply(data_paths, length) %>% max %>% equals(0)) {

      paste0(format(month_start, "%Y-%m"), " ", satellite, ": no data") %>% report
      next

    } else if (file.exists(out_path) && !OVERWRITE) {

      next

    }

    # Load the 1 km grid and project to MODIS sinusoidal
    if (!exists("grd")) {
      format(month_start, "%Y-%m") %>% paste0(" ", satellite, ": loading 1 km grid") %>% report
      grd <- readRDS("1.work/grids/grd_1km_pts.rds") %>% spTransform(constants$proj_sinu)
    }

    # Load a code table to aid in deciphering MODIS LST QA bit codes
    if (!exists("qa_codes")) {
      qa_codes <- modis_lst_qa_code_table() %>% as.data.table
    }

    # Extract day and night LST, emissivity, and QA for each date in the month
    # Some of these dates may have no data - in that case, return a template filled with NA

    format(month_start, "%Y-%m") %>% paste0(" ", satellite, ": extracting in parallel") %>% report
    result <- mclapply(names(data_paths), function(month_day) {

      # Create a template for the extracted data
      template <- data.table(
        "date" = as.Date(month_day, format = "A%Y%j"),
        "grd_1km_id" = grd$id,
        "day_lst" = NA,
        "day_lst_error" = NA,
        "night_lst" = NA,
        "night_lst_error" = NA,
        "emissivity" = NA,
        "day_emis_error" = NA,
        "night_emis_error" = NA
      )
      # 72.8 MB

      # Get the tiles for the date
      tiles <- data_paths[[month_day]]

      # Extract any data for the date to the template
      if (length(tiles) > 0) {

        # Load all datasets of interest from the tiles, merge into a single stack, and crop to the
        # 1 km grid. It's ok to merge because MODIS tiles align without overlap.
        datasets <- lapply(tiles, function(tile) {
          dat <-  paste0("HDF4_EOS:EOS_GRID:", tile, ":MODIS_Grid_Daily_1km_LST:", to_extract) %>%
                  stack() %>%
                  magrittr::set_names(to_extract)
          dat
        })
        layer_names <- names(datasets[[1]])
        datasets <- do.call(raster::merge, datasets) %>% crop(., grd, snap = "out") %>%
                    magrittr::set_names(layer_names)
        # 16 seconds
        # 54.1 MB

        # Calculate mean of bands 31 and 32 emissivity
        datasets$emissivity <- mean(datasets$Emis_31, datasets$Emis_32)
        datasets <- dropLayer(datasets, c("Emis_31", "Emis_32"))
        # 45.1 MB

        # Confirm the grid is in the same projection as the data
        if (!identical(projection(datasets), projection(grd))) {
          grd <- spTransform(grd, projection(datasets))
        }

        # Extract the data by the reference grid points
        # TODO test disaggregating data and extracting by buffers -> slower but
        # more accurate, is the extra accuracy worth the effort?
        datasets <- raster::extract(datasets, grd, df = TRUE)
        # 7 seconds
        # 30.1 MB

        # Add LST and emissivity to the template
        template[, day_lst := datasets$LST_Day_1km]
        template[, night_lst := datasets$LST_Night_1km]
        template[, emissivity := datasets$emissivity]
        # 80.4 MB

        # Look up the day LST and emissivity error from the QA codes
        errors <-
          data.table("code" = datasets$QC_Day) %>%
          qa_codes[., c("lst_error", "emis_error"), on = "code", with = FALSE]
        template[, c("day_lst_error", "day_emis_error") := errors]
        # 85.4 MB

        # Look up the night LST and emissivity error from the QA codes
        errors <-
          data.table("code" = datasets$QC_Night) %>%
          qa_codes[., c("lst_error", "emis_error"), on = "code", with = FALSE]
        template[, c("night_lst_error", "night_emis_error") := errors]
        # 90.4 MB

      }
      # 40 seconds
      # 90.4 MB

      # Return the template (which may have no data)
      template

    }, mc.cores = ncores) %>% rbindlist
    # 60 seconds
    # 1.45 GB

    # Key on date then grd_1km id
    setkeyv(result, c("date", "grd_1km_id"))

    # Save the extracted data
    paste0(format(month_start, "%Y-%m"), " ", satellite, ": saving") %>% report
    write_fst(result, out_path, compress = 100)
    paste("  ->", out_path) %>% report
    # 60 seconds, 135 MB for RDS
    # 40 seconds, 210 MB for fst compress = 100

    # Clear memory
    rm(result)

  }

}

# Cleanup
rm(data_paths, dataset, datasets, lst_out_dir, month_days, month_start, month_starts,
   out_path, satellite, to_extract, modis_lst_qa_code_table)
if (exists("grd")) {
  rm(grd, qa_codes)
}

report("Done")
