# Extract MODIS NDVI data for the 1 km grid cells
#
# For each satellite (aqua / terra)
# * For each month with data
#   - Load and properly scale the monthly NDVI data
#   - Extract the NDVI and QA data by the 1 km grid points
#   - Add the grd_1km and save

# Load packages, helper functions, and constants
source("init.R")
source("2.scripts/helpers/modis_ndvi_qa_code_table.R")

report("---")
report("Extracting MODIS NDVI data")


########################
# MODIS Monthly 1km NDVI
########################

# Create a directory to hold the extracted data
ndvi_out_dir <- "1.work/extracts/modis_ndvi"
dir.create(ndvi_out_dir, showWarnings = FALSE)

# Identify the aqua and terra datasets
datasets <- c(aqua = "MYD13A3.006", terra = "MOD13A3.006")

# Iterate over years to extract the data
# This limits parallelization to 24, but allows progress reporting
for (year in constants$model_years) {

  # List all data files for the year (Aqua and Terra)
  data_paths <- file.path(constants$data_dir, "ndvi", datasets, "raw", year) %>%
                list.files(pattern = "\\.hdf$", full.names = TRUE)

  # Split the files by date and satellite
  data_paths %<>% basename %>%
                  strsplit("\\.") %>%
                  sapply(function(x) { paste(x[1:2], collapse = ".") }) %>%
                  split(data_paths, .)

  # Determine the paths to extract the data to
  out_paths <- map_chr(data_paths, function(dp) {
                 dataset <- dp[1] %>% dirname %>% dirname %>% dirname %>% basename
                 satellite <- names(datasets)[datasets == dataset]
                 month <- dp[1] %>%
                          basename %>%
                          strsplit("\\.") %>%
                          sapply("[", 2) %>%
                          as.Date(format = "A%Y%j") %>%
                          format("%Y-%m")
                 paste0("grd_1km_", satellite, "_ndvi_", month, ".fst") %>%
                   file.path(ndvi_out_dir, .)
               })

  # Skip if data already extracted
  if (all(file.exists(out_paths)) && !OVERWRITE) {

    next

  }

  # Load the 1 km grid as spatial pixels in the MODIS sinusoidal projection
  if (!exists("grd")) {
    report("Loading 1 km grid")
    grd <- readRDS("1.work/grids/grd_1km_pxs_sinu.rds")
  }

  # Load a code table to aid in deciphering MODIS NDVI QA bit codes
  # 16 seconds
  if (!exists("qa_codes")) {
    report("Loading QA code table")
    qa_codes <- modis_ndvi_qa_code_table() %>% as.data.table
  }

  # Extract all missing data in parallel
  missing <- names(data_paths)[!file.exists(out_paths) | OVERWRITE]
  mclapply(missing, function(m) {

    tiles <- data_paths[[m]]
    out_path <- out_paths[m]

    # Skip if data already extracted
    if (file.exists(out_path) && !OVERWRITE) { return(NULL) }

    # Create a template to hold the extracted data
    result <- data.table(
      "grd_1km_id" = grd$id,
      "ndvi" = NA,
      "ndvi_usefulness" = NA
    )
    # 12.6 MB

    # Load and properly scale monthly NDVI and QA, merge into a single stack, and crop to the grid.
    # It's ok to merge because MODIS tiles align without overlap.
    datasets <- lapply(tiles, function(tile) {

      # Identify the monthly NDVI dataset
      ndvi_sds <- paste0("HDF4_EOS:EOS_GRID:", tile, ":MOD_Grid_monthly_1km_VI:1 km monthly NDVI")

      # Get the scale factor from the dataset metadata
      ndvi_metadata <- gdalinfo(ndvi_sds)
      ndvi_scale_factor <-
        grep("  scale_factor=", ndvi_metadata) %>%
        ndvi_metadata[.] %>%
        gsub("  scale_factor=(.+)", "\\1", .) %>%
        as.numeric

      # Load the NDVI dataset as a raster and apply the scale factor
      # The raster package does this automatically, but for some reason in the metadata the NDVI
      # scale factor is stored as 10000 when it should be 0.0001. So to properly scale NDVI we must
      # divide by the square of the scale factor
      ndvi <- raster(ndvi_sds) / (ndvi_scale_factor^2)

      # Load the QA data, stack with NDVI, and return the stack
      paste0("HDF4_EOS:EOS_GRID:", tile, ":MOD_Grid_monthly_1km_VI:1 km monthly VI Quality") %>%
        raster %>%
        stack(ndvi, .)

    }) %>% do.call(raster::merge, .) %>% crop(., grd, snap = "out")
    names(datasets) <- c("ndvi", "qa")

    # Confirm the grid is in the same projection as the data
    if (!identical(projection(datasets), projection(grd))) {
      grd <- spTransform(grd, projection(datasets))
    }

    # Extract NDVI and QA by the 1 km grid points
    # TODO test extracting data by buffers -> slower but more accurate, is it worth the effort?
    extracted <- raster::extract(datasets, grd, df = TRUE)

    # Add NDVI to the result template
    result[, ndvi := extracted$ndvi]

    # Look up the usefulness from the QA codes table and add to the template
    qa <- data.table("code" = extracted$qa) %>% qa_codes[., usefulness, on = "code"]
    result[, ndvi_usefulness := qa]
    rm(qa)

    # Key on grd_1km id
    result %>% setkey("grd_1km_id")

    # Save the extracted data
    write_fst(result, out_path, compress = 100)

    # Return NULL to save memory
    NULL
  }, mc.cores = ncores, mc.preschedule = FALSE)

}

if (exists("grd")) {
  rm(grd, qa_codes)
}
rm(ndvi_out_dir, datasets, year, data_paths, out_paths, modis_ndvi_qa_code_table)

report(glue("Done {paste(range(constants$model_years), collapse = '-')}"))
