# Create a 1 square km buffer around each point of the 1 km reference grid.
# The buffers are used to extract data for association with the grid points.
#
# * Project the grid to EPSG:3035, an equal-area projection
# * Create square and round buffers of various widths around each grid point
# * Save in EPSG:2154

# Load packages, helper functions, and constants
source("init.R")

report("---")
report("Generating buffers around the 1 km grid points")


# Skip if all output files exist
out_paths <- c("grd_1km_pys.rds", "grd_1km_pys.shp") %>%
             file.path("1.work/grids", .)
skip_if_exist(out_paths)

# Load the 1 km grid and project to EPSG:3035 (LAEA Europe)
# This is an equal area projection, which will ensure buffers have same area throughout France
grid_3035 <- readRDS("1.work/grids/grd_1km_pts.rds") %>%
             spTransform(constants$epsg_3035)

# Define a function to buffer in parallel
buffer_parallel <- function(splits, width, capStyle, ncores, out_proj4) {

  # Buffer the points in each group, reproject to EPSG:2154, and collect the
  # results into a single SpatialPolygonsDataFrame
  buffers <- mclapply(splits, function(splt) {
    gBuffer(splt, width = width * 500, capStyle = capStyle, byid = TRUE) %>%
      spTransform(out_proj4)
  }, mc.cores = ncores) %>% do.call(rbind, .)

  # Set the plot order and return the resulting buffers
  buffers@plotOrder <- 1:nrow(buffers)
  buffers

}

# Split the grid points into ncores groups for parallel processing
splits <- split_into_n(grid_3035, ncores)

# Generate 1 km square buffers around each grid point
paste("Creating 1 km square buffers") %>% report
buffers <- buffer_parallel(splits, 1, "SQUARE", ncores, constants$epsg_2154)

# Save as RDS
report("Saving")
path <- file.path(constants$grids_dir, "grd_1km_pys.rds")
saveRDS(buffers, path)
paste("->", path) %>% report

# Save as shapefile
path <- "1.work/grids/grd_1km_pys.shp"
writeOGR(buffers, dsn = path, driver = "ESRI Shapefile", layer = "grd_1km_pys")
paste("->", path) %>% report

# # Generate circular buffers with radius of 1, 25, and 50 km around each grid point
# for (dist in c(1, 25, 50)) {
#   paste("Creating", dist, "km radius buffers") %>% report
#   buffers <- buffer_parallel(splits, dist * 2, "ROUND", ncores, constants$epsg_2154)
#   path <- file.path(constants$grids_dir, paste0("grd_1km_rnd_", dist, "km.rds"))
#   paste("  Saving to", path) %>% report
#   saveRDS(buffers, path)
#   rm(buffers)
# }

report("Done")
