# Check for misalignment between the different MODIS products

source("init.R")

report("---")
report("Checking alignement of MODIS products")


# Function to return the first tile matching a regex for a MODIS product
first_tile <- function(path, regexp) {

  year_dir <- dir(path, full.names = TRUE)[1]
  tiles <- list.files(year_dir, full.names = TRUE, pattern = regexp)
  tiles[1]

}

# Function to compare the alignment of two rasters
check_alignment <- function(a, b) {
  d <- c(
    max(origin(a) - origin(b)),
    max(res(a) - res(b)),
    max(coordinates(a) - coordinates(b))
  )
  names(d) <- c("origin", "resolution", "coordinates")
  d
}

# MODIS products to check
products <- list(
  "aqua_lst"   = file.path(constants$data_dir, "lst", "MYD11A1.006", "raw"),
  "aqua_ndvi"  = file.path(constants$data_dir, "ndvi", "MYD13A3.006", "raw"),
  "terra_lst"  = file.path(constants$data_dir, "lst", "MOD11A1.006", "raw"),
  "terra_ndvi" = file.path(constants$data_dir, "ndvi", "MOD13A3.006", "raw")
)

# Get the location of the first Aqua LST tile
tile_id <-
  first_tile(products$aqua_lst, "\\.hdf$")[1] %>%
  regmatches(., regexpr("h\\d{2}v\\d{2}", .))

# Load the the first dataset of the corresponding tile for each MODIS product
rasters <- sapply(products, function(path) {
  paste0(tile_id, ".+\\.hdf$") %>%
  first_tile(path, .) %>%
  get_subdatasets %>%
  .[1] %>%
  raster
})
rm(tile_id, products)

# Compare alignment
results <- rbind(
  "Aqua LST  - Terra LST"  = check_alignment(rasters$aqua_lst, rasters$terra_lst),
  "Aqua NDVI - Terra NDVI" = check_alignment(rasters$aqua_ndvi, rasters$terra_ndvi),
  "Aqua LST  - Aqua  NDVI" = check_alignment(rasters$aqua_lst, rasters$aqua_ndvi),
  "Terra LST - Terra NDVI" = check_alignment(rasters$terra_lst, rasters$terra_ndvi)
) %>% as.data.frame
report(results)

# Confirm the rasters are aligned to within 1 mm
misalignment <- max(results$coordinates)
if (misalignment < 0.001) {
  report("MODIS products are aligned to within 1 mm")
} else {
  stop(paste("MODIS products are misaligned by", round(misalignment, 2), "m"))
}

# Cleanup
rm(rasters, results, misalignment, check_alignment, first_tile)

report("Done")
