#
# Associate stations with nearest 1 km and 200 m grid cell
#

# Load packages, helper functions, and constants
source("init.R")

report("---")
report("Matching meteo stations to 200 m and 1 km grid cells")


# Skip if all output files exist
out_path <- "1.work/grids/stations_gridded.shp"
skip_if_exist(out_path)

# Load 200 m grid raster
grd_rst <- brick("1.work/grids/grd_200m_inputs.tif")
names(grd_rst) <- read.csv("1.work/grids/grd_200m_inputs.tif.bandnames") %>% names

# Load meteo station IDs and locations
stn_pts <- fread("1.work/meteo_obs/stations_after_cleaning.csv") %>%
           st_as_sf(coords = c("longitude", "latitude"), crs = 4326) %>%
           st_transform(projection(grd_rst))

# Polygonize urban areas
urb <- mask(grd_rst$clump_pop, grd_rst$include) %>%
       magrittr::set_names("pop") %>%
       rasterToPolygons(dissolve = TRUE) %>%
       st_as_sf %>%
       st_cast("POLYGON")
rm(grd_rst)

# Add an index
urb$idx <- 1:nrow(urb)

# Add name of largest city intersecting each urban area
cities <- st_read("0.raw/ign/ADMIN-EXPRESS_2-0__SHP__FRA_2019-01-18/ADMIN-EXPRESS/1_DONNEES_LIVRAISON_2019-01-18/ADE_2-0_SHP_LAMB93_FR/COMMUNE.shp",
                  stringsAsFactors = FALSE) %>%
          st_transform(st_crs(urb)) %>%
          st_intersection(urb)
city_names <- cities[, c("idx", "pop", "NOM_COM", "POPULATION")] %>%
              as.data.table %>%
              .[, geometry := NULL] %>%
              .[order(-POPULATION), lapply(.SD, first), by = idx]
urb$name[city_names$idx] <- city_names$NOM_COM

# Save urban area polygons with names
st_write(urb, "1.work/grids/grd_200m_urb_areas.shp", delete_dsn = TRUE)
rm(cities, city_names)


# Calculate the distance from each station to the nearest urban area
stn_urb <- st_nn(stn_pts, urb, k = 1, returnDist = TRUE)
stn_pts$urb_name <- urb$name[unlist(stn_urb$nn)]
stn_pts$urb_dist <- round(stn_urb$dist)
rm(stn_urb, urb)

# Associate each station with the nearest 1 km and 200 m grid cell
# NOTE: use short column names b/c ESRI shapefile column names are limited to 10 chars!
grd_1km <- st_read("1.work/grids/grd_1km_pts.shp")
stn_grd_1km <- st_crs(grd_1km) %>%
               st_transform(stn_pts, .) %>%
               st_nn(grd_1km, k = 1) %>%
               unlist(use.names = FALSE)
stn_pts$grd_1km <- grd_1km$id[stn_grd_1km]
rm(grd_1km, stn_grd_1km)

# Associate each station with the nearest 1 km and 200 m grid cell
# NOTE: use short column names b/c ESRI shapefile column names are limited to 10 chars!
grd_200m <- st_read("1.work/grids/grd_200m_pts.shp")
stn_grd_200m <- st_crs(grd_200m) %>%
                st_transform(stn_pts, .) %>%
                st_nn(grd_200m, k = 1) %>%
                unlist(use.names = FALSE)
stn_pts$grd_200m <- grd_200m$id[stn_grd_200m]
rm(grd_200m, stn_grd_200m)

# Save the stations with distances and grid ids
st_write(stn_pts, out_path, delete_dsn = TRUE)
