# Evaluate predictions in data.table that also contains observations

source("2.scripts/helpers/rmse.R")

assess_preds <- function(dtable, obs_col, pred_col, loc_id_col) {

  # Make an empty list to hold summary statistics
  stats <- list()

  # Mean, sd, and number of observations
  stats$obs.mean <- dtable[, get(obs_col)] %>% mean
  stats$obs.sd   <- dtable[, get(obs_col)] %>% sd
  stats$obs.n    <- dtable[, .N]

  # RMSE and MAE
  res <- dtable[, get(obs_col) - get(pred_col)]
  stats$rmse <- rmse(res)
  stats$mae <- abs(res) %>% mean

  # R^2, intercept + slope
  # obs_col ~ pred_col
  all_lm <- paste(obs_col, "~", pred_col) %>% as.formula %>% lm(data = dtable)
  stats$r2       <- summary(all_lm)$r.squared
  stats$i        <- summary(all_lm)$coef["(Intercept)", "Estimate"]
  stats$i.se     <- summary(all_lm)$coef["(Intercept)", "Std. Error"]
  stats$slope    <- summary(all_lm)$coef[pred_col, "Estimate"]
  stats$slope.se <- summary(all_lm)$coef[pred_col, "Std. Error"]

  # Spatial R^2, RMSE, MAE
  # For each station, mean obs_col ~ mean pred_col
  spatial <- dtable[, .(
                         obs_bar  = mean(get(obs_col)),
                         pred_bar = mean(get(pred_col))
                       ), by = loc_id_col]
  spatial_lm <- lm(obs_bar ~ pred_bar, data = spatial)
  stats$r2.space <- summary(spatial_lm)$r.squared
  stats$rmse.space <- spatial_lm %>% residuals %>% rmse
  stats$mae.space <- spatial_lm %>% residuals %>% abs %>% mean

  # Temporal R^2, RMSE, MAE
  # For each station, delta obs_col vs delta pred_col
  # Delta obs_col  = obs_col  - mean obs_col  at the station
  # Delta pred_col = pred_col - mean pred_col at the station
  temporal <- dtable[spatial, .(
                                 obs = get(obs_col),
                                 obs_bar,
                                 pred = get(pred_col),
                                 pred_bar
                               ), on = loc_id_col]
  temporal[, delta_obs := obs - obs_bar]
  temporal[, delta_pred := pred - pred_bar]
  temporal_lm <- lm(delta_obs ~ delta_pred, data = temporal)
  stats$r2.time <- summary(temporal_lm)$r.squared
  stats$rmse.time <- temporal_lm %>% residuals %>% rmse
  stats$mae.time <- temporal_lm %>% residuals %>% abs %>% mean

  # Mean, sd, quantiles of residuals
  stats$res.mean <- mean(res)
  stats$res.sd <- sd(res)
  stats$res.median <- median(res)
  stats$res.iqr <- IQR(res)
  stats[c("res.min", "res.1Q", "res.3Q", "res.max")] <- quantile(res, c(0, 0.25, 0.75, 1))

  # Return the statistics as a named vector
  unlist(stats)

}
