# Create folds test sets for the dataset (all records not in a test set are in the train set)
make_cv_test_sets <- function(dataset, folds = 10, method = "kfold", group_by = NULL, seed = NULL) {
  if (!is.null(seed)) { set.seed(seed) }

  # Index the records
  index <- 1:nrow(dataset)

  # K-fold cross validation: randomly assign each record to one of folds test sets
  if (method == "kfold") {
    test_sets <- suppressWarnings(split(sample(index), 1:folds)) # don't warn if index %% folds != 0
  }
  # Monte Carlo cross validation: randomly sample 1 / folds records; repeat folds times
  else if (method == "monte carlo") {
    test_sets <- lapply(1:folds, function(fold) {
      sample(index, trunc(length(index) / folds))
    })
  }
  # K-fold grouped cross-validation: randomly assign groups of records to folds test sets
  # Note: dataset must be a data.frame; groups may vary in size (will warn if they vary by > 5%)
  else if (method == "kfold grouped") {
    if (length(group_by) != 1 || !(group_by %in% names(dataset))) {
      stop("`group_by` must be the name of a `dataset` column")
    }

    # Randomly assign groups to folds test sets, attempting to create sets of roughly equal size
    index <- dataset[, .(i = 1:nrow(dataset), group_id = get(group_by))]
    records_per_group <- index[, .N, by = "group_id"]
    n_groups <- nrow(records_per_group)
    if (n_groups < folds) {
      stop(paste0("Cannot split ", n_groups, " groups into ", folds, " folds"))
    }
    i <- 0
    coef_var <- 1
    while (coef_var > 0.05 && i < 10) {
      # Assign each grouping id to one of folds test sets
      test_sets <- suppressWarnings(split(sample(nrow(records_per_group)), 1:folds))
      test_sets <- lapply(test_sets, function(splt) { records_per_group[splt] })

      # Calculate the size of each test set and the coefficient of variation of the sizes
      records_per_split <- sapply(test_sets, function(dt) { sum(dt$N) })
      coef_var <- sd(records_per_split) / mean(records_per_split)
      i <- i + 1
    }
    if (coef_var > 0.05) {
      rnd <- round(coef_var, 3)
      warning(paste0("Test sets vary in size by > 5% (coef. var. = ", rnd, ")"))
    } else if (i > 1) {
      warning(paste(i, "tries were needed to create test sets of roughly equal size"))
    }

    # Replace each group id with the row indexes of its records
    test_sets <- lapply(test_sets, function(dt) {
      index[dt, i, on = "group_id"]
    })
  }
  else {
    stop(paste("Unrecognized cross-validation method:", method))
  }

  # Return a list with the fold number and test indexes for each fold
  lapply(1:folds, function(fold) {
    list(fold = fold, test = test_sets[[fold]])
  })
}
