# Assess the m2 model predictions
# Goal is to determine whether smoothing with GAMs improves predictions or reduces spatial
# autocorrelation of residuals

library(magrittr)
library(parallel)
library(data.table)
library(fst)
library(nlme)
library(mgcv)
library(spdep)   # Moran's I
library(sp)      # for visualisation
library(raster)  # for visualisation
library(mapview) # for visualisation

source("2.scripts/helpers/constants.R")
source("2.scripts/helpers/get_ncores.R")
source("2.scripts/helpers/report.R")
source("2.scripts/helpers/abbreviate_colnames.R")
source("2.scripts/helpers/assess_preds.R")

# Setup -----------------------------------------------------------------------

ncores <- get_ncores()
threads_fst(ncores)

year <- 2008
t_col <- "tmin"
lst_col <- "aqua_night_lst"
paste(t_col, year, ": does smoothing reduce spatial autocorrelation of residuals?") %>% report

# Load mod2 data including pred_m2 and pred_m2_rough
# ~ 2 min, ~ 7.5 GB
path <-
  paste("mod2", t_col, year, "pred_m2_final.fst", sep = "_") %>%
  file.path(constants$predictions_dir, .)
paste(year, t_col, ": loading", basename(path)) %>% report
mod2 <- read_fst(path, as.data.table = TRUE)
mod2[, c("pred_m2_smooth", "pred_m2_final") := NULL] # remove inaccurate predictions

# Load the m2 model and reconstruct its formula
# ~ 6 min, ~ 11.5 GB
path <-
  paste("m2", t_col, year, "model.rds", sep = "_") %>%
  file.path(constants$models_dir, .)
paste(year, t_col, ": loading", basename(path)) %>% report
m2 <- readRDS(path)
m2_fml <- list(fixed = as.formula(m2$terms))
attributes(m2_fml$fixed) <- NULL

# Load the splines
# ~ 4 min, ~ 8 GB
path <-
  paste("m2", t_col, year, "splines.rds", sep = "_") %>%
  file.path(constants$models_dir, .)
paste(year, t_col, ": loading", basename(path)) %>% report
splines <- readRDS(path)

# Load temperature observations and station id from mod1
mod1_path <-
  sub("_lst$", "", lst_col) %>%
  paste0("mod1_", ., "_", year, ".fst") %>%
  file.path(constants$joined_dir, .)
col_dict <- read_fst(mod1_path, to = 1) %>% names
names(col_dict) <- abbreviate_colnames(col_dict)
cols_to_load <- col_dict[c("date", "grd_1km_id", "stn_id", t_col)]
paste(year, t_col, ": loading", basename(mod1_path)) %>% report
mod1 <-
  read_fst(mod1_path, as.data.table = TRUE, columns = cols_to_load) %>%
  setnames(names(cols_to_load))
mod1[, date := format(date, "%Y-%m-%d")] %>% setkeyv(c("date", "grd_1km_id"))

# Generate pred_m2_final from m2 and the splines
pred_m2_tps <- lapply(splines, function(tps) { tps$fitted.values }) %>% unlist %>% as.vector
pred_m2_mix <- predict(m2) %>% as.vector
mod2[, pred_m2_final := pred_m2_tps + pred_m2_mix]
rm(pred_m2_tps, pred_m2_mix)#, m2, m2_fml, splines)
gc()


# Join the various m2 predictions from mod2 to the temperature data from mod1, omit NA, and take
# the mean of every value for each location-day (b/c there are stations that have same location)
# FIXME TODO why are there rows without data?
# FIXME TODO really deal with stations that share a location
paste(year, t_col, ": joining m2 predictions to observed temperature") %>% report
dat <- mod2[mod1, on = key(mod1)] %>% na.omit
dat <- dat[, lapply(.SD, mean), by = .(date, grd_1km_id)]

# Calculate residuals for each model
paste(year, t_col, ": calculating m2 residuals") %>% report
pred_cols <- c("pred_m2", "pred_m2_rough", "pred_m2_final")
res_cols <- sub("pred_", "res_", pred_cols)
for (i in 1:length(pred_cols)) {
  pred_col <- pred_cols[i]
  res_col <- res_cols[i]
  dat[, (res_col) := get(t_col) - get(pred_col)]
  rm(pred_col, res_col)
}

# Which predictions are closest to the observed values? -----------------------

paste(year, t_col, ": calculating model summary statistics") %>% report
stats <- lapply(pred_cols, function(pred_col) {
  assess_preds(dtable = dat, obs_col = t_col, pred_col = pred_col, loc_id_col = "grd_1km_id")
})
names(stats) <- pred_cols
stats_smry <- do.call(rbind, stats)
print("Which m2 model produces predictions that are closest to the observed values?")
print(stats_smry[, c("r2", "rmse", "r2.space", "r2.time", "rmse.space", "rmse.time")] %>% round(3))
print("Values for pred_m2 are high b/c IDW temperature is equal to observed value at stations")
print("")

# tmin 2008
#                   r2  rmse r2.space r2.time rmse.space rmse.time
# pred_m2        0.919 1.801    0.895   0.937      1.499     1.456
# pred_m2_rough  0.975 1.050    0.957   0.998      0.956     0.245
# pred_m2_final  0.974 1.042    0.957   0.998      0.960     0.275


# Does smoothing with GAM reduce spatial autocorrelation of residuals? --------
# ~ 35 min for 3 models

paste(year, t_col, ": checking daily residuals for autocorrelation") %>% report
autocorr <- lapply(res_cols, function(res_col) {
  paste(year, t_col, res_col, ": calculating daily Moran's I of residuals") %>% report
  split(dat, by = "date") %>%
  mclapply(function(splt) {
    # Convert to SpatialPointsDataFrame
    coordinates(splt) <- ~ grd_1km_x + grd_1km_y
    proj4string(splt) <- constants$epsg_2154

    # Weight by inverse distance
    weights <- 1 / (coordinates(splt) %>% dist %>% as.matrix)
    diag(weights) <- 0

    # Calculate Moran's I by random permutation
    # Using permutation is faster when there are many points and will not fail if the distribution
    # of the residuals does not meet test assumptions. Alex's code uses 10000 permutations but the
    # GeoDa guide suggests 999 is usually sufficient; use the max possible if nrow(splt) < 7:
    # https://geodacenter.github.io/workbook/5a_global_auto/lab5a.html#permutation-inference
    n_perm <- nrow(splt) %>% lfactorial %>% exp %>% min(999)
    result <- moran.mc(splt[[res_col]], mat2listw(weights), n_perm)

    # Return Moran's I and its pseudo p value. The latter is the fraction of permutations which
    # had an I equal to or greater than the observed I (each permutation is created by randomly
    # assigning the observed values to the points)
    data.table(
      date = splt$date[1],
      I = round(result$statistic, 3),
      p = result$p.value,
      n = nrow(splt)
    )
  }, mc.cores = ncores, mc.preschedule = FALSE) %>%
  rbindlist
})
names(autocorr) <- res_cols
autocorr_smry <- sapply(autocorr, function(dtable) { summary(dtable$I) })

# Save the list of Moran's I
paste(year, t_col, ": saving Moran's I") %>% report
path <-
  paste("mod1", t_col, year, "autocorr.rds", sep = "_") %>%
  file.path(constants$predictions_dir, .)
saveRDS(autocorr, path)
paste(year, t_col, "  ->", basename(path)) %>% report

print("Are the models' residuals spatially autocorrelated?")
print(autocorr_smry %>% round(3))
print("")

# tmin 2008
#         res_m2 res_m2_rough res_m2_final
# Min.    -0.998       -0.397       -0.503
# 1st Qu.  0.034        0.020        0.022
# Median   0.056        0.026        0.029
# Mean     0.066        0.029        0.032
# 3rd Qu.  0.094        0.035        0.039
# Max.     0.289        0.121        0.128

# Are there patterns to residuals on days with high m2 autocorrelation? -------

paste(year, t_col, ": Mapping three days when m2 residuals are most autocorrelated") %>% report
test_dates <- autocorr$res_m2[order(I), ]$date %>% tail(3) %>% sort
for (test_date in test_dates) {
  # Get the Moran's I of the residuals of each model on the test date
  test_i <-
    lapply(autocorr, function(dtable) { dtable[date == test_date, ] }) %>%
    rbindlist %>%
    as.data.frame %>%
    set_rownames(res_cols)
  print("Example day when m2 residuals are spatially autocorrelated")
  print(test_i)
  print("")

  # Map the residuals of each model on the test date
  points <- dat[date == test_date, ]
  coordinates(points) <- ~ grd_1km_x + grd_1km_y
  proj4string(points) <- constants$epsg_2154
  map <- mapview(points, zcol = res_cols, layer.name = paste(res_cols, test_date), legend = T,
                 col.regions = colorRampPalette(c("blue", "white", "red")), alpha.regions = 1)
  print(map)
}

# tmin 2008
#                     date     I     p   n
# res_m2        2008-03-09 0.289 0.001 385
# res_m2_rough  2008-03-09 0.009 0.097 385
# res_m2_final  2008-03-09 0.012 0.051 385
#
# res_m2        2008-05-28 0.236 0.001 357
# res_m2_rough  2008-05-28 0.035 0.001 357
# res_m2_final  2008-05-28 0.051 0.001 357
#
# res_m2        2008-12-30 0.230 0.001 703
# res_m2_rough  2008-12-30 0.058 0.001 703
# res_m2_final  2008-12-30 0.034 0.001 703


# Are there patterns to residuals on days with high m2 autocorrelation? -------

paste(year, t_col, ": Mapping three days when m2 residuals are most autocorrelated") %>% report
test_dates <-
  autocorr[names(autocorr) != "res_m2"] %>%
  lapply(function(dtable) {
    dtable[n >= 100, ][order(I), ]$date %>% tail(2)
  }) %>% unlist %>% unique %>% sort
for (test_date in test_dates) {
  # Get the Moran's I of the residuals of each model on the test date
  test_i <-
    lapply(autocorr, function(dtable) { dtable[date == test_date, ] }) %>%
    rbindlist %>%
    as.data.frame %>%
    set_rownames(res_cols)
  print("Example day when m2 residuals are spatially autocorrelated")
  print(test_i)
  print("")

  # Map the residuals of each model on the test date
  points <- dat[date == test_date, ]
  coordinates(points) <- ~ grd_1km_x + grd_1km_y
  proj4string(points) <- constants$epsg_2154
  map <- mapview(points, zcol = res_cols, layer.name = paste(res_cols, test_date), legend = T,
                 col.regions = colorRampPalette(c("blue", "white", "red")), alpha.regions = 1)
  print(map)
}

# tmin 2008
#                     date     I     p   n
# res_m2        2008-04-02 0.184 0.001 167
# res_m2_rough  2008-04-02 0.094 0.001 167
# res_m2_final  2008-04-02 0.122 0.001 167
#
# res_m2        2008-11-05 0.099 0.002 124
# res_m2_rough  2008-11-05 0.121 0.001 124
# res_m2_final  2008-11-05 0.089 0.003 124
#
# res_m2        2008-12-01 0.088 0.001 168
# res_m2_rough  2008-12-01 0.076 0.002 168
# res_m2_final  2008-12-01 0.100 0.001 168


# Conclusions -----------------------------------------------------------------

# * Smoothing with GAMs does not substantially improve R2 or RMSE
# * m2 residuals generally have less spatial autocorrelation than m2 residuals
# * There are sometimes patterns in m2 residuals, but autocorrelation remains low (I < 0.13)
# * Smoothing with GAMs does not change patterns or reduce spatial autocorrelation of residuals

report("Done")
