#
# Assess all model predictions
#

# Load packages, helper functions, and constants
source("init.R")

report("---")
report("Assessing model predictions")


# Skip if all output files exist -----
out_paths <- range(constants$model_years) %>%
             paste(collapse = "-") %>%
             glue("3.outcomes/model_performance/{c('m1', 'm4')}_{yrs}_cv_perf.csv", yrs = .)
if (all(file.exists(out_paths)) && !OVERWRITE) {
  rm(out_paths)
  stop("Skip existing")
}


# Functions -----

# Calculate and format performance metrics for a list of data.tables
eval_perf <- function(dtables, lab_cols, obs_col, pred_col, loc_id_col, cores = 1) {

  report(glue("  [{length(dtables)} groups] [{cores} cores]"))

  # Calculate performance metrics
  perf <- mclapply(dtables, assess_preds, obs_col = obs_col, pred_col = pred_col,
                   loc_id_col = loc_id_col, mc.cores = cores, mc.preschedule = FALSE)

  # Set up labels
  labs <- names(perf) %>%
          tstrsplit("\\.") %>%
          as.data.table %>%
          setnames(lab_cols)

  # Convert the list of metrics into a labeled data.table
  perf %<>% lapply(as.list) %>%
            lapply(as.data.table) %>%
            rbindlist %>%
            round(7) %>%
            cbind(labs, .)

  # Return
  perf

}

# Convert a vector of dates to seasons
to_season <- function(d) {

  seasons <- rep(NA, length(d))
  months <- month(d)
  seasons[months %in% c(12, 1, 2)] <- "winter"
  seasons[months %in% 3:5] <- "spring"
  seasons[months %in% 6:8] <- "summer"
  seasons[months %in% 9:11] <- "fall"
  factor(seasons, levels = c("winter", "spring", "summer", "fall"))

}

# Convert a vector of distance to nearest urban area to urban / rural / peri
to_urbanity <- function(urb_dist) {

  urbanity <- rep(NA, length(urb_dist))
  urbanity[urb_dist == 0] <- "urban"
  urbanity[urb_dist > 0 & urb_dist <= 5000] <- "peri"
  urbanity[urb_dist > 5000] <- "rural"
  factor(urbanity, levels = c("urban", "peri", "rural"))

}

# Load saved CV predictions
load_cv <- function(path, model, yrs, tvars, force = FALSE) {

  if (!model %in% c("m1", "m4")) {
    stop(glue("Don't know how to load CV predictions for {model}"))
  }

  year_range <- paste(range(yrs), collapse = "-")

  if (file.exists(path) && !force) {

    report(glue("Loading {model} CV predictions for {year_range}"))
    cv <- read_fst(path, as.data.table = TRUE)
    return(cv)

  }

  report(glue("Combining {model} CV predictions for {year_range}"))
  cv <- lapply(tvars, function(tvar) {

          cv <- lapply(yrs, function(yr) {

                  if (model == "m1") {

                    # Identify the best model
                    prd <- if_else(tvar == "tmax", "day", "night")
                    sat <- if_else(tvar == "tmean" || yr < 2003, "terra", "aqua")

                    # HACK old m1 cv uses different column names
                    if (yr < 2017) {
                      ref_cols <- c("date", "lst_1km_id", "stn_id", "climate_type")
                    } else {
                      ref_cols <- c("date", "grd_1km_id", "stn_id", "climate")
                    }
                    pred_cols <- c(as.character(tvar), "pred_m1_cv")

                    cv <- glue("1.work/models/m1_{tvar}_{sat}_{prd}_lst_{yr}_cv.fst") %>%
                          read_fst(columns = c(ref_cols, pred_cols), as.data.table = TRUE)

                    if (yr < 2017) {
                      setnames(cv, c("lst_1km_id", "climate_type"), c("grd_1km_id", "climate"))
                    }
                    setnames(cv, as.character(tvar), "obs")

                  } else {

                    models <- c("rf", "xgb", "ens")
                    tcols <- c(glue("res_{tvar}"), glue("pred_res_{tvar}_{models}"))
                    pcols <- sub("ens", "m4", models) %>%
                             glue("pred_{m}_cv", m = .)
                    cv <- glue("1.work/models/m4_{tvar}_{yr}_cv.fst") %>%
                          read_fst(
                            as.data.table = TRUE,
                            columns = c("date", "grd_200m_id", tcols)
                          ) %>%
                          setnames(old = tcols, new = c("res", pcols))

                  }

                  cv

                }) %>%
                rbindlist

          cv[, tvar := tvar]
          setcolorder(cv, c("tvar", "date"))

        }) %>%
        rbindlist

  if (model == "m1") {
    setkeyv(cv, c("tvar", "date", "grd_1km_id"))
  } else {
    setkeyv(cv, c("tvar", "date", "grd_200m_id"))
  }

  # Save
  write_fst(cv, path, compress = 100)
  report(glue("-> {path}"))

  # Return
  cv

}

## Calculate m1 performance (1 km temperature) -----

tvars <- c("tmin", "tmean", "tmax") %>%
         factor(., levels = .)

m1_cv <- range(constants$model_years) %>%
         paste(collapse = "-") %>%
         glue("1.work/summaries/m1_{yrs}_cv.fst", yrs = .) %>%
         load_cv(model = "m1", yrs = constants$model_years, tvars = tvars, force = OVERWRITE)

report("Assessing m1 performance")

report("  All years")
overall <- split(m1_cv, by = "tvar") %>%
           eval_perf(lab_cols = "tvar", obs_col = "obs", pred_col = "pred_m1_cv",
                     loc_id_col = "grd_1km_id", cores = ncores)

report("  By year")
annual <- m1_cv[, .(year = year(date), tvar)] %>%
          split(m1_cv, .) %>%
          eval_perf(lab_cols = c("year", "tvar"), obs_col = "obs", pred_col = "pred_m1_cv",
                    loc_id_col = "grd_1km_id", cores = ncores)

report("  By month")
monthly <- m1_cv[, .(month = month(date), tvar)] %>%
           split(m1_cv, .) %>%
           eval_perf(lab_cols = c("month", "tvar"), obs_col = "obs", pred_col = "pred_m1_cv",
                     loc_id_col = "grd_1km_id", cores = ncores)

report("  By season")
seasonal <- m1_cv[, .(season = to_season(date), tvar)] %>%
            split(m1_cv, .) %>%
            eval_perf(lab_cols = c("season", "tvar"), obs_col = "obs", pred_col = "pred_m1_cv",
                      loc_id_col = "grd_1km_id", cores = ncores)

report("  By climate region")
climatic <- m1_cv[climate != 9, ] %>%
            .[order(tvar, climate), ] %>%
            split(by = c("climate", "tvar"), drop = TRUE) %>%
            eval_perf(lab_cols = c("climate", "tvar"), obs_col = "obs", pred_col = "pred_m1_cv",
                      loc_id_col = "grd_1km_id", cores = ncores)

report("  By urban / rural")
stns <- st_read("1.work/grids/stations_gridded.shp", quiet = TRUE) %>%
        as.data.table %>%
        .[, .(insee_id, urb_dist)]
urb <- stns[, .(urbanity = to_urbanity(mean(urb_dist))), by = "insee_id"] %>%
       .[m1_cv, on = c(insee_id = "stn_id")]
urbanity <- urb[, .(urbanity, tvar)] %>%
            split(urb, .) %>%
            eval_perf(lab_cols = c("urbanity", "tvar"), obs_col = "obs", pred_col = "pred_m1_cv",
                      loc_id_col = "grd_1km_id", cores = ncores)
rm(urb, stns)

# Save
report("Saving m1 performance summary")
m1_smry <- list(overall, annual, monthly, seasonal, climatic, urbanity) %>%
           rbindlist(fill = TRUE) %>%
           .[, model := "m1"] %>%
           setcolorder(c("model", "tvar", "year", "month", "season", "climate", "urbanity"))
path <- range(constants$model_years) %>%
        paste(collapse = "-") %>%
        glue("3.outcomes/model_performance/m1_{yrs}_cv_perf.csv", yrs = .)
fwrite(m1_smry, path)
report(glue("-> {path}"))

# Cleanup m1
rm(m1_cv, path, overall, annual, monthly, seasonal, climatic, urbanity)


## Calculate m4 performance (200 m residual, ensemble) -----

m4_cv <- range(constants$model_years) %>%
         paste(collapse = "-") %>%
         glue("1.work/summaries/m4_{yrs}_cv.fst", yrs = .) %>%
         load_cv(model = "m4", yrs = constants$model_years, tvars = tvars, force = OVERWRITE)

report("Assessing m4 performance")

report("  All years")
overall <- split(m4_cv, by = "tvar") %>%
           eval_perf(lab_cols = "tvar", obs_col = "res", pred_col = "pred_m4_cv",
                     loc_id_col = "grd_200m_id", cores = ncores)

report("  By year")
annual <- m4_cv[, .(year = year(date), tvar)] %>%
          split(m4_cv, .) %>%
          eval_perf(lab_cols = c("year", "tvar"), obs_col = "res", pred_col = "pred_m4_cv",
                    loc_id_col = "grd_200m_id", cores = ncores)

report("  By month")
monthly <- m4_cv[, .(month = month(date), tvar)] %>%
           split(m4_cv, .) %>%
           eval_perf(lab_cols = c("month", "tvar"), obs_col = "res", pred_col = "pred_m4_cv",
                     loc_id_col = "grd_200m_id", cores = ncores)

report("  By season")
seasonal <- m4_cv[, .(season = to_season(date), tvar)] %>%
            split(m4_cv, .) %>%
            eval_perf(lab_cols = c("season", "tvar"), obs_col = "res", pred_col = "pred_m4_cv",
                      loc_id_col = "grd_200m_id", cores = ncores)

report("  By climate region")
clim <- read_fst("1.work/extracts/grd_200m_climate.fst", as.data.table = TRUE) %>%
        .[m4_cv, on = "grd_200m_id"] %>%
        .[climate != 9, ]
climatic <- clim[order(tvar, climate), ] %>%
            split(by = c("climate", "tvar"), drop = TRUE) %>%
            eval_perf(lab_cols = c("climate", "tvar"), obs_col = "res", pred_col = "pred_m4_cv",
                      loc_id_col = "grd_200m_id", cores = ncores)
rm(clim)

report("  By urban / rural")
{
  # Calculate the distance between each 200 m grid cell with a residual and the closest urban area
  report("    Categorizing m4 CV points as urban / rural")
  grd_200m <- read_fst("1.work/grids/grd_200m.fst", as.data.table = TRUE) %>%
              setnames("id", "grd_200m_id")
  res_pts <- m4_cv[, lapply(.SD, unique), .SDcols = "grd_200m_id"] %>%
             grd_200m[., on = "grd_200m_id"] %>%
             st_as_sf(coords = c("x", "y"), crs = 3035)
  urb_areas <- read_sf("1.work/grids/grd_200m_urb_areas.shp")

  res_urb_nn <- st_nn(res_pts, urb_areas, k = 1, returnDist = TRUE)
  res_pts$urb_dist <- round(unlist(res_urb_nn$dist))

  urb <- as.data.table(res_pts) %>%
         .[, .(grd_200m_id, urbanity = to_urbanity(urb_dist))] %>%
         m4_cv[., on = "grd_200m_id"]
}
report("    Evaluating performance")
urbanity <- urb[, .(urbanity, tvar)] %>%
            split(urb, .) %>%
            eval_perf(lab_cols = c("urbanity", "tvar"), obs_col = "res", pred_col = "pred_m4_cv",
                      loc_id_col = "grd_200m_id", cores = ncores)
rm(grd_200m, res_pts, urb_areas, res_urb_nn, urb)

# Save
report("Saving m4 performance summary")
m4_smry <- list(overall, annual, monthly, seasonal, climatic, urbanity) %>%
           rbindlist(fill = TRUE) %>%
           .[, model := "m4"] %>%
           setcolorder(c("model", "tvar", "year", "month", "season", "climate", "urbanity"))
path <- range(constants$model_years) %>%
        paste(collapse = "-") %>%
        glue("3.outcomes/model_performance/m4_{yrs}_cv_perf.csv", yrs = .)
fwrite(m4_smry, path)
report(glue("-> {path}"))

# Cleanup m4
rm(m4_cv, path, overall, annual, monthly, seasonal, climatic, urbanity)


# Cleanup
rm(m1_smry, m4_smry, tvars, eval_perf, to_season, to_urbanity, load_cv)

report(glue("Done {paste(range(constants$model_years), collapse = '-')}"))
