#
# Fit stage 1 models (m1). These will predict 1 km Ta where LST is available
#

# Load packages, helper functions, and constants
source("init.R")
source("2.scripts/helpers/make_cv_test_sets.R")

report("---")
report("Fitting stage 1 models")


# Functions -------------------------------------------------------------------

# Cross-validate a model variant
cross_validate <- function(variant, folds = 10, method = "kfold", group_by = NULL, return_cv = TRUE,
                           ncores)
{
  # Load the data for the variant
  dataset <- load_mod1(variant)

  # Make test sets for k-folds grouped cross-validation
  cv_splits <- make_cv_test_sets(dataset, folds, method = method, group_by = group_by)

  # For each test set, train a model on the other data and then predict for the test set
  run_cv_fold <- function(cv_split, predictors, formula_string, dataset) {
    # Split the data into train and test sets
    test <- dataset[cv_split$test, ]
    train <- dataset[-cv_split$test, ]
    rm(dataset)

    # Scale the train set, then apply the same scaling to the test set
    # If scaled before splitting into train and test info from train would leak into test set
    # Don't return the scaling factors; in each fold the test set is scaled and predicted, and the
    # result is merged into a single table so there is no need to save each fold's scaling factors
    train[, (predictors) := lapply(.SD, scale), .SDcols = predictors]
    scales <- get_scales(train)
    for (col in predictors) {
      s <- scales[[col]]
      test[, (col) := scale(get(col), center = s[["scaled:center"]], scale = s[["scaled:scale"]])]
    }

    # Fit the formula to the training set
    # Disable some convergence tests to speed up and avoid spurious warnings; a spot check using
    # allFit showed that models with "failed to converge" warnings had the same effect estimates
    # as models fit using other optimizers which did converge.
    # See ?convergence and https://cran.r-project.org/web/packages/lme4/vignettes/lmerperf.html
    model <- lmer(formula_string, train, control = lmerControl(calc.derivs = FALSE))

    # Raise if the model fails to converge
    if (!is.null(model@optinfo$conv$lme4$messages)) {
      stop(model@optinfo$conv$lme4$messages)
    }

    # Predict for the test set and return it
    preds <- predict(model, newdata = test, allow.new.levels = TRUE, re.form = NULL)
    test[, pred_m1_cv := preds]
    test[, cv_fold := cv_split$fold]
  }
  cv <- mclapply(cv_splits, run_cv_fold, dataset = dataset, predictors = variant$predictors,
                 formula_string = variant$formula_string, mc.cores = ncores, mc.preschedule = FALSE)

  # Raise if errors were encountered
  for (cv_split in cv) {
    if (class(cv_split)[1] == "try-error") {
      stop(cv_split)
    }
    rm(cv_split)
  }

  # Combine into a single data frame and assess the predictions
  cv <- rbindlist(cv)
  variant$cv_stats <- assess_preds(dtable = cv, obs_col = variant$t_col, pred_col = "pred_m1_cv",
                                   loc_id_col = "stn_id")

  # Optionally add the CV predictions to the variant
  if (isTRUE(return_cv)) {
    cv %>% setkeyv(cols = c("cv_fold", "date", "stn_id"))
    variant$cv <- cv
  }

  # Return the variant
  variant
}

# Fit a model variant and assess its predictions
fit_variant <- function(variant, REML = TRUE, return_model = TRUE, backfit = FALSE) {
  # Load and format the mod1 data needed by the variant
  mod1 <- load_mod1(variant)

  # Scale the data
  # Note: scaling does not seem to affect model performance, but it speeds up fitting ~3x
  # (50 sec to fit on scaled data vs. 170 sec to fit on unscaled data)
  mod1[, (variant$predictors) := lapply(.SD, scale), .SDcols = variant$predictors]

  # Construct a formula object for the variant's formula string
  # lmer requires a formula object and bfFixefLMER_t.fnc requires the formula's environment to
  # contain mod1
  fml <- as.formula(variant$formula_string)

  # If backfitting, fit initial model without REML
  # This is needed to allow valid comparisons of AIC between nested model variants
  if (backfit == TRUE) { REML <- FALSE }

  # Disable some convergence tests to speed up and avoid spurious warnings; a spot check using
  # allFit showed that models with "failed to converge" warnings had the same effect estimates
  # as models fit using other optimizers which did converge.
  # See ?convergence and https://cran.r-project.org/web/packages/lme4/vignettes/lmerperf.html
  lmer_ctrl <- lmerControl(calc.derivs = FALSE)

  # Fit the model; use REML = FALSE to compare AIC of nested variants e.g. when backfitting
  m1 <- lmer(formula = fml, data = mod1, REML = REML, control = lmer_ctrl)

  # Optionally iterate over the predictors checking whether removing each predictor improves the
  # model fit as evaluated by the relative likelihood of the simpler model's AIC
  if (isTRUE(backfit)) {
    logfile <- paste("m1", variant$t_col, variant$lst_col, variant$year, "log.txt", sep = "_") %>%
               file.path(constants$models_dir, .)

    # Backfit to find the best set of predictors
    # * Don't need to set.REML.FALSE if initial model was fit with REML = FALSE
    # * Don't reset.REML.TRUE; update the variant and reload mod1 first
    # * Use bfFixefLMER_t.fnc b/c t values seem to be better indicators of which predictors are
    #   unhelpful than F values. This will produce a warning that t values are inappropriate b/c
    #   "factors have > 2 levels"; I think it's safe to ignore this b/c we only use t values to
    #   choose the order in which we evaluate predictors (all are evaluated and included or not
    #   based on AIC)
    m1 <- bfFixefLMER_t.fnc(model = m1, method = "AIC", log.file = logfile, set.REML.FALSE = REML,
                            reset.REML.TRUE = FALSE)

    # Update the variant to reflect the best set of predictors
    new_predictors <- fixef(m1) %>% names %>% setdiff("(Intercept)")
    variant$predictors <- variant$predictors[variant$predictors %in% new_predictors]
    new_rand_effect <- paste0("(", strsplit(as.character(formula(m1))[3], "\\(")[[1]][2])
    variant$rand_effect <- new_rand_effect
    variant$formula_string <- formula_string(variant)

    # Re-load the data, scale, and re-fit the model with REML = TRUE. This restores any rows that
    # were dropped due to missing data for predictors that the final model does not use
    mod1 <- load_mod1(variant)
    mod1[, (variant$predictors) := lapply(.SD, scale), .SDcols = variant$predictors]
    fml <- as.formula(variant$formula_string)
    m1 <- lmer(formula = fml, data = mod1, REML = TRUE, control = lmer_ctrl)
  }

  # Store any convergence warnings
  variant$warnings <- m1@optinfo$conv$lme4$messages

  # Store the scaling factors used on mod1
  variant$scales <- get_scales(mod1)

  # Store the Akaike information criterion and the t values of the estimated coefficients
  variant$aic <- AIC(m1) %>% round
  variant$bic <- BIC(m1) %>% round
  variant$tvals <- coef(summary(m1))[, "t value"] %>% round(2)
  variant$fixef <- fixef(m1) %>% round(3)

  # # Assess the model's predictions
  # mod1[, "pred_m1" := predict(m1)]
  # variant$stats <- assess_preds(dtable = mod1, obs_col = variant$t_col, pred_col = "pred_m1",
  #                               loc_id_col = "stn_id")

  # Optionally add the model to the variant
  if (isTRUE(return_model)) {
    variant$model <- m1
  }

  # Return the variant
  variant
}

# Construct a formula for a model variant and return it as a character vector
# Using a character vector is convenient for printing and is memory efficient (a formula object
# includes a copy of the environment it was created in, including all objects in that environment)
formula_string <- function(variant) {
  c(variant$predictors, variant$rand_effect) %>%
    paste(collapse = " + ") %>%
    paste(variant$t_col, "~", .)
}

# Return the scaling factors that were applied to the data columns
get_scales <- function(mod1) {
  # Get the scaling factors for every column (even columns that have not been scaled)
  scales <- lapply(mod1, function(col) {
    if (!is.null(attr(col, "scaled:center"))) {
      scales <- attributes(col)[c("scaled:center", "scaled:scale")]
    }
  })

  # Return the scaling factors for all columns that have been scaled
  missing <- sapply(scales, is.null)
  scales[!missing]
}

# Initialize a model variant (a list of all the info needed to fit a model)
init_variant <- function(variant) {
  # Confirm required info is present
  for (item in c("year", "t_col", "lst_col", "predictors", "rand_effect")) {
    if (is.null(variant[[item]])) {
      stop(paste("Variant is missing", item))
    }
  }

  # Replace placeholders (e.g. lst_col) with actual column names
  variant <- replace_placeholders(variant)

  # Set the formula as a string. Do not create a formula object here; for backfitting with
  # LMERConvenienceFunctions::bfFixefLMER_t.fnc the formula's environment must contain the model
  # data so we must construct the formula object in an environment where the data is loaded.
  variant$formula_string <- formula_string(variant)

  # Set paths
  base <- paste("m1", variant$t_col, variant$lst_col, variant$year, sep = "_")
  variant$model_path <- paste0(base, ".rds") %>%
                        file.path(constants$models_dir, .)
  variant$cv_path <- paste0(base, "_cv.fst") %>%
                     file.path(constants$models_dir, .)
  variant$meta_path <- paste0(base, "_metadata.rds") %>%
                       file.path(constants$models_dir, .)
  variant$data_path <- sub("_lst$", "", variant$lst_col) %>%
                       paste0("mod1_", ., "_", variant$year, ".fst") %>%
                       file.path(constants$joined_dir, .)

  # Return the variant
  variant
}

# Load and format mod1 data for a model variant
load_mod1 <- function(variant, verbose = FALSE) {
  # Load the data
  if (length(variant$data_path) == 0) {
    msg <- paste(variant$year, collapse = " ") %>% paste("No data for variant", .)
    stop(msg)
  }
  mod1 <- lapply(variant$data_path, function(path) {
    if (isTRUE(verbose)) { paste("Loading", basename(path)) %>% report }
    threads_fst(ncores)
    read_fst(path, as.data.table = TRUE)
  }) %>% rbindlist

  # Abbreviate column names
  colnames(mod1) %>% shorten %>% setnames(mod1, .)
  setkeyv(mod1, c("date", "stn_id", "grd_1km_id"))

  # Drop LST where error > 2 K (~ 1.5% of data)
  lst_err_col <- paste0(variant$lst_col, "_err")
  mod1 <- mod1[get(lst_err_col) %in% c("<= 1K", "<= 2K"), ]

  # Subset the required data and drop rows with missing data
  info_cols <- c("date", "stn_id", "lat", "lon", "stn_type", "stn_elev", "grd_1km_id")
  model_cols <- c("climate", variant$t_col, variant$predictors)
  mod1 <- mod1[, c(info_cols, model_cols), with = FALSE] %>% na.omit

  # Convert LST from Kelvin to Celsius
  if (variant$lst_col %in% names(mod1)) {
    mod1[, (variant$lst_col) := get(variant$lst_col) - 273.15] # K -> deg C
  }

  # NOTE: do NOT scale data here as that would leak information into the CV test set

  # Return the data
  mod1
}

# Replace any placeholders in a variant's predictors and random effect with actual column names
replace_placeholders <- function(variant) {
  # Determine the columns that placeholders refer to
  sat <- strsplit(variant$lst_col, "_")[[1]][1]
  dict <- c(
    lst_col = variant$lst_col,
    emis_col = paste0(sat, "_emis"),
    ndvi_col = paste0(sat, "_ndvi"),
    sim_col = paste0("sim_", variant$t_col)
  )

  # Replace placeholders in the predictors
  names(variant$predictors) <- variant$predictors
  placeholders <- dict[names(dict) %in% variant$predictors]
  variant$predictors[names(placeholders)] <- placeholders

  # Replace any placeholders in the random effect
  for (pattern in names(dict)) {
    variant$rand_effect <- gsub(pattern, dict[pattern], variant$rand_effect)
  }

  # Return the modified variant
  variant
}

# Return a single-column data frame summarizing a model variant
summarize <- function(variant, report_all_predictors = FALSE) {
  # Identify the variant
  smry <- variant[c("year", "t_col", "lst_col")]
  if (length(variant$year) > 1) {
    min_year <- min(smry$year)
    max_year <- max(smry$year)
    if (identical(smry$year, min_year:max_year)) {
      smry$year <- paste(min_year, max_year, sep = ":")
    } else {
      smry$year <- paste(smry$year, collapse = ",")
    }
  }

  # Get the random effect nesting of the variant
  rand_eff_nesting <- strsplit(variant$rand_effect, "\\| ")[[1]]
  smry$rand_eff <- sub("\\(", "", rand_eff_nesting[1])
  smry$nesting <- sub("\\)", "", rand_eff_nesting[2])

  # Get AIC and BIC
  smry$aic <- variant$aic
  smry$bic <- variant$bic

  # Get model statistics if they exist
  stats <- c("r2", "r2.space", "r2.time", "rmse", "rmse.space", "rmse.time", "mae", "mae.space",
             "mae.time", "i", "i.se", "slope", "slope.se", "res.mean", "res.sd", "res.min",
             "res.1Q", "res.med", "res.3Q", "res.max")
  # if (!is.null(variant$stats)) {
  #   smry[stats] <- variant$stats[stats] %>% sprintf("%.3f", .)
  # }

  # Get cross-validation statistics if they exist
  if (!is.null(variant$cv_stats)) {
    smry[paste0("cv.", stats)] <- variant$cv_stats[stats] %>% sprintf("%.3f", .)
  }

  # Get model coefficients and their t values if they exist
  if (!is.null(variant$tvals)) {
    if (isTRUE(report_all_predictors)) {
      # Report t values for all possible predictors (including any not used in the model). This
      # allows printing a data frame summarizing all model variants.
      report_for <- c("(Intercept)", all_predictors)
      names(report_for) <- report_for
      report_for[names(variant$predictors)] <- variant$predictors # substitute placeholders
    } else {
      # Report t values for predictors used in the model
      report_for <- names(variant$tvals)
      names(report_for) <- report_for
    }
    # It's important to suffix the coefficient names (e.g. lst_col.v); otherwise the lst_col
    # coefficient will overwrite the lst_col name (e.g. aqua_day_lst) in the summary report
    smry[paste0(names(report_for), ".t")] <- variant$tvals[report_for] %>% sprintf("%.2f", .)
    smry[paste0(names(report_for), ".v")] <- variant$fixef[report_for] %>% sprintf("%.3f", .)
  }

  # Get any warnings
  if (!is.null(variant$warnings)) {
    smry$warnings <- variant$warnings
  }

  # Return as a single-row data frame
  smry %>% as.data.frame %>% set_colnames(names(smry))
}

# Return a data frame summarizing all model variants (one per column)
summarize_variants <- function(variants) {
  lapply(variants, summarize, report_all_predictors = TRUE) %>%
    do.call(rbind, .) %>%
    set_rownames(1:length(variants))
}

# Construct model variants for all combinations of the specified parameters
variants_for_combo <- function(combo) {
  expand.grid(combo) %>%
    apply(1, as.list) %>%
    lapply(init_variant) %>%
    purrr::discard(~ grepl("^aqua", .x$lst_col) && .x$year %in% c(2000, 2001))
}

# Setup -------------------------------------------------------------------------

# All possible predictors. List clc_vegetation last so that it will be dropped if model matrix is
# rank deficient (happens b/c clc_vegetation is highly correlated with clc_artificial). Predictors
# ending in '_col' are placeholders that will be replaced by specific column names. For example,
# when modeling tmin ~ aqua_night_lst:
# 'lst_col'  -> 'aqua_night_lst'
# 'emis_col' -> 'aqua_emis'
# 'ndvi_col' -> 'aqua_ndvi'
# 'sim_col'  -> 'sim_tmin'
all_predictors <- c("lst_col", "emis_col", "ndvi_col", "elev", "pop", "clc_artificial", "clc_water",
                    "clc_bare", "clc_vegetation") # , "sim_col")

# # 1. Compare sets of predictors -----------------------------------------------
#
# # Sets of predictors to test. Note that these exclude clc_vegetation b/c it is highly correlated
# # (< -0.9) with clc_artificial; if all clc predictors are included the model is rank-deficient and
# # the last predictor is automatically excluded.
# predictor_sets <- list(
#   c("lst_col", "emis_col"),
#   c("lst_col", "emis_col", "ndvi_col"),
#   c("lst_col", "emis_col", "ndvi_col", "elev"),
#   c("lst_col", "emis_col", "ndvi_col", "elev", "pop"),
#   c("lst_col", "emis_col", "ndvi_col", "elev", "pop", "sim_col"),
#   c("lst_col", "emis_col", "ndvi_col", "elev", "pop", "sim_col", "clc_artificial"),
#   c("lst_col", "emis_col", "ndvi_col", "elev", "pop", "sim_col", "clc_artificial", "clc_bare"),
#   c("lst_col", "emis_col", "ndvi_col", "elev", "pop", "sim_col", "clc_artificial", "clc_bare", "clc_water"),
#   setdiff(all_predictors, "emis_col"),
#   setdiff(all_predictors, "sim_col")
# )
#
# # Combinations test e.g tmin ~ aqua_night_lst + base_predictors
# combos <- list(
#   list(
#     year = 2008,
#     t_col = "tmin",
#     rand_effect = c("(1 + lst_col | date)", "(1 + lst_col | date/climate)"),
#     predictors = predictor_sets,
#     lst_col = c("aqua_night_lst", "terra_night_lst")
#   ),
#   list(
#     year = 2008,
#     t_col = "tmax",
#     rand_effect = c("(1 + lst_col | date)", "(1 + lst_col | date/climate)"),
#     predictors = predictor_sets,
#     lst_col = c("aqua_day_lst", "terra_day_lst")
#   ),
#   list(
#     year = 2008,
#     t_col = "tmean",
#     rand_effect = c("(1 + lst_col | date)", "(1 + lst_col | date/climate)"),
#     predictors = predictor_sets,
#     lst_col = c("aqua_day_lst", "terra_day_lst", "aqua_night_lst", "terra_night_lst")
#   ),
#   # without lst or emis
#   list(
#     year = 2008,
#     predictors = list(setdiff(all_predictors, c("lst_col", "emis_col"))),
#     rand_effect = c("(1 | date)", "(1 | date/climate)"),
#     lst_col = c("aqua_day_lst", "terra_day_lst", "aqua_night_lst", "terra_night_lst"),
#     t_col = c("tmin", "tmax", "tmean")
#   )
# )
#
# for (combo in combos) {
#   paste(combo$t_col, combo$year, collapse = " ") %>% report
#
#   # Construct the model variants to fit
#   variants <- variants_for_combo(combo)
#
#   # Fit the variants in parallel; don't use REML or return the fit models
#   paste("  Fitting", length(variants), "model variants in parallel") %>% report
#   variants <- mclapply(variants, fit_variant, REML = FALSE, return_model = FALSE, mc.cores = ncores)
#
#   # Summarize the variants
#   report("  Saving summary")
#   {
#     smry <- summarize_variants(variants)
#     year <- sapply(variants, function(v) { v$year })
#     year_range <- paste0(min(year), "-", max(year))
#
#     if (!is.null(suffix)) { suffix <- paste0(suffix, "_") }
#     path <- Sys.time() %>%
#             format("%Y-%m-%d_%H%M%S") %>%
#             paste0("m1_", smry$t_col[1], "_", year_range, "_", suffix, ., ".csv") %>%
#             file.path(constants$models_dir, .)
#     paste0("    ", basename(path)) %>% report
#     write.csv(smry, file = path)
#   }
# }
#
# # Conclusions
# # 1a. Does nesting by climate type improve fit?
# # * Tested on 2003, 2008, 2012
# # * Nesting improves fit (slightly):
# #   - RMSE decreases ~ 0.1
# #   - R2   increases ~ 0.02
# # 1b. Are initial stats similar to cross-validated stats?
# # * Tested on 2003, 2008, 2012
# # * Initial stats are similar to cross-validated stats:
# #   - RMSE / R2 decrease slightly with cross-validation
# #   - Relative performance of model variants does not change
# #   - kfold grouped is slightly more pessimistic than kfold
# # 1c. What are the best models? (initial)
# # * Tested on 2003, 2008, 2005
# # * tmin
# #   - aqua_night: all > leave out emis
# #   - terra_night: leave out emis > all
# # * tmax
# #   - aqua_day: all > leave out pop (sometimes leave out clc_water or clc_artificial is best)
# #   - terra_day: all > leave out pop (sometimes leave out clc_water or clc_artificial is best)
# # * tmean
# #   - aqua_day: all > leave out emis
# #   - terra_day: all > leave out emis
# #   - aqua_night: leave out emis > all (sometimes reverse)
# #   - terra_night: all > leave out pop
# # 1d. What are the best models? (initial2)
# # * Tested on each year in 2003:2016
# # * tmin
# #   - aqua_night: all seems best; a few years benefit from leaving out emis; a few might benefit from leaving out ndvi or elev or emis and ndvi
# #   - terra_night: all seems best; a few years benefit from leaving out emis; a few might benefit from leaving out elev
# #   - overall: terra better for space (RMSE space diff ~ 0.1); aqua better for time (RMSE time diff ~ 0.07)
# # * tmax
# #   - aqua_day: all seems best; a few years benefit from leaving out clc_artificial or clc_water
# #   - terra_day: fails to converge; all seems best; a few years benefit from leaving out pop or clc_artificial; a few might benefit from leaving out emis
# #   - overall: aqua better (RMSE diff ~ 0.05); aqua more consistent
# # * tmean
# #   - aqua_day: all seems best; a few years benefit from leaving out emis; 2014 might benefit from leaving out ndvi
# #   - terra_day: all seems best; 2007 benefits from leaving out emis; a few might benefit from leaving out ndvi
# #   - aqua_night: all seems best; a few years benefit from leaving out emis; a few might benefit from leaving out pop
# #   - terra_night: all seems best; 2004 benefits from leaving out emis; a few might benefit from leaving out pop
# #   - overall: terra night better (RMSE diff ~ 0.04 vs terra day; RMSE diff ~ 0.02 vs aqua night; RMSE diff ~ 0.04 vs aqua day); terra more consistent
# # 1e. What are the best models? (initial3)
# # * Fit on entire period 2003:2016
# # * tmin
# #   - terra model performs if leave out emis
# #   - terra better for space; aqua for time (RMSE space diff ~ 0.1; RMSE time diff ~ 0.05)
# #   - aqua elev effect positive (i.e. higher elev -> higher tmin) - maybe main elev effect captured by sim?
# # * tmax
# #   - aqua better (RMSE diff ~ 0.03)
# #   - ndvi effect positive (i.e. higher ndvi -> higher tmax) - maybe related to season?
# #   - clc_bare effect negative (i.e. more bare -> lower tmax) - maybe Relativeated to elevation?
# # * tmean
# #   - terra better (RMSE diff ~ 0.02)
# #   - aqua emis effect negative (i.e. higher emis -> lower tmean) -
# #   - clc_bare effect negtive (i.e. more bare -> lower tmean) - maybe related to elevation?
# # * overall
# #   - sim improves RMSE by 0.23 - 0.42
# #   - lst improves RMSE by 0.05 - 0.13
# #   - leaving out sim improves RMSE space ~ 0.05 for tmax and tmean
# #   - leaving out sim hurts RMSE space ~ 0.05 for tmin
# #   - lst + sim alone are almost as good as all predictors (RMSE diff ~ 0.02)
#
# 2. Find best model for each year by backfitting -----------------------------

# Combinations to backfit
combos <- list(
  list(
    t_col = "tmin",
    year = constants$model_years,
    lst_col = c("aqua_night_lst", "terra_night_lst"),
    predictors = list(setdiff(all_predictors, "sim_col")),
    rand_effect = "(1 + lst_col | date/climate)"
  ),
  list(
    t_col = "tmax",
    year = constants$model_years,
    lst_col = c("aqua_day_lst", "terra_day_lst"),
    predictors = list(setdiff(all_predictors, "sim_col")),
    rand_effect = "(1 + lst_col | date/climate)"
  ),
  list(
    t_col = "tmean",
    year = constants$model_years,
    lst_col = c("aqua_day_lst", "terra_day_lst", "aqua_night_lst", "terra_night_lst"),
    predictors = list(setdiff(all_predictors, "sim_col")),
    rand_effect = "(1 + lst_col | date/climate)"
  )
)

for (combo in combos) {

  # Construct the model variants
  variants <- variants_for_combo(combo)

  # Prune variants that have already been saved
  if (!OVERWRITE) {
    check <- map_lgl(variants, ~ c(.x$model_path, .x$cv_path, .x$meta_path) %>% file.exists %>% all)
    variants <- variants[!check]
  }
  rm(check)

  # Skip if all variants are already saved
  if (length(variants) == 0) {
    paste(combo$t_col, ": all variants exist")
    next
  }

  # Ensure mod1 exists for all variants
  purrr::map(variants, ~ stopifnot(file.exists(.x$data_path)))

  # Backfit in parallel, adding the model to each variant
  # ~ 45 minutes for 32 variants (all in parallel)
  map_int(variants, "year") %>% unique %>% paste(collapse = " ") %>% paste(combo$t_col, ":", .) %>%
    report
  paste(combo$t_col, ": backfitting", length(variants), "variants") %>% report
  variants <- mclapply(variants, fit_variant, backfit = TRUE, return_model = TRUE,
                       mc.cores = ncores, mc.preschedule = FALSE)

  # Save the models to disk
  # ~ 7 minutes for 32 variants (~ 13 seconds / variant)
  paste(combo$t_col, ": saving", length(variants), "models") %>% report
  variants <- lapply(variants, function(variant) {
    paste0("  ", variant$model_path) %>% report
    saveRDS(variant$model, variant$model_path)

    # Remove the model from the variant and return
    variant$model <- NULL
    variant
  })

  # Cross-validate and save the cross-validation and variant metadata
  # ~ 60 minutes for 32 variants
  paste(combo$t_col, ": cross-validating", length(variants), "models") %>% report
  variants <- lapply(variants, function(variant) {

    # Cross-validate
    variant <- cross_validate(variant, folds = 10, method = "kfold", return_cv = TRUE,
                              ncores = ncores)

    # # K-fold grouped by station id
    # variant <- cross_validate(variant, folds = 10, method = "kfold grouped", group_by = "stn_id",
    #                           return_cv = TRUE, ncores = ncores)

    # Save the cross-validation
    paste0("  ", variant$cv_path) %>% report
    threads_fst(ncores)
    write_fst(variant$cv, variant$cv_path, compress = 100)

    # Remove the cross-validation from the variant and save as metadata
    variant$cv <- NULL
    paste0("  ", variant$meta_path) %>% report
    saveRDS(variant, variant$meta_path)

    # Return the variant
    variant
  })

  # Load all saved variants and save a summary
  paste(combo$t_col, ": summarizing all saved variants") %>% report
  variants <- variants_for_combo(combo) %>%
              map(~ readRDS(.x$meta_path))
  {
    smry <- summarize_variants(variants)

    year_range <- purrr::map_int(variants, "year") %>%
                  range() %>%
                  paste(collapse = "-")
    path <- Sys.time() %>%
            format("%Y-%m-%d_%H%M%S") %>%
            paste0("m1_", smry$t_col[1], "_", year_range, "_smry_", ., ".csv") %>%
            file.path(constants$work_dir, "summaries", .)
    paste("->", basename(path)) %>% report
    write.csv(smry, file = path)
    rm(smry, path, year_range)
  }

}

# Cleanup
rm(variants, combo, combos, all_predictors, cross_validate, fit_variant, formula_string, get_scales,
   init_variant, load_mod1, make_cv_test_sets, replace_placeholders, summarize, summarize_variants,
   variants_for_combo)

report(glue("Done {paste(range(constants$model_years), collapse = '-')}"))

# # For reference - load and check all fitted models, calculate statistics, and save metadata
# m1_paths <- list.files(constants$models_dir, pattern = "m1_.+_model\\.rds", full.names = TRUE)
# for(m1_path in m1_paths) {
#   basename(m1_path) %>% paste("Checking", .) %>% report
#
#   # Load the model and its metadata
#   m1 <- readRDS(m1_path)
#   m1.meta <- sub("_model\\.rds", "_metadata.rds", m1_path) %>% readRDS
#
#   # Load mod1
#   mod1 <- load_mod1(m1.meta)
#
#   # Confirm the model's data frame matches mod1
#   if (!identical(nrow(m1@frame), nrow(mod1))) {
#     stop("Model's data.frame does not match m1")
#   }
#
#   # Load cv
  # cv_path <- paste("m1", m1.meta$t_col, m1.meta$lst_col, m1.meta$year, "cv.fst", sep = "_") %>%
  #            file.path(constants$models_dir, .)
#   cv <- read_fst(cv_path, as.data.table = TRUE)
#
#   # Confirm the cv data matches mod1
#   if (!identical(cv[order(date, stn_id), .(date, stn_id)], mod1[, .(date, stn_id)])) {
#     stop("CV data does not match mod1")
#   }
#
#   # Assess the model's performance
#   cv[, "m1_pred" := predict(m1, newdata = cv)]
#   m1.meta$stats <- assess_preds(dtable = cv, obs_col = m1.meta$t_col, pred_col = "pred_m1",
#                                 loc_id_col = "stn_id")
#   m1.meta$cv_stats <- assess_preds(dtable = cv, obs_col = m1.meta$t_col, pred_col = "pred_m1_cv",
#                                    loc_id_col = "stn_id")
#
#   # Save the updated metadata
  # path <- paste("m1", m1.meta$t_col, m1.meta$lst_col, m1.meta$year, "metadata.rds", sep = "_") %>%
  #         file.path(constants$models_dir, .)
#   paste0("    ", basename(path)) %>% report
#   saveRDS(m1.meta, path)
# }
