#
# Fit stage 2 models (m2). These will be used to generate stage 3 predictions
#

# Load packages, helper functions, and constants
source("init.R")

report("---")
report("Fitting stage 2 models")


# Functions -------------------------------------------------------------------

# Construct the fixed and random parts of a m2 formula as needed for nlme
construct_m2_formula <- function(response, predictor) {
  list(
    fixed = paste(response, "~", predictor) %>% as.formula,           # response ~ predictor
    random = paste("~ 1 +", predictor, "| grd_1km_id") %>% as.formula # ~ 1 + predictor | grd_1km_id
  )
}

# Fit m2 on stage 2 predictions -----------------------------------------------

# Iterate over years
# ~ 60 min / year
for (year in constants$model_years) {

  # Skip if all m2 models already exist
  t_cols <- c("tmin", "tmean", "tmax")
  m2_paths <- paste0("m2_", t_cols, "_", year, ".rds") %>% file.path(constants$models_dir, .)
  names(m2_paths) <- t_cols
  present <- m2_paths[file.exists(m2_paths)]
  to_fit <- m2_paths[!file.exists(m2_paths)]
  if (length(to_fit) == 0 && !OVERWRITE) {
    next
  }

  # Load mod3 data including the m1 predictions
  # ~ 5 min, ~ 23 GB
  mod3_path <- paste("mod3", year, "pred_m1.fst", sep = "_") %>%
               file.path(constants$predictions_dir, .)
  paste(year, ": loading", basename(mod3_path)) %>% report
  threads_fst(ncores)
  mod3 <- read_fst(mod3_path, as.data.table = TRUE)

  # Fit m2 models in parallel (tmin / tmean / tmax)
  # ~ 30 min, ~ 170 GB
  paste(year, ": fitting stage 2 models in parallel") %>% report
  mclapply(names(to_fit), function(t_col) {

    # Select rows with m1 predictions
    m1_col <- paste0("m1_", t_col)
    idw_col <- paste0("idw_", t_col)
    paste(year, t_col, ": selecting rows with", m1_col) %>% report
    mod2 <- mod3[!is.na(get(m1_col)), c("date", "grd_1km_id", idw_col, m1_col), with = FALSE]

    # Fit m2 : Ta_pred ~ Ta_idw + (1 + Ta_idw | cell_id)
    # ~ 20 min, ~ 11.5 GB
    msg <- paste(": fitting m2:", m1_col, "~", idw_col, "+ (1 +", idw_col, "| grd_1km_id)")
    paste(year, t_col, msg) %>% report
    fml <- construct_m2_formula(response = m1_col, predictor = idw_col)
    m2 <- lme(
      fixed = fml$fixed,   # m1_txxx ~ idw_txxx
      random = fml$random, # ~ 1 + idw_txxx | grd_1km_id
      data = mod2,
      keep.data = FALSE,   # don't include a copy of mod2 in the model object
      control = lmeControl(opt = "optim")
    )

    # HACK set the model's formula (e.g. fixed = fml$fixed -> fixed = m1_tmax ~ idw_tmax)
    # This is necessary b/c nlme saves the exact call to `lme` (e.g. fixed = fml$fixed); setting the
    # formula allows us to use the model in an environment that does not contain the `fml` object
    m2$call$fixed <- fml$fixed
    m2$call$random <- fml$random

    # Reduce model size by ~ 6.5 GB by removing unneeded rownames
    rownames(m2$fitted) <- NULL
    rownames(m2$groups) <- NULL
    rownames(m2$residuals) <- NULL

    # Save the model
    # ~ 15 min, ~ 3 GB file
    paste(year, t_col, ": saving m2") %>% report
    m2_path <- to_fit[t_col]
    saveRDS(m2, m2_path)
    paste(year, t_col, ": ->", basename(m2_path)) %>% report

    # Cleanup
    rm(m2, mod2)
    gc()

    # Return NULL to save memory
    NULL

  }, mc.cores = ncores)

  rm(mod3)
  gc()

}

rm(m2_paths, present, t_cols, to_fit, year, construct_m2_formula)

report(glue("Done {paste(range(constants$model_years), collapse = '-')}"))


# # For reference: m2 including smoothing splines
#
# library(mgcv)    # GAM thin plate spline for m2
#
# source("2.scripts/helpers/summarize_as_raster.R")
#
# # Save an m2 variant
# save_m2 <- function(m2_variant, name) {
#   paste(year, t_col, ": saving", name) %>% report
#   path <- sub("^m2_", "", name) %>%
#           paste0(".rds") %>%
#           paste("m2", t_col, year, ., sep = "_") %>%
#           file.path(constants$models_dir, .)
#   saveRDS(m2_variant, path)
#   paste(year, t_col, "  ->", basename(path)) %>% report
# }
#
# # Save a mod2 data table that includes predictions from a m2 variant
# save_m2_preds <- function(mod2, name) {
#   paste(year, t_col, ": saving mod2 with", name, "predictions") %>% report
#   path <- paste0(name, ".fst") %>%
#           paste("mod2", t_col, year, ., sep = "_") %>%
#           file.path(constants$predictions_dir, .)
#   write_fst(mod2, path, compress = 100)
#   paste(year, t_col, "  ->", basename(path)) %>% report
# }
#
# # Save a summary raster of an m2 variant
# save_m2_summary <- function(smry, name) {
#   paste(year, t_col, ": saving summary raster of", name) %>% report
#   path <- paste0(name, "_smry.tif") %>%
#           paste("mod2", t_col, year, ., sep = "_") %>%
#           file.path(constants$predictions_dir, .)
#   save_geotiff(smry, path, overwrite = TRUE)
#   paste(year, t_col, ": ->", basename(path)) %>% report
# }
#
# # ~ 2 hours / model (half of time is to fit thin plate splines)
#
# # Ta_pred ~ Ta_idw + smooth(x, y) + (1 + Ta_idw | cell_id)
# # This cannot be fit due to computational limitations so we fit in three stages:
# # 1. Fit without the smoothing term: Ta_pred ~ Ta_idw + (1 + Ta_idw | cell_id)
# # 2. Smooth the residuals with a thin plate spline of x and y and adjust the predictions
# # 3. Model the smoothed predictions: Ta_smooth ~ Ta_idw + (1 + Ta_idw | cell_id)
#
# # Fit without the smoothing term: pred_m2 ~ idw_txxx + (1 + idw_txxx | grd_1km_id)
# # ~ 30 min, ~ 11.5 GB
# paste(year, t_col, ": fitting m2_rough (m2 without smoothing term)") %>% report
# fixed <- paste0("pred_m2 ~ idw_", t_col)
# random <- paste0("~ 1 + idw_", t_col, " | grd_1km_id")
# m2_rough <- lme(
#   fixed = as.formula(fixed),   # pred_m2 ~ idw_txxx
#   random = as.formula(random), # ~ 1 + idw_xxx | grd_1km_id
#   data = mod2,
#   keep.data = FALSE,            # don't include a copy of mod2 in the model object
#   control = lmeControl(opt = "optim")
# )
# rm(fixed, random)
#
# # Generate rough (unsmoothed) m2 predictions for mod2 using m2_rough. Use as.vector to remove
# # attributes from the predictions (which has a 'names' attribute that lits the grd_1km_id
# # corresponding to each predicted value)
# # ~ 1 min, 5.5 GB
# paste(year, t_col, ": predicting mod2 using m2_rough -> pred_m2_rough") %>% report
# mod2[, pred_m2_rough := predict(m2_rough) %>% as.vector]
#
# # Save m2_rough
# # ~ 15 min, ~ 3 GB file
# save_m2(m2_rough, "m2_rough")
#
# # Save the m2_rough predictions
# # ~ 3 min, ~ 3 GB file
# save_m2_preds(mod2, "pred_m2_rough")
#
# # Save a raster summary of the m2_rough predictions
# # ~ 30 seconds
# smry <- summarize_as_raster(mod2, "pred_m2_rough")
# save_m2_summary(smry, "pred_m2_rough")
# # plot(smry)
#
# # For each bimonthly period, fit a thin plate spline of x and y to the residuals of the rough
# # predictions, then adjust the rough predictions by *subtracting* the spline values. I don't
# # understand why we would subtract these rather than adding - subtracting would seem to make the
# # predictions *worse* than before - but this is what Alex and Adar did.
# # ~ 60 min, ~ 8 GB for splines, ~ 750 MB for smoothed preds
# paste(year, t_col, ": fitting thin plate splines") %>% report
# mod2[, bimon := (as.Date(date) %>% month %>% add(1) %>% divide_by(2) %>% as.integer)] # 3 min
# result <- split(mod2, by = "bimon") %>%
#           mclapply(function(splt) {
#             # Calculate residuals
#             splt[, res_m2_rough := pred_m2 - pred_m2_rough]
#
#             # Fit thin plate spline GAM (bam is ~4x slower than gam!? no speedup with nthreads > 1)
#             spline <- gam(res_m2_rough ~ s(grd_1km_x, grd_1km_y), data = splt)
#
#             # Adjust the predictions by *subtracting* the spline values
#             smoothed <- splt$pred_m2_rough - spline$fitted.values
#
#             # Return the GAM and the mod2 split
#             list(spline = spline, smoothed = smoothed)
#           }, mc.cores = ncores)
# splines <- lapply(result, function(x) { x$spline })
# smoothed <- lapply(result, function(x) { x$smoothed }) %>% unlist(use.names = FALSE)
# rm(result)
# gc()
#
# # Add the smoothed predictions to mod2
# # ~ 1 min, 6.5 GB
# paste(year, t_col, ": smoothing pred_m2_rough with thin plate splines -> pred_m2_smooth")
# mod2[, pred_m2_smooth := smoothed]
# rm(smoothed)
# gc()
#
# # Use multi-threaded fst mode (may run single-threaded after fork?)
# threads_fst(ncores)
#
# # Save the list of thin plate splines
# # ~ 20 min, ~ 5.5 GB file
# save_m2(splines, "splines")
#
# # Save the m2_smooth predictions
# # ~ 3 min, 3.7 GB
# save_m2_preds(mod2, "pred_m2_smooth")
#
# # Save a raster summary of the m2_smooth predictions
# # ~ 30 seconds
# smry <- summarize_as_raster(mod2, "pred_m2_smooth")
# save_m2_summary(smry, "pred_m2_smooth")
#
# # Model the smoothed predictions: pred_m2_smooth ~ idw_txxx + (1 + idw_txxx | grd_1km_id)
# # ~ 20 min, ~ 10.5 GB
# paste(year, t_col, ": fitting m2_final (predicts pred_m2_smooth)") %>% report
# fixed <- paste0("pred_m2_smooth ~ idw_", t_col)
# random <- paste0("~ 1 + idw_", t_col, " | grd_1km_id")
# m2_final <- lme(
#   fixed = as.formula(fixed),   # pred_m2_smooth ~ idw_txxx
#   random = as.formula(random), # ~ 1 + idw_xxx | grd_1km_id
#   data = mod2,
#   keep.data = FALSE,      # don't include a copy of mod2 in the model object
#   control = lmeControl(opt = "optim")
# )
# rm(fixed, random)
#
# # Generate final predictions for mod2 using m2_final and the splines
# # ~ 75 seconds, 7.5 GB
# paste(year, t_col, ": predicting mod2 using m2_final and splines") %>% report
# pred_m2_tps <- lapply(splines, function(tps) { tps$fitted.values }) %>% unlist %>% as.vector
# pred_m2_mix <- predict(m2) %>% as.vector
# mod2[, pred_m2_final := pred_m2_tps + pred_m2_mix]
# rm(pred_m2_tps, pred_m2_mix)
#
# # Save the final m2
# # ~ 15 min, ~ 3 GB file
# save_m2(m2, "m2_final")
#
# # Save the final m2 predictions
# # ~ 3 min, 4.4 GB file
# save_m2_preds(mod2, "pred_m2_final")
#
# # Save a raster summary of the m2 predictions
# # ~ 30 seconds
# smry <- summarize_as_raster(mod2, "pred_m2_final")
# save_m2_summary(smry, "final")
# plot(smry)
