#
# Fit stage 3 models (m3). These will generate 200 m estimates of the 1 km (m2) residuals
#

# Load packages, helper functions, and constants
source("init.R")

report("---")
report("Fitting stage 3 models")


# Functions -----

# A handler to suppress "generateDesign could only produce X points instead of 1000" warnings
# These occur when tuning random forest b/c MLR finds < 1000 reasonable hyperparameter combinations
suppressMLRTuneWarning <- function(w) {

  pattern <- "generateDesign could only produce \\d+ points instead of 1000"

  if (grepl(pattern, w)) {

    invokeRestart("muffleWarning")

  }

}

# Set up the learners -----

# Random forest
rf <- makeTuneWrapper(

        # Random forest from ranger
        learner = makeLearner(
                    cl = "regr.ranger",
                    importance = "impurity", # estimate feature importance by impurity
                    num.threads = ncores,    # let ranger run in parallel
                    num.trees = 400          # fix num.trees; more is better but takes longer
                  ),

        # Maximize R squared; also report RMSE, MAE, and time
        measures = list(mlr::rsq, mlr::rmse, mlr::mae, mlr::timetrain),

        # When tuning, train on two 20% samples of the data
        resampling = makeResampleDesc("Subsample", split = 0.2, iters = 2),

        # Hyperparameters to tune
        par.set = makeParamSet(
                    makeIntegerParam("mtry", 3, 10) # num features evaluated at each node
                  ),

        # Tune using mlrMBO with 6 iterations (initial design grid of 4 + 2)
        control = makeTuneControlMBO(budget = 6),

        # Don't report each tuning iteration
        show.info = FALSE

      )

# XGBoost
xg <- makeTuneWrapper(

        # XGBoost with fixed number of rounds and gamma
        learner = makeLearner(
                    cl = "regr.xgboost",
                    nthread = ncores,    # let xgboost run in parallel
                    nrounds = 100,       # number of iterations
                    gamma = 5            # regularization
                  ),

        # Maximize R squared; also report RMSE, MAE, and time
        measures = list(mlr::rsq, mlr::rmse, mlr::mae, mlr::timetrain),

        # When tuning, train on two 20% samples of the data
        resampling = makeResampleDesc("Subsample", split = 0.2, iters = 2),

        # Hyperparameters to tune
        par.set = makeParamSet(
                    makeNumericParam("eta", 0.1, 0.3),            # learning rate (small = slow)
                    makeIntegerParam("max_depth", 5, 20),         # max tree depth
                    makeIntegerParam("min_child_weight", 3, 30),  # min obs per node
                    makeNumericParam("colsample_bytree", 0.75, 1) # pct features evaluated per tree
                  ),

        # Tune using mlrMBO with 24 iterations (initial design grid of 16 + 8)
        control = makeTuneControlMBO(budget = 24),

        # Don't report each tuning iteration
        show.info = FALSE

      )

algos = list("rf" = rf, "xgb" = xg)

# Tune, train, and cross-validate models for all years -----

for (year in constants$model_years) {

  for (tvar in c("tmin", "tmean", "tmax")) {

    for (algo in names(algos)) {

      # Skip if model already exists
      model_path <- glue("1.work/models/m3_{tvar}_{year}_{algo}_model.rds")
      if (file.exists(model_path) && !OVERWRITE) {
        rm(model_path)
        next
      }

      # Load mod4 data
      if (!exists("mod4") || !identical(year, mod4[, year(date) %>% unique])) {

        glue("{year} {tvar} {algo}: Loading data") %>% report
        mod4 <- glue("1.work/joined/mod4_{year}.fst") %>%
                read_fst(as.data.table = TRUE)

        # Convert date to julian day
        mod4[, day := format(date, "%j") %>% as.integer]

      }

      # Define the response variable and features to use for prediction
      id_cols <- c("date", "stn_id", "grd_200m_id")
      target <- glue("res_{tvar}")
      features <- c(glue("pred_{tvar}"), "climate", "elev", "pop", "artificial", "vegetation",
                    "bare", "water", "ndvi", "bt", "x", "y", "day")

      # Get observations where the target and features are not NA
      task_data <- mod4[, c(id_cols, target, features), with = FALSE] %>%
                   na.omit
      rm(features)

      # Save the date, stn id, and grd id of each observation then remove these columns
      # Must save stn id because some cell-days are associated with multiple stations
      template <- task_data[, id_cols, with = FALSE]
      task_data <- task_data[, -id_cols, with = FALSE]
      rm(id_cols)

      # Define the task
      task <- makeRegrTask(
                id = glue("{tvar}_{year}_{algo}"),
                data = as.data.frame(task_data),
                target = target,
                coordinates = task_data[, c("x", "y")]
              )

      learner <- algos[[algo]]

      # Create dummy features for xgboost (which can't handle factors)
      if (getLearnerShortName(learner) %like% "xgboost") {

        task <- createDummyFeatures(task)

      }

      # Cross-validate the model, tuning within each CV fold to maximize r squared
      # Suppress mlrMBO warning when < 1000 possible hyperparameter combinations
      glue("{year} {tvar} {algo}: Cross-validating (with nested tuning)") %>% report
      cv <- withCallingHandlers(
              mlr::resample(
                learner = learner,
                task = task,
                resampling = makeResampleDesc("CV", iters = 5),
                measures = learner$measures,
                models = TRUE,                    # keep models (to get feature importances)
                extract = function(m) {
                            tr <- getTuneResult(m)
                            tr$mbo.result <- NULL # remove the (large) MBO result to save memory
                            tr
                          },
                show.info = TRUE
              ),
              warning = suppressMLRTuneWarning
            )

      # Get feature importances and tuning results then remove the models and extracts
      # This reduces the size on disk from ~ 100 MB to ~ 20 MB
      cv$importances <- lapply(cv$models, getFeatureImportance)
      cv$tune.results.x <- getNestedTuneResultsX(cv)
      partial <- ncol(cv$tune.results.x) > 1
      cv$hped <- generateHyperParsEffectData(cv, partial.dep = partial)
      cv$models <- NULL
      cv$extract <- NULL
      rm(partial)

      # Replace the preds object with a data.table including the date, station id, and grid cell id
      # Need station id b/c some cell-days are associated with multiple stations so date and grid
      # id alone are not enough to uniquely identify
      cv$pred <- getRRPredictions(cv) %>%
                 as.data.table() %>%
                 .[order(id), .(iter, truth, response)] %>%
                 setnames(
                   old = c("truth", "response"),
                   new = c(target, glue("pred_{target}"))
                 ) %>%
                 cbind(template, .)

      # Save the cross-validation
      cv_path <- glue("1.work/models/m3_{tvar}_{year}_{algo}_cv.rds")
      tmp_path <- glue("{cv_path}.tmp")
      saveRDS(cv, tmp_path)
      file.rename(tmp_path, cv_path)
      glue("{year} {tvar} {algo}: -> {cv_path}") %>% report
      rm(cv_path)

      # Plot the results
      dir.create("1.work/summaries/tuning", showWarnings = FALSE)
      glue("{year} {tvar} {algo}: Generating plots") %>% report
      {

        # Effect of each parameter
        for (par in cv$hped$hyperparams) {

          p <- plotHyperParsEffect(cv$hped, x = par, y = "rsq.test.mean",
                                   partial.dep.learn = "regr.ranger") +
                 geom_smooth(method = "loess") +
                 labs(title = cv$task.id)
          plot_path <- glue("1.work/summaries/tuning/m3_{tvar}_{year}_{algo}_rsq-{par}.png")
          ggsave(plot_path, p, width = 7, height = 7)
          glue("{year} {tvar} {algo}: -> {plot_path}") %>% report

        }

        # Optimization path, rsq
        p <- ggplot(cv$hped$data, aes(x = iteration, y = rsq.test.mean)) +
               geom_point() +
               geom_line() +
               facet_grid("nested_cv_run", labeller = label_both) +
               labs(title = cv$task.id)
        plot_path <- glue("1.work/summaries/tuning/m3_{tvar}_{year}_{algo}_tune-rsq.png")
        ggsave(plot_path, p, width = 7, height = 7)
        glue("{year} {tvar} {algo}: -> {plot_path}") %>% report

        # Optimization path, rmse
        p <- ggplot(cv$hped$data, aes(x = iteration, y = rmse.test.rmse)) +
               geom_point() +
               geom_line() +
               facet_grid("nested_cv_run", labeller = label_both) +
               labs(title = cv$task.id)
        plot_path <- glue("1.work/summaries/tuning/m3_{tvar}_{year}_{algo}_tune-rmse.png")
        ggsave(plot_path, p, width = 7, height = 7)
        glue("{year} {tvar} {algo}: -> {plot_path}") %>% report

        # Feature importance
        {

          # Get feature importance from each model
          imp <- seq_along(cv$importances) %>%
                 lapply(function(i) {
                   cv$importances[[i]] %$%
                     res %>%
                     gather("feature", "importance") %>%
                     cbind(fold = i)
                 }) %>%
                 rbindlist

          # Sort decreasing by mean importance across all models
          lvl <- imp[, .(imp = mean(importance)), by = "feature"] %>%
                 .[order(imp), feature]

          # Plot
          p <- ggplot(imp, aes(x = factor(feature, levels = lvl), y = importance)) +
                 geom_col(position = "dodge2") +
                 coord_flip() +
                 labs(x = "feature") +
                 facet_wrap("fold", labeller = label_both) +
                 labs(title = cv$task.id)
          plot_path <- glue("1.work/summaries/tuning/m3_{tvar}_{year}_{algo}_features.png")
          ggsave(plot_path, p, width = 7, height = 7)
          glue("{year} {tvar} {algo}: -> {plot_path}") %>% report

        }

      }
      rm(p, imp, lvl)

      # Tune and train a final model on the entire dataset
      glue("{year} {tvar} {algo}: Training final model") %>% report
      model <- withCallingHandlers(
                 train(learner, task = task),
                 warning = suppressMLRTuneWarning
               )

      # Save the model
      tmp_path <- glue("{model_path}.tmp")
      saveRDS(model, tmp_path)
      file.rename(tmp_path, model_path)
      glue("{year} {tvar} {algo}: -> {model_path}") %>% report
      rm(tmp_path, model_path)

      # Cleanup
      rm(cv, model, learner, task, task_data, template, target)

    }

  }

}

if (exists("mod4")) {
  rm(mod4)
}
rm(year, tvar, algo, rf, xg, algos, suppressMLRTuneWarning)

report(glue("Done {paste(range(constants$model_years), collapse = '-')}"))
