# Benchmark alternate stage 3 models (m3 = local 200 m)

library(magrittr)
library(data.table)
library(fst)
library(glue)
library(ranger)
library(mlr)
library(mlrMBO)
library(ggplot2)

source("2.scripts/helpers/constants.R")
source("2.scripts/helpers/report.R")
source("2.scripts/helpers/get_ncores.R")


# Setup -----

ncores <- get_ncores()
threads_fst(ncores)


# -----

year <- 2010
tvar <- "tmin"


# Load mod4
mod4 <- glue("1.work/joined/mod4_{year}.fst") %>%
        read_fst(as.data.table = TRUE)

# Convert date to julian day
mod4[, day := format(date, "%j") %>% as.integer]

# Define the columns to use in the model
model_cols <- c("climate", "elev", "pop", "artificial", "vegetation", "bare", "water", "ndvi", "bt",
                "x", "y", "day")

# Subset data
res_col <- glue("res_{tvar}")
pred_col <- glue("pred_{tvar}")
dt <- mod4[!is.na(get(res_col)), c(res_col, pred_col, model_cols), with = FALSE]
df <- dt %>%
      na.omit %>%
      as.data.frame

# Define the task
task <- makeRegrTask(
          id = glue("{tvar}_{year}"),
          data = df,
          target = res_col,
          coordinates = df[, c("x", "y")]
        )

# Set up the random forest learner
learner <- makeLearner(
             cl = "regr.ranger",
             par.vals = list(
               importance = "impurity", # estimate feature importance by impurity method
               num.threads = ncores     # let ranger grow trees and predict in parallel
             )
           )

#
# 1. Is tuning result stable with different subsamples?
# -> Yes, tuning seems fairly stable; var(sample) < var(parameter value)
{

  tune1_path <- "1.work/models/m3_explore_tune1.rds"

  if (file.exists(tune1_path)) {

    tune1 <- readRDS(tune1_path)

  } else {

    tune1 <- lapply(1:5, function(i) {
               glue("Iteration {i}") %>% print
               tuneParams(
                 learner = learner,
                 task = task,
                 resampling = makeResampleDesc("Holdout", split = 0.8),
                 measures = list(rsq, rmse, mae, timetrain),
                 par.set = makeParamSet(
                             makeIntegerParam("num.trees", 100, 500),
                             makeIntegerParam("mtry", 2, 6)
                           ),
                 control = makeTuneControlGrid(resolution = 3)
               )
             })
    saveRDS(tune1, tune1_path)

  }

  tune1

  hped1 <- generateHyperParsEffectData(tune1[[1]])
  hped1$nested <- TRUE
  hped1$data <- lapply(1:5, function(i) {
                  h <- generateHyperParsEffectData(tune1[[i]])
                  h$data$nested_cv_run <- i
                  h$data
                }) %>%
                do.call(rbind, .)

  # Effect of varying num.trees
  plotHyperParsEffect(hped1, x = "num.trees", y = "rsq.test.mean")

  # Effect of varying mtry
  plotHyperParsEffect(hped1, x = "mtry", y = "rsq.test.mean")

  # Best combinations of mtry and num.trees
  plotHyperParsEffect(hped1, x = "mtry", y = "num.trees", z = "rsq.test.mean", plot.type = "heatmap")

  # Best combinations of mtry and num.trees, interpolated
  plotHyperParsEffect(hped1, x = "mtry", y = "num.trees", z = "rsq.test.mean", plot.type = "contour",
                      interpolate = "regr.ksvm")


  # Performance vs. num.trees and holdout, faceted by num.trees
  ggplot(hped1$data, aes(x = mtry, y = rsq.test.mean, color = as.factor(nested_cv_run))) +
    geom_line() +
    geom_point() +
    facet_wrap("num.trees", labeller = "label_both")

  # Time to train model vs. num.trees and num.trees, faceted by holdout
  ggplot(hped1$data, aes(x = num.trees, y = timetrain.test.mean, color = as.factor(mtry))) +
    geom_line() +
    geom_point() +
    facet_wrap("nested_cv_run", labeller = "label_both")

  # Performance vs mtry and holdout, faceted by num.trees
  hped1_zoom <- hped1$data[hped1$data$mtry > 2, ]
  ggplot(hped1_zoom, aes(x = mtry, y = rsq.test.mean, color = as.factor(nested_cv_run))) +
    geom_point() +
    facet_wrap("num.trees", labeller = "label_both")

}


# 2. Is tuning result stable with different subsample sizes?
# -> Unstable for 7% sample; stable for larger samples
{

  tune2_path <- "1.work/models/m3_explore_tune2.rds"

  if (file.exists(tune2_path)) {

    tune2 <- readRDS(tune2_path)

  } else {

    tune2 <- lapply(1:5, function(i) {
               splt <- i * 0.1
               glue("Holdout of {splt}") %>% print
               tuneParams(
                 learner =learner,
                 task = task,
                 resampling = makeResampleDesc("Holdout", split = splt),
                 measures = list(rsq, rmse, mae, timetrain),
                 par.set = makeParamSet(
                             makeIntegerParam("num.trees", 100, 500),
                             makeIntegerParam("mtry", 2, 6)
                           ),
                 control = makeTuneControlGrid(resolution = 3)
               )
             })
    saveRDS(tune2, tune2_path)

  }

  tune2

  hped2 <- generateHyperParsEffectData(tune2[[1]])
  hped2$nested <- TRUE
  hped2$data <- lapply(1:5, function(i) {
                  h <- generateHyperParsEffectData(tune2[[i]])
                  h$data$nested_cv_run <- i
                  h$data
                }) %>%
                do.call(rbind, .)

  # Performance vs mtry and holdout, faceted by num.trees
  ggplot(hped2$data, aes(x = mtry, y = rsq.test.mean, color = as.factor(nested_cv_run))) +
    geom_line() +
    geom_point() +
    facet_wrap("num.trees", labeller = "label_both")

  # Performance vs. mtry and num.trees, faceted by holdout
  ggplot(hped2$data, aes(x = mtry, y = rsq.test.mean, color = as.factor(num.trees))) +
    geom_line() +
    geom_point() +
    facet_wrap("nested_cv_run", labeller = "label_both")

}


# WIP 3. Can we minimize both MAE and timetrain?
{

  tune3_path <- "1.work/models/m3_explore_tune3.rds"

  tune3 <- tuneParamsMultiCrit(
             learner = learner,
             task = downsample(task, 0.5),
             resampling = makeResampleDesc("Subsample", split = 0.33, iters = 3),
             measures = list(rsq, timetrain, rmse, mae),
             par.set = makeParamSet(
                         makeIntegerParam("num.trees", 100, 500),
                         makeIntegerParam("mtry", 2, 6)
                       ),
             control = makeTuneMultiCritControlMBO(n.objectives = 2, budget = 24)
           )
  saveRDS(tune3, tune3_path)

  tune3
  plotTuneMultiCritResult(tune3)
  plotHyperParsEffect(hped1, x = "mtry", y = "num.trees", z = "rsq.test.mean", plot.type = "contour",
                      interpolate = "regr.ksvm")

}
