#
# Fit stage 4 models (m4). These will ensemble prediction of 200 m residuals
#

# Load packages, helper functions, and constants
source("init.R")

report("---")
report("Fitting stage 4 models")


# Fit ensemble model -----
for (year in constants$model_years) {

  for (tvar in c("tmin", "tmean", "tmax")) {

    # Skip if model already exists
    model_path <- glue("1.work/models/m4_{tvar}_{year}_model.rds")
    if (file.exists(model_path) && !OVERWRITE) {

      next

    }

    # Load the predictions from all algorithms
    if (!exists("preds")) {

      glue("{year} {tvar}: loading m3 predictions") %>% report
      preds_rf <- glue("1.work/predictions/mod4_{year}_pred_m3_rf.fst") %>%
                  read_fst(as.data.table = TRUE)
      names(preds_rf) %>% gsub("(pred_res_.+)", glue("\\1_rf"), .) %>%
                          setnames(preds_rf, .)

      preds_xgb <- glue("1.work/predictions/mod4_{year}_pred_m3_xgb.fst") %>%
                   read_fst(as.data.table = TRUE) %>%
                   .[, .(date, grd_200m_id, pred_res_tmin, pred_res_tmean, pred_res_tmax)]
      names(preds_xgb) %>% gsub("(pred_res_.+)", glue("\\1_xgb"), .) %>%
                           setnames(preds_xgb, .)

      preds <- preds_rf[preds_xgb, on = c("date", "grd_200m_id")]
      rm(preds_rf, preds_xgb)

    }

    # Subset the data
    target <- glue("res_{tvar}")
    rf_col <- glue("pred_{target}_rf")
    xgb_col <- glue("pred_{target}_xgb")
    dat <- preds[, c("date", "grd_200m_id", "x", "y", target, rf_col, xgb_col), with = FALSE] %>%
           na.omit

    # Name of column for ensemble predictions
    ens_col <- glue("pred_{target}_ens")

    # Formula for ensemble model
    # Note: explanation of tensor product https://stats.stackexchange.com/questions/45446/intuition-behind-tensor-product-interactions-in-gams-mgcv-package-in-r#
    ens_fml <- glue("{target} ~ te(x, y, by = {rf_col}) + te(x, y, by = {xgb_col})") %>%
               as.formula

    # Cross-validate
    glue("{year} {tvar}: cross-validating") %>% report
    resamp <- makeResampleInstance("CV", size = nrow(dat), iters = 10)
    cv <- 1:length(resamp$train.inds) %>%
          mclapply(function(fold) {

            # Ensemble the predictions using a GAM with geographically varying coefficients
            # This takes into account if there are areas where one model is better than the other
            # Note: bam with tensor smoothers is ~40x faster than gam with thin plate splines
            model <- bam(ens_fml, data = dat[resamp$train.inds[[fold]], ])

            # Predict for the test data
            test_dat <- dat[resamp$test.inds[[fold]], ]
            preds <- predict(model, newdata = test_dat)

            # Return test data with predictions
            test_dat[, c(ens_col) := preds]
            test_dat[, "fold" := fold]

          }, mc.cores = ncores, mc.preschedule = FALSE) %>%
          rbindlist

    # Save CV
    write_fst(cv, glue("1.work/models/m4_{tvar}_{year}_cv.fst"), compress = 100)

    # Assess performance
    glue("{year} {tvar}: assessing performance") %>% report
    perf <- assess_preds(cv, target, ens_col, "grd_200m_id") %>%
            as.list %>%
            c(year = year, tvar = tvar, .) %>%
            as.data.table
    perf[, lapply(.SD, round, 4), by = .(year, tvar)] %>%
      .[, .(year, tvar, r2, rmse, mae, r2.space, r2.time, res.mean, res.sd, i, slope)] %>%
      report()
    fwrite(perf, glue("1.work/models/m4_{tvar}_{year}_perf.csv"))

    # Fit a final model
    glue("{year} {tvar}: fitting final ensemble") %>% report
    model <- bam(ens_fml, data = dat)

    # Save the model
    tmp_path <- glue("{model_path}.tmp")
    saveRDS(model, tmp_path)
    file.rename(tmp_path, model_path)
    glue("{year} {tvar}: -> {model_path}") %>% report

    rm(dat, cv, model)

  }

  if (exists("preds")) {
    rm(preds)
  }
  rm(tvar)

}

rm(model_path, year)

report(glue("Done {paste(range(constants$model_years), collapse = '-')}"))
