# Compare thin plate spline vs. tensor product interaction for ensembling

library(magrittr)
library(glue)
library(data.table)
library(fst)
library(mgcv) # for glm
library(mlr)  # for resampling framework

source("2.scripts/helpers/report.R")
source("2.scripts/helpers/get_ncores.R")
source("2.scripts/helpers/assess_preds.R")


year <- 2001
tvar <- "tmin"

ncores <- get_ncores()
threads_fst(ncores)

# Load the predictions from all algorithms
preds_rf <- glue("1.work/predictions/mod4_{year}_pred_m3_rf.fst") %>%
            read_fst(as.data.table = TRUE)
names(preds_rf) %>% gsub("(pred_res_.+)", glue("\\1_rf"), .) %>%
                    setnames(preds_rf, .)

preds_xgb <- glue("1.work/predictions/mod4_{year}_pred_m3_xgb.fst") %>%
             read_fst(as.data.table = TRUE) %>%
             .[, .(date, grd_200m_id, pred_res_tmin, pred_res_tmean, pred_res_tmax)]
names(preds_xgb) %>% gsub("(pred_res_.+)", glue("\\1_xgb"), .) %>%
                     setnames(preds_xgb, .)

preds <- preds_rf[preds_xgb, on = c("date", "grd_200m_id")]
rm(preds_rf, preds_xgb)


# Subset the data
target <- glue("res_{tvar}")
rf_col <- glue("pred_{target}_rf")
xgb_col <- glue("pred_{target}_xgb")
dat <- preds[, c("date", "grd_200m_id", "x", "y", target, rf_col, xgb_col), with = FALSE] %>%
       na.omit

# Make cv sets
resamp <- makeResampleInstance("CV", size = nrow(dat), iters = 10)

ens_col <- glue("pred_{target}_ens")


# 575 sec
report("CVing gam with s")
start <- Sys.time()
cv_gam_s <- 1:length(resamp$train.inds) %>%
            mclapply(function(fold) {

              # Fit the model
              model <- glue("{target} ~ s(x, y, by = {rf_col}) + s(x, y, by = {xgb_col})") %>%
                       as.formula %>%
                       gam(data = dat[resamp$train.inds[[fold]], ])

              # Predict for the test data
              test_dat <- dat[resamp$test.inds[[fold]], ]
              preds <- predict(model, newdata = test_dat)

              # Return test data with predictions
              test_dat[, c(ens_col) := preds]
              test_dat[, "fold" := fold]

            }, mc.cores = ncores) %>%
            rbindlist
end <- Sys.time()
results <- assess_preds(cv_gam_s, target, ens_col, "grd_200m_id") %>%
           round(4) %>%
           as.list %>%
           c(model = "gam_s", time = round(difftime(end, start, units = "secs")), .) %>%
           as.data.table

# 28 sec
report("CVing gam with te")
start <- Sys.time()
cv_gam_t <- 1:length(resamp$train.inds) %>%
              mclapply(function(fold) {

                # Fit the model
                model <- glue("{target} ~ te(x, y, by = {rf_col}) + te(x, y, by = {xgb_col})") %>%
                         as.formula %>%
                         gam(data = dat[resamp$train.inds[[fold]], ])

                # Predict for the test data
                test_dat <- dat[resamp$test.inds[[fold]], ]
                preds <- predict(model, newdata = test_dat)

                # Return test data with predictions
                test_dat[, c(ens_col) := preds]
                test_dat[, "fold" := fold]

              }, mc.cores = ncores) %>%
              rbindlist
end <- Sys.time()
results <- assess_preds(cv_gam_t, target, ens_col, "grd_200m_id") %>%
           round(4) %>%
           as.list %>%
           c(model = "gam_te", time = round(difftime(end, start, units = "secs")), .) %>%
           rbind(results, .)

# 14 sec
report("CVing bam with te")
start <- Sys.time()
cv_bam <- 1:length(resamp$train.inds) %>%
          mclapply(function(fold) {

            # Fit the model
            model <- glue("{target} ~ te(x, y, by = {rf_col}) + te(x, y, by = {xgb_col})") %>%
                     as.formula %>%
                     bam(data = dat[resamp$train.inds[[fold]], ])

            # Predict for the test data
            test_dat <- dat[resamp$test.inds[[fold]], ]
            preds <- predict(model, newdata = test_dat)

            # Return test data with predictions
            test_dat[, c(ens_col) := preds]
            test_dat[, "fold" := fold]

          }, mc.cores = ncores) %>%
          rbindlist
end <- Sys.time()
results <- assess_preds(cv_bam, target, ens_col, "grd_200m_id") %>%
           round(4) %>%
           as.list %>%
           c(model = "bam_te", time = round(difftime(end, start, units = "secs")), .) %>%
           rbind(results, .)

results



# -----




year <- 2008
tvar <- "tmax"

# Load the predictions from all algorithms
preds_rf <- glue("1.work/predictions/mod4_{year}_pred_m3_rf.fst") %>%
            read_fst(as.data.table = TRUE)
names(preds_rf) %>% gsub("(pred_res_.+)", glue("\\1_rf"), .) %>%
                    setnames(preds_rf, .)

preds_xgb <- glue("1.work/predictions/mod4_{year}_pred_m3_xgb.fst") %>%
             read_fst(as.data.table = TRUE) %>%
             .[, .(date, grd_200m_id, pred_res_tmin, pred_res_tmean, pred_res_tmax)]
names(preds_xgb) %>% gsub("(pred_res_.+)", glue("\\1_xgb"), .) %>%
                     setnames(preds_xgb, .)

preds <- preds_rf[preds_xgb, on = c("date", "grd_200m_id")]
rm(preds_rf, preds_xgb)


# Subset the data
target <- glue("res_{tvar}")
rf_col <- glue("pred_{target}_rf")
xgb_col <- glue("pred_{target}_xgb")
dat <- preds[, c("date", "grd_200m_id", "x", "y", target, rf_col, xgb_col), with = FALSE] %>%
       na.omit

# Make cv sets
resamp <- makeResampleInstance("CV", size = nrow(dat), iters = 10)

ens_col <- glue("pred_{target}_ens")

report("CVing bam of rf and xgb")
start <- Sys.time()
cv_bam_both <- 1:length(resamp$train.inds) %>%
               mclapply(function(fold) {

                 # Fit the model
                 model <- glue("{target} ~ te(x, y, by = {rf_col}) + te(x, y, by = {xgb_col})") %>%
                          as.formula %>%
                          gam(data = dat[resamp$train.inds[[fold]], ])

                 # Predict for the test data
                 test_dat <- dat[resamp$test.inds[[fold]], ]
                 preds <- predict(model, newdata = test_dat)

                 # Return test data with predictions
                 test_dat[, c(ens_col) := preds]
                 test_dat[, "fold" := fold]

               }, mc.cores = ncores) %>%
               rbindlist
end <- Sys.time()
results2 <- assess_preds(cv_bam_both, target, ens_col, "grd_200m_id") %>%
            round(4) %>%
            as.list %>%
            c(model = "bam_both", time = round(difftime(end, start, units = "secs")), .) %>%
            as.data.table

report("CVing bam of rf only")
start <- Sys.time()
cv_bam_rf <- 1:length(resamp$train.inds) %>%
             mclapply(function(fold) {

               # Fit the model
               model <- glue("{target} ~ te(x, y, by = {rf_col})") %>%
                        as.formula %>%
                        gam(data = dat[resamp$train.inds[[fold]], ])

               # Predict for the test data
               test_dat <- dat[resamp$test.inds[[fold]], ]
               preds <- predict(model, newdata = test_dat)

               # Return test data with predictions
               test_dat[, c(ens_col) := preds]
               test_dat[, "fold" := fold]

             }, mc.cores = ncores) %>%
             rbindlist
end <- Sys.time()
results2 <- assess_preds(cv_bam_rf, target, ens_col, "grd_200m_id") %>%
            round(4) %>%
            as.list %>%
            c(model = "bam_rf", time = round(difftime(end, start, units = "secs")), .) %>%
            rbind(results2, .)

report("CVing bam of xgb only")
start <- Sys.time()
cv_bam_xgb <- 1:length(resamp$train.inds) %>%
              mclapply(function(fold) {

                # Fit the model
                model <- glue("{target} ~ te(x, y, by = {xgb_col})") %>%
                         as.formula %>%
                         gam(data = dat[resamp$train.inds[[fold]], ])

                # Predict for the test data
                test_dat <- dat[resamp$test.inds[[fold]], ]
                preds <- predict(model, newdata = test_dat)

                # Return test data with predictions
                test_dat[, c(ens_col) := preds]
                test_dat[, "fold" := fold]

              }, mc.cores = ncores) %>%
              rbindlist
end <- Sys.time()
results2 <- assess_preds(cv_bam_xgb, target, ens_col, "grd_200m_id") %>%
            round(4) %>%
            as.list %>%
            c(model = "bam_xgb", time = round(difftime(end, start, units = "secs")), .) %>%
            rbind(results2, .)

results2
