#
# Generate and plot rasters summarizing each annual model's predictions
#

# Load packages, helper functions, and constants
source("init.R")

report("---")
report("Summarizing annual model predictions")

# Functions ----

save_pred_plots <- function(year, model, force = FALSE) {

  if (!model %in% c("m1", "m2", "1km", "m3_rf", "m3_xgb", "m4", "200m")) {
    stop(glue("Unrecognized model: {model}"))
  }

  # Skip existing unless forced
  tvars <- c("tmin", "tmean", "tmax") %>%
           purrr::set_names()
  plot_paths <- glue("1.work/summaries/preds/pred_{model}_{year}_{tvars}.png") %>%
                purrr::set_names(tvars)
  if (all(file.exists(plot_paths)) && !force) {
    return(plot_paths)
  }

  # Save *all* summaries
  # Need to plot all to create an overall plot showing all 3 tvars
  report(glue("{year} {model}: plotting summary rasters"))
  plots <- map(tvars, function(tvar) {

             # Plot each layer
             smry <- glue("1.work/summaries/preds/pred_{model}_{year}_{tvar}.tif") %>%
                     load_geotiff()
             label <- ifelse(model %in% c("m3_rf", "m3_xgb", "m4"), glue("res_{tvar}"), tvar)
             plot <- plot_summary_raster(smry, label)

             # Save as png
             path <- plot_paths[tvar]
             jpeg(path, width = 1920, height = 1280, res = 150)
             print(plot$mean, split = c(1, 1, 3, 2), more = TRUE)
             print(plot$sd,   split = c(2, 1, 3, 2), more = TRUE)
             print(plot$N,    split = c(3, 1, 3, 2), more = TRUE)
             print(plot$min,  split = c(1, 2, 3, 2), more = TRUE)
             print(plot$max,  split = c(2, 2, 3, 2))
             dev.off()
             report(glue("{year} {model}: -> {path}"))

             # Return the plot
             plot

           })

  # Save an overall plot
  path <- glue("1.work/summaries/preds/pred_{model}_{year}.png")
  jpeg(path, width = 1920, height = 1920, res = 150)
  print(plots$tmin$mean,  split = c(1, 1, 3, 3), more = TRUE)
  print(plots$tmin$sd,    split = c(1, 2, 3, 3), more = TRUE)
  print(plots$tmin$N,     split = c(1, 3, 3, 3), more = TRUE)
  print(plots$tmean$mean, split = c(2, 1, 3, 3), more = TRUE)
  print(plots$tmean$sd,   split = c(2, 2, 3, 3), more = TRUE)
  print(plots$tmean$N,    split = c(2, 3, 3, 3), more = TRUE)
  print(plots$tmax$mean,  split = c(3, 1, 3, 3), more = TRUE)
  print(plots$tmax$sd,    split = c(3, 2, 3, 3), more = TRUE)
  print(plots$tmax$N,     split = c(3, 3, 3, 3))
  dev.off()
  report(glue("{year} {model}: -> {path}"))

  return(plot_paths)

}

save_pred_summaries <- function(year, model, force = FALSE) {

  if (!model %in% c("m1", "m2", "1km", "m3_rf", "m3_xgb", "m4", "200m")) {
    stop(glue("Unrecognized model: {model}"))
  }

  # Skip if exists unless forced
  tvars <- c("tmin", "tmean", "tmax")
  smry_paths <- glue("1.work/summaries/preds/pred_{model}_{year}_{tvars}.tif") %>%
                purrr::set_names(tvars)
  if (force) {
    missing <- smry_paths
  } else {
    missing <- smry_paths[!file.exists(smry_paths)]
  }
  if (length(missing) == 0) {
    return(smry_paths)
  }
  tvars <- names(missing)

  # Load predictions
  report(glue("{year} {model}: loading predictions"))
  if (model %in% c("m1", "m2", "1km")) {

    pred_path <- glue("1.work/predictions/mod3_{year}_pred_{model}.fst")
    ref_cols <- c("date", "grd_1km_id", "grd_1km_x", "grd_1km_y")

    if (model == "1km") {
      pred_cols <- tvars
    } else {
      pred_cols <- glue("{model}_{tvars}")
    }

  } else {

    pred_path <- glue("1.work/predictions/mod5_{year}_pred_{model}.fst")
    ref_cols <- c("date", "grd_200m_id", "x", "y")

    if (model %like% "m3") {
      pred_path <- sub(model, "m3", pred_path)
      pred_cols <- strsplit(model, "_")[[1]][2] %>%
                   glue("pred_res_{tvars}_", .)
    } else if (model == "m4") {
      pred_cols <- glue("pred_res_{tvars}_ens")
    } else {
      pred_cols <- tvars
    }

  }
  names(pred_cols) <- tvars
  preds <- read_fst(pred_path, columns = c(ref_cols, pred_cols), as.data.table = TRUE)

  # Ensure the output dir exists
  dir.create(dirname(missing[1]), recursive = TRUE, showWarnings = FALSE)

  report(glue("{year} {model}: creating summary rasters"))
  for (tvar in tvars) {

    if (model %in% c("m1", "m2", "1km")) {

      smry <- summarize_1km_preds(preds, smry_col = pred_cols[tvar])

    } else {

      smry <- summarize_200m_preds(preds, smry_col = pred_cols[tvar])

    }

    # Save
    smry_path <- missing[tvar]
    save_geotiff(smry, smry_path, overwrite = TRUE)
    report(glue("{year} {model}: -> {smry_path}"))

  }

  return(smry_paths)

}

# Summarize all models -----

for (year in constants$model_years) {

  for (model in c("m1", "m2", "1km", "m3_rf", "m3_xgb", "m4", "200m")) {

    save_pred_summaries(year, model, force = OVERWRITE)
    save_pred_plots(year, model, force = OVERWRITE)

  }

}

rm(year, model, save_pred_plots, save_pred_summaries)

report(glue("Done {paste(range(constants$model_years), collapse = '-')}"))
