# Generate final 1 km predictions:
# * Fill gaps in stage 1 predictions with stage 2 predictions
# * Fill remaining gaps (over water / snow) with value of closest cell on same day

# Load packages, helper functions, and constants
source("init.R")

report("---")
report("Generating final stage 3 predictions (m1, m2, mod3)")


# Generate final predictions --------------------------------------------------

# Iterate over years
for (year in constants$model_years) {

  # Skip if predictions already exist
  pred_1km_path <- paste("mod3", year, "pred_1km.fst", sep = "_") %>%
                     file.path(constants$predictions_dir, .)
  if (file.exists(pred_1km_path) && !OVERWRITE) {
    next
  }

  # Load mod3 data including m1 and m2 predictions
  # ~
  mod3_path <- paste("mod3", year, "pred_m2.fst", sep = "_") %>%
               file.path(constants$predictions_dir, .)
  paste(year, ": loading", basename(mod3_path)) %>% report
  threads_fst(ncores)
  mod3 <- read_fst(mod3_path, as.data.table = TRUE)

  # Get the best prediction for each cell-day and fill any gaps
  # ~ 120 seconds
  t_cols <- c("tmin", "tmean", "tmax")
  for (t_col in t_cols) {

    paste(year, t_col, ": getting best prediction for each cell-day") %>% report
    src_col <- paste0(t_col, "_src")

    # Start with m1 predictions
    m1_col <- paste0("m1_", t_col)
    mod3[, c(t_col, src_col) := .(get(m1_col), 1L)]
    rm(m1_col)

    # Fill blanks with m2 predictions
    m2_col <- paste0("m2_", t_col)
    mod3[get(t_col) %>% is.na, c(t_col, src_col) := .(get(m2_col), 2L)]
    rm(m2_col)

    # Fill any remaining gaps with the best prediction for the nearest grid cell. Most gaps occur
    # when a cell never has LST, but some gaps in 2001 - 2004 only occur on some days; unclear why
    # ~ 30 seconds
    if (mod3[get(t_col) %>% is.na, .N] > 0) {

      paste(year, t_col, ": filling gaps (mostly over snow and water)") %>% report

      # Get all cell-days that are missing predictions
      gaps <- mod3[get(t_col) %>% is.na, .(date, grd_1km_id, grd_1km_x, grd_1km_y)]

      # Get locations of missing predictions
      missing <- gaps[, .(x = mean(grd_1km_x), y = mean(grd_1km_y)), by = grd_1km_id] %>%
                 st_as_sf(coords = c("x", "y"), crs = 2154)

      # Get locations that have a prediction for every day
      present <- mod3[date == first(date), .(grd_1km_id, x = grd_1km_x, y = grd_1km_y)] %>%
                 .[!(grd_1km_id %in% missing$grd_1km_id), ] %>%
                 st_as_sf(coords = c("x", "y"), crs = 2154)

      # Match each missing location to the nearest location that always has predictions
      nn <- st_nn(missing, present)
      matches <- data.table(
                   gap_id = missing$grd_1km_id,
                   fill_id = present$grd_1km_id[nn %>% unlist]
                 )
      rm(missing, present, nn)
      # mapview(st_connect(gaps, pts, ids = nn))

      # Associate the nearest location with each cell-day missing a prediction
      gaps <- gaps[matches, .(date, grd_1km_id, fill_id), on = c(grd_1km_id = "gap_id")] %>%
              setkeyv(c("date", "grd_1km_id", "fill_id"))

      # Look up the predictions for the matched cell-days
      fill <- mod3[gaps, .(fill_val = get(t_col)), on = c("date", grd_1km_id = "fill_id")]
      gaps[, fill_val := fill]

      # Fill the gaps in mod3
      mod3[gaps, (t_col) := fill_val, on = c("date", "grd_1km_id")]
      mod3[gaps, (src_col) := 3L, on = c("date", "grd_1km_id")]
      rm(matches, gaps, fill)

    }

    # Round
    mod3[, (t_col) := round(get(t_col), 2)]

  }

  # Drop unneeded columns
  src_cols <- paste0(t_cols, "_src")
  mod3 <- mod3[, c("date", "grd_1km_id", "grd_1km_x", "grd_1km_y", t_cols, src_cols), with = FALSE]

  # Save the final predictions
  # ~ 5 min, ~ 4.6 GB file
  paste(year, ": saving final 1 km predictions") %>% report
  threads_fst(ncores)
  write_fst(mod3, pred_1km_path, compress = 100)
  paste(year, ": ->", basename(pred_1km_path)) %>% report

  # Cleanup
  paste(year, ": cleaning environment") %>% report
  rm(mod3)
  gc()

}

rm(year, pred_1km_path)

report(glue("Done {paste(range(constants$model_years), collapse = '-')}"))
