#
# Generate stage 4 predictions (mod4) using stage 3 models (m3)
#

# Load packages, helper functions, and constants
source("init.R")

report("---")
report("Generating stage 4 predictions (m3, mod4)")


# Predict using m3 ------------------------------------------------------------

# Iterate over years
for (year in constants$model_years) {

  for (algo in c("rf", "xgb")) {

    # Skip if predictions already exist
    pred_path <- glue("1.work/predictions/mod4_{year}_pred_m3_{algo}.fst")
    if (file.exists(pred_path) && !OVERWRITE) {

      next

    }

    # Load mod4 data
    if (!exists("mod4") || !identical(year, mod4[, year(date) %>% unique])) {

      glue("{year} {algo}: loading data") %>% report
      mod4 <- glue("1.work/joined/mod4_{year}.fst") %>%
              read_fst(as.data.table = TRUE)

      # Convert date to julian day
      mod4[, day := format(date, "%j") %>% as.integer]

    }

    # Predict for each temperature
    for (tvar in c("tmin", "tmean", "tmax")) {

      # Load the model
      # ~ 20 sec, 1.2 GB
      glue("{year} {algo} {tvar}: loading model") %>% report
      m3 <- glue("1.work/models/m3_{tvar}_{year}_{algo}_model.rds") %>%
            readRDS

      # Get observations where the model features are not NA
      if (algo == "xgb") {

        # Create dummy features for xgboost (which can't handle factors)
        features <- m3$features[!grepl("climate", m3$features)]
        dat <- mod4[, c("date", "grd_200m_id", features, "climate"), with = FALSE] %>%
               na.omit %>%
               as.data.frame %>%
               createDummyFeatures %>%
               as.data.table
        rm(features)

      } else {

        dat <- mod4[, c("date", "grd_200m_id", m3$features), with = FALSE] %>%
               na.omit

      }

      # Predict
      glue("{year} {algo} {tvar}: predicting") %>% report
      preds <- dat[, m3$features, with = FALSE] %>%
               as.data.frame %>%
               predict(m3, newdata = .) %>%
               getPredictionResponse

      # Join the predictions to mod4
      to_join <- dat[, .(date, grd_200m_id)] %>%
                 cbind(pred_m3 = preds) %>%
                 .[, .(pred_m3 = mean(pred_m3)), by = c("date", "grd_200m_id")] %>%
                 .[mod4, .(pred_m3), on = c("date", "grd_200m_id")]
      mod4[, glue("pred_{m3$task.desc$target}") := to_join]

      # Cleanup
      rm(m3, dat, preds, to_join)

    }
    rm(tvar)

    # Save the predictions
    glue("{year} {algo}: saving predictions") %>% report
    tmp_path <- glue(pred_path, ".tmp")
    mod4[, .(date, grd_200m_id, x, y, res_tmin, res_tmean, res_tmax, pred_res_tmin, pred_res_tmean,
             pred_res_tmax)] %>%
      write_fst(tmp_path, compress = 100)
    file.rename(tmp_path, pred_path)
    glue("{year} {algo}: -> {pred_path}") %>% report
    rm(tmp_path)

  }

}

if (exists("mod4")) {
  rm(mod4)
}
rm(algo, year, pred_path)

report(glue("Done {paste(range(constants$model_years), collapse = '-')}"))
