#
# Generate stage 5 predictions (mod5) using stage 3 models (m3)
# Step 1: predict for entire 200 m grid with RF and XGB
#

# Load packages, helper functions, and constants
source("init.R")

report("---")
report("Generating stage 5 predictions, step 1: predict with RF and XGB (m3, mod5)")


# Predict using m3 ------------------------------------------------------------

# Iterate over years
for (year in constants$model_years) {

  # Skip if predictions already exist
  pred_path <- glue("1.work/predictions/mod5_{year}_pred_m3.fst")
  if (file.exists(pred_path) && !OVERWRITE) {

    next

  }

  # Load mod5 data
  # ~ 3 min, ~ 10.1 GB
  glue("{year}: Loading data") %>% report
  mod5 <- glue("1.work/joined/mod5_{year}.fst") %>%
          read_fst(as.data.table = TRUE)

  # Convert date to julian day
  mod5[, day := format(date, "%j") %>% as.integer]

  # Ensure mod5 includes x and y coordinates
  if (length(setdiff(c("x", "y"), names(mod5))) != 0) {

    # Add coordinates of 200 m grid centroids
    grd <- read_fst("1.work/grids/grd_200m.fst", as.data.table = TRUE)
    mod5[, c("x", "y") := grd[mod5, .(x, y), on = c(id = "grd_200m_id")]]
    rm(grd)

  }

  # Predict for each temperature
  for (tvar in c("tmin", "tmean", "tmax")) {

    # Predict with each algorithm
    for (algo in c("rf", "xgb")) {

      # Load the model
      # ~ 20 sec, 1.2 GB
      glue("{year} {tvar} {algo}: loading model") %>% report
      m3 <- glue("1.work/models/m3_{tvar}_{year}_{algo}_model.rds") %>%
            readRDS

      # Get observations where the model features are not NA
      glue("{year} {tvar} {algo}: formatting data") %>% report
      if (algo == "xgb") {

        # Create dummy features for xgboost (which can't handle factors)
        features <- m3$features[!grepl("climate", m3$features)]
        dat <- mod5[, c("date", "grd_200m_id", features, "climate"), with = FALSE] %>%
               na.omit
        for (clim in levels(dat$climate)) {
          dat[, paste0("climate.", clim) := 0L]
          dat[climate == clim, paste0("climate.", clim) := 1L]
        }
        dat[, climate := NULL]

        rm(clim,features)

      } else {

        dat <- mod5[, c("date", "grd_200m_id", m3$features), with = FALSE] %>%
               na.omit

      }

      # Predict
      # Process in chunks so as not to exceed available memory
      # ~ 48 GB needed for data + m3
      # ~ 5 kB / row for rf prediction; ~ 1 kB / row for xgb
      nsplits <- dat[, .N] * ifelse(algo == "rf", 5, 1) / (get_mem() - 50000000)
      nsplits <- ceiling(nsplits)
      dat[, m3_pred_grp := ceiling(.I / (.N / nsplits))]
      preds <- lapply(1:nsplits, function(grp) {
                 glue("{year} {tvar} {algo}: predicting for chunk {grp} of {nsplits}") %>% report
                 dat[m3_pred_grp == grp, m3$features, with = FALSE] %>%
                   as.data.frame() %>%
                   predict(m3, newdata = .) %>%
                   getPredictionResponse()
               }) %>%
               unlist()

      # Join the predictions to mod5
      glue("{year} {tvar} {algo}: joining predictions to mod5") %>% report
      dat %<>% .[, pred_m3 := preds] %>%
               .[, .(pred_m3 = mean(pred_m3)), by = .(date, grd_200m_id)] %>%
               .[mod5, .(pred_m3), on = .(date, grd_200m_id)]
      mod5[, glue("pred_{m3$task.desc$target}_{algo}") := dat]

      # Cleanup
      rm(dat, m3, preds)
      gc()

    }

  }

  # Save the predictions
  glue("{year}: saving predictions") %>% report
  pred_cols <- names(mod5) %>%
               grepl("pred_", .) %>%
               magrittr::extract(names(mod5), .)
  tmp_path <- glue(pred_path, ".tmp")
  write_fst(mod5[, c("date", "grd_200m_id", "x", "y", pred_cols), with = FALSE], tmp_path,
            compress = 100)
  file.rename(tmp_path, pred_path)
  glue("{year}: -> {pred_path}") %>% report

  # Cleanup
  rm(mod5, tmp_path, pred_cols, tvar, algo)

}

rm(pred_path, year)

report(glue("Done {paste(range(constants$model_years), collapse = '-')}"))
