#
# Generate stage 5 predictions (mod5) using stage 4 models (m4)
# Step 2: predict for entire 200 m grid with ensemble
#

# Load packages, helper functions, and constants
source("init.R")

report("---")
report("Generating stage 5 predictions, step 2: predict with ensemble (m4, mod5)")


# Predict using m4 ------------------------------------------------------------

# Iterate over years
for (year in constants$model_years) {

  # Skip if predictions already exist
  pred_path <- glue("1.work/predictions/mod5_{year}_pred_m4.fst")
  if (file.exists(pred_path) && !OVERWRITE) {

    next

  }

  # Load mod5 data with m3 predictions
  # ~ 60 sec, ~ 8.7 GB
  glue("{year}: Loading data") %>% report
  mod5 <- glue("1.work/predictions/mod5_{year}_pred_m3.fst") %>%
          read_fst(as.data.table = TRUE)

  # Predict for each temperature
  # Requires ~ 35 GB memory
  for (tvar in c("tmin", "tmean", "tmax")) {

    # Load the m4 ensemble model
    glue("{year}: {tvar}: loading m4") %>% report
    m4 <- glue("1.work/models/m4_{tvar}_{year}_model.rds") %>%
          readRDS

    # Predict residuals
    glue("{year}: {tvar}: predicting residuals") %>% report
    resid <- predict(m4, newdata = mod5, block.size = 100000) %>%
             as.numeric
    mod5[, glue("pred_res_{tvar}_ens") := resid]
    rm(m4, resid)

  }

  # Calculate final 200 m Ta predictions by adding 1 km predicted Ta and 200 m predicted residual
  for (tvar in c("tmin", "tmean", "tmax")) {

    # Add predicted residual to predicted temperature and round
    glue("{year}: {tvar}: calculating final 200 m Ta") %>% report
    mod5[, (tvar) := round(get(glue("pred_{tvar}")) + get(glue("pred_res_{tvar}_ens")), 2)]

    # Fill gaps in final 200 m predictions with nearest prediction within 1 km
    # These are always along coast / border of France; unsure why they occur
    glue("{year}: {tvar}: filling gaps in final 200 m Ta") %>% report
    gaps <- mod5[is.na(get(tvar)), .(date, grd_200m_id, x, y)]
    gap_pts <- gaps[, lapply(.SD, first), .SDcols = c("x", "y"), by = grd_200m_id] %>%
               st_as_sf(coords = c("x", "y"), crs = 3035)
    fill_pts <- mod5[date == first(date) & !is.na(get(tvar)), .(grd_200m_id, x, y)] %>%
                st_as_sf(coords = c("x", "y"), crs = 3035)
    nn <- st_nn(gap_pts, fill_pts, returnDist = TRUE, progress = FALSE)
    matches <- data.table(
                 gap_id = gap_pts$grd_200m_id,
                 fill_id = fill_pts$grd_200m_id[unlist(nn$nn)],
                 dist = unlist(nn$dist)
               ) %>%
               .[dist < 1000, ]

    # # DEBUG plot distant gaps
    # grd_pts <- read_fst("1.work/grids/grd_200m.fst", as.data.table = TRUE) %>%
    #            st_as_sf(coords = c("x", "y"), crs = 3035)
    # far <- matches[dist > 800, ] %>%
    #        .[, pair_id := 1:.N]
    # pairs <- rbind(
    #            as.data.table(gap_pts)[far, .(grd_200m_id, pair_id, type = "gap", dist, geometry), on = .(grd_200m_id = gap_id)],
    #            as.data.table(fill_pts)[far, .(grd_200m_id, pair_id, type = "fill", dist, geometry), on = .(grd_200m_id = fill_id)]
    #          ) %>%
    #          .[order(pair_id), ] %>%
    #          st_as_sf()
    # lines <- dplyr::group_by(pairs, pair_id, dist) %>%
    #          dplyr::summarise() %>%
    #          st_cast("LINESTRING")
    # near <- filter(pairs, type == "gap") %>%
    #         st_nn(grd_pts, k = 9, maxdist = 1000) %>%
    #         unlist() %>%
    #         grd_pts[., ]
    # mapview(near, col.regions = "green") +
    #   mapview(filter(pairs, type == "fill"), col.regions = "blue", layer.name = "fill") +
    #   mapview(lines, zcol = "dist") +
    #   mapview(filter(pairs, type == "gap"), col.regions = "red", layer.name = "gap")

    gaps <- gaps[matches, .(date, grd_200m_id, fill_id), on = c(grd_200m_id = "gap_id")] %>%
            setkeyv(c("date", "grd_200m_id", "fill_id"))
    fill <- mod5[gaps, .(fill_val = get(tvar)), on = c("date", grd_200m_id = "fill_id")]
    gaps[, fill_val := fill]
    mod5[gaps, (tvar) := fill_val, on = .(date, grd_200m_id)]
    rm(gaps, matches, fill, gap_pts, fill_pts, nn)

  }
  rm(tvar)

  # Save m4 predictions
  glue("{year}: saving m4 predictions") %>% report
  tmp_path <- glue(pred_path, ".tmp")
  mod5[, .(date, grd_200m_id, x, y, pred_res_tmin_ens, pred_res_tmean_ens, pred_res_tmax_ens)] %>%
    write_fst(tmp_path, compress = 100)
  file.rename(tmp_path, pred_path)
  glue("{year}: -> {pred_path}") %>% report

  # Save final 200m predictions
  glue("{year}: saving final 200 m predictions") %>% report
  final_path <- sub("m4.fst", "200m.fst", pred_path)
  tmp_path <- glue(final_path, ".tmp")
  mod5[, .(date, grd_200m_id, x, y, tmin, tmean, tmax)] %>%
    write_fst(tmp_path, compress = 100)
  file.rename(tmp_path, final_path)
  glue("{year}: -> {final_path}") %>% report

  # Cleanup
  rm(mod5, tmp_path, final_path)

}

rm(pred_path, year)

report(glue("Done {paste(range(constants$model_years), collapse = '-')}"))
