# Clean the Meteo France station data -----------------------------------------

source("init.R")

report("---")
report("Cleaning Meteo France station data")


# Skip if all output files exist
out_paths <- paste0("meteo_data_", constants$model_years, "-clean.fst") %>%
             file.path(constants$work_dir, "meteo_obs", "clean", .)
if (all(file.exists(out_paths)) && !OVERWRITE) {
  rm(out_paths)
  stop("Skip existing")
}


# Setup -----------------------------------------------------------------------

# An empty vector to hold a description of each cleaning step
step_descriptions <- character()

# Number of cores to use for parallel computation
ncores <- (length(constants$model_years) + 1) %>% min(., get_ncores())

# The directories for raw and processed station data
meta_dir <- file.path(constants$data_dir, "meteo", "meteo_france", "metadata")
processed_dir <- file.path(constants$work_dir, "meteo_obs")

# Functions -------------------------------------------------------------------

# Function to calculate and report total observations and number just removed
tally_and_report <- function(old_tally = 0) {
  new_tally <- nrow(obs)
  msg <- paste(" ", new_tally, "observations")
  if (new_tally < old_tally) {
    msg <- paste0(msg, " (", old_tally - new_tally, " removed)")
  }
  report(msg)
  new_tally
}

# Function to calculate and save the number of observations per station location
save_station_stats <- function(dt, out_dir, filename) {
  report("Calculating number of observations per station")

  obs_counts <- dt[, .N, by = .(insee_id, latitude, longitude)]
  min_counts <- dt[!is.na(temperature_min), .N, by = .(insee_id, latitude, longitude)]
  mean_counts <- dt[!is.na(temperature_mean), .N, by = .(insee_id, latitude, longitude)]
  max_counts <- dt[!is.na(temperature_max), .N, by = .(insee_id, latitude, longitude)]
  obs_counts %>% setnames(old = "N", new = "n_obs")
  obs_counts[min_counts, tmin_obs := i.N, on = c("insee_id", "latitude", "longitude")]
  obs_counts[mean_counts, tmean_obs := i.N, on = c("insee_id", "latitude", "longitude")]
  obs_counts[max_counts, tmax_obs := i.N, on = c("insee_id", "latitude", "longitude")]

  # Add the number of locations per station
  location_counts <- obs_counts[, .N, by = .(insee_id)]
  obs_counts[location_counts, n_loc := i.N, on = "insee_id"]
  rm(location_counts)

  # Add the station type
  if ("station_type" %in% names(dt)) {
    types <- dt[, .N, by = c("insee_id", "station_type")]
    obs_counts[, type := types[obs_counts, station_type, on = "insee_id"]]
  }

  # Save as a csv
  path <- file.path(out_dir, filename)
  paste("  Saving to", path) %>% report
  fwrite(obs_counts[order(insee_id), ], path)
}

# Function to split data by year and save
save_by_year <- function(df, out_dir, filename_suffix) {
  paste("Saving", filename_suffix, "observations to", out_dir) %>% report

  # Split by year and save in parallel
  split(df, df$year) %>%
    mclapply(function(splt, filename_suffix, out_dir) {
      splt$year[1] %>%
        paste0("meteo_data_", ., "-", filename_suffix, ".fst") %>%
        file.path(out_dir, .) %>%
        write_fst(splt, ., compress = 100)
    }, filename_suffix = filename_suffix, out_dir = out_dir, mc.cores = ncores)

  report("  OK")
}

# Function to write a summary of the cleaning steps to a readme file
write_readme <- function(step_descriptions, out_dir) {
  paste("Writing readme in", out_dir) %>% report
  c(
    "Cleaned by removing observations that are:",
    paste("*", step_descriptions)
  ) %>%
  write(file.path(clean_dir, "readme.txt"))
}


# Load observations for all years ---------------------------------------------



paste(
  "Loading parsed observations for", first(constants$model_years), "to", last(constants$model_years)
) %>% report
obs <- mclapply(constants$model_years, function(year) {
  paste0("meteo_data_", year, "-parsed.fst") %>%
    file.path(processed_dir, "parsed", .) %>%
    read_fst(as.data.table = TRUE)
}, mc.cores = ncores) %>% rbindlist(fill = TRUE)

# Round coordinates to 5 decimal places
# This improves grouping of rows by location: without rounding seemingly identical coordinates
# (e.g. 46.54333) may be unequal due to floating point issues. Prior to 2017, Meteo France reported
# coordinates to the nearest second (0.0002778 degrees, about 31 m). Post 2017 they report to the
# nearest tenth of a decimal minute (0.001667 degrees, about 185 m). Thus rounding to to 5 decimal
# places (~ 1 m) does not increase imprecision.
obs[, latitude := round(latitude, 5)]
obs[, longitude := round(longitude, 5)]

# Set keys
obs %>% setkeyv(c("date", "insee_id", "latitude", "longitude"))

obs_tally <- tally_and_report()
# 17553444 observations


# Calculate and save the number of observations per station location ----------

save_station_stats(obs, processed_dir, "stations_before_cleaning.csv")


# Add year, month, and day of year columns to facilitate filtering ------------

report("Adding year, month, and day of year columns for filtering")
obs[, year := year(date)]
obs[, month := month(date)]
obs[, doy := yday(date)]
# 50 seconds


# Add station type from Meteo France station metadata -------------------------

report("Adding current station type")

# Load the metadata file layout
layout <- file.path(meta_dir, "station_metadata_file_layout.csv") %>% read.csv(as.is = TRUE)

# Read station type from the metadata and add it to the observations
metadata <-
  file.path(meta_dir, "station_metadata.csv") %>%
  fread(col.names = layout$colname, colClasses = layout$coltype)
metadata[, type := factor(type)]
obs[metadata, station_type := i.type, on = "insee_id"]
rm(layout, metadata)

# # ALTERNATIVE
# # Meteo France makes it very difficult to export the historical type of a station. This code
# # adds the current station type to any station that was open for an entire calendar year from 2000
# # to 2017 and had a station type of 0, 1, or 2 at some point during that year.
# metadata <-
#   file.path(meta_dir, "station_types_by_year", "stations_0-2_2000-2017.csv") %>%
#   fread %>%
#   setnames(c("insee_id", "name", "elevation", "type", "status", "lat", "lon", "year"))
# obs[metadata, station_type := i.type, on = c("insee_id", "year")]
# rm(metadata)


# Add elevation from EU-DEM v1.1 based on observation locations ---------------

report("Adding elevation")

# Extract unique observation locations and transform to SpatialPoints
obs_pts <-
  obs[, .(longitude, latitude)] %>%
  unique %>%
  SpatialPoints(proj4string = CRS(constants$epsg_4326))

# Load the Copernicus EU-DEM v1.1 elevation data
elevation <-
  file.path(constants$work_dir, "elevation", "eu-dem_v1-1", "eu_dem_v11_france.tif") %>%
  raster

# Extract elevation for each observation location
obs_pts$elevation <-
  spTransform(obs_pts, projection(elevation)) %>% # project to match elevation data
  raster::extract(elevation, .)                   # 40 seconds; velox is much slower

# Add elevation to the observations data frame
obs[as.data.table(obs_pts), station_elevation := i.elevation, on = c("latitude", "longitude")]
rm(elevation)


# Create an empty data frame to hold removed observations ---------------------

removed <- obs[0, ]
removed[, problem := ""] # add column to hold details of why station was removed


# Remove observations with no temperature data --------------------------------

description <- "missing temperature data"
paste("Removing observations", description) %>% report

# Remember observations with no temperature data
removed <-
  obs[is.na(temperature_min) & is.na(temperature_max) & is.na(temperature_mean), ] %>%
  .[, problem := description] %>%
  rbind(removed, .)

# Remove observations with no temperature data
obs <- obs[!(is.na(temperature_min) & is.na(temperature_max) & is.na(temperature_mean)), ]
step_descriptions %<>% c(description)

obs_tally <- tally_and_report(obs_tally)
#   924694 removed
# 16628750 observations

rm(description)


# Remove observations outside the study area ----------------------------------

description <- "outside study area"
paste("Removing observations", description) %>% report

# Flag locations as inside or outside study area (including a 1 km buffer)
study_area <-
  file.path(constants$work_dir, "borders", "france_buf1km_epsg-4326.shp") %>%
  shapefile
obs_pts$outside <- obs_pts %>% over(study_area) %>% is.na %>% as.logical
paste("  Found", length(obs_pts[obs_pts$outside, ]), "locations outside study area") %>% report
# 196 locations outside

# Add the flag to the observations
obs[as.data.table(obs_pts), outside := i.outside, on = c("latitude", "longitude")]

# Remember observations outside study area
removed <-
  obs[(outside), ] %>%
  .[, problem := description] %>%
  .[, outside := NULL] %>%
  rbind(removed, .)

# Remove observations outside the study area and the flag column
obs <- obs[!(outside), ]
obs[, outside := NULL]
step_descriptions %<>% c(description)

obs_tally <- tally_and_report(obs_tally)
#   376315 removed
# 16252435 observations

rm(obs_pts, study_area, description)


# Remove observations from unknown stations -----------------------------------

description <- "from unknown station"
paste("Removing observations", description) %>% report

# Remember observations with no known station type
removed <-
  obs[station_type %>% as.character %>% is.na, ] %>%
  .[, problem := description] %>%
  rbind(removed, .)

# Remove observations with no known station type
obs <- obs[station_type %>% as.character %>% is.na %>% not, ]
step_descriptions %<>% c(description)

obs_tally <- tally_and_report(obs_tally)
#  1431039 removed
# 14821396 observations

rm(description)


# Remove low-quality observations from co-located stations --------------------

description <- "co-located with a higher-quality station"
paste("Removing observations", description) %>% report

# Select observations from stations that have the same location as another station
colocated <-
  obs[, .N, by = c("date", "latitude", "longitude")] %>%
  .[N > 1, ] %>%
  obs[., on = c("date", "latitude", "longitude")]

# For each location and date, identify the lowest station type (should be highest quality)
colocated[
  ,
  best_station_type := levels(.SD$station_type)[min(as.integer(.SD$station_type))],
  by = c("date", "latitude", "longitude")
]

# Mark all observations from lower-quality co-located stations
obs[
  colocated[station_type != best_station_type, c("date", "insee_id"), with = FALSE],
  low_quality := TRUE,
  on = c("date", "insee_id")
]

# Remember observations from lower-quality co-located stations
removed <-
  obs[(low_quality), ] %>%
  .[, problem := description] %>%
  .[, low_quality := NULL] %>%
  rbind(removed, .)

# Remove observations from lower-quality co-located stations and the flag column
obs <- obs[(is.na(obs$low_quality)), ]
obs[, low_quality := NULL]
step_descriptions %<>% c(description)

obs_tally <- tally_and_report(obs_tally)
#    46009 removed
# 14775387 observations


# Remove inconsistent observations from co-located stations -------------------

description <- "from co-located stations with temperature discrepancy > 2 degrees"
paste("Removing observations", description) %>% report

# Select colocated observations that have the same station type
colocated <-
  colocated[station_type == best_station_type, ] %>%
  .[, .N, by = c("date", "latitude", "longitude")] %>%
  .[N > 1] %>%
  obs[., on = c("date", "latitude", "longitude")]

# Get the date and insee id of co-located observations whose temperatures differ by > 2 degrees
bad <-
  colocated[
    ,
    lapply(.SD, function(x) { max(x, na.rm = TRUE) - min(x, na.rm = TRUE) }),
    by = c("date", "latitude", "longitude"),
    .SDcols = c("temperature_min", "temperature_max")
  ] %>%
  .[temperature_min > 2 | temperature_max > 2, ] %>%
  colocated[., on = c("date", "latitude", "longitude")] %>%
  .[, c("date", "insee_id"), with = FALSE]

# Mark the observations from the bad station dates
obs[bad, tdiff := TRUE, on = c("date", "insee_id")]

# Remember the bad observations
removed <-
  obs[(tdiff), ] %>%
  .[, problem := description] %>%
  .[, tdiff := NULL] %>%
  rbind(removed, .)

# Remove the bad observations and flag column
obs <- obs[(is.na(tdiff)), ]
obs[, tdiff := NULL]
step_descriptions %<>% c(description)

obs_tally <- tally_and_report(obs_tally)
#     1502 removed
# 14773885 observations

# Cleanup
rm(bad, colocated, description)


# Remove observations from stations with < 21 observations in a month ---------

description <- "from station with < 21 observations this month"
paste("Removing observations", description) %>% report

# Calculate the number of observations per station for each month
obs[, .N, by = c("year", "month", "insee_id")] %>%
  obs[., station_obs_this_month := i.N, on = c("year", "month", "insee_id")]

# Remember observations from stations with < 21 observations in a month
removed <-
  obs[station_obs_this_month < 21, ] %>%
  .[, problem := description] %>%
  .[, station_obs_this_month := NULL] %>%
  rbind(removed, .)

# Remove observations from stations with < 21 observations in a month
obs <- obs[station_obs_this_month >= 21, ]
obs[, station_obs_this_month := NULL]
step_descriptions %<>% c(description)

obs_tally <- tally_and_report(obs_tally)
#    33077 removed
# 14740808 observations

# Cleanup
rm(description)


# Calculate and save the number of observations per station location ----------

save_station_stats(obs, processed_dir, "stations_after_cleaning.csv")


# Save the removed and the clean data -----------------------------------------

# Save removed data
file.path(processed_dir, "removed") %T>%
  dir.create(recursive = TRUE, showWarnings = FALSE) %>%
  save_by_year(removed, ., "removed")

# Save clean data
clean_dir <- file.path(processed_dir, "clean")
dir.create(clean_dir, recursive = TRUE, showWarnings = FALSE)
save_by_year(obs, clean_dir, "clean")

# Save a readme describing how the data was cleaned
write_readme(step_descriptions, clean_dir)

# Save a shapefile of station locations
obs_pts <-  obs[, .(
                    insee_id = first(insee_id),
                    type = first(station_type),
                    n_obs = length(date)
                   ), by = c("latitude", "longitude")]
obs_pts <- SpatialPointsDataFrame(
             coords = obs_pts[, .(longitude, latitude)],
             data = obs_pts[, .(insee_id, type, n_obs)],
             proj4string = CRS(constants$epsg_4326)
           )
report("Saving shapefile of station locations")
path <- "1.work/meteo_obs/clean/station_locations.shp"
shapefile(obs_pts, path, overwrite = TRUE)
paste("->", path) %>% report


report("Done")
