# Parse Meteo France station data

source("init.R")

report("---")
report("Parsing Meteo France station data")


# To parse latitude or longitude stored as degrees, minutes, and seconds without leading zeros
# e.g. -01°02'12" is stored as -10212
ddmmss_to_decimal <- function(ddmmss) {

  # Extract the sign and left pad with zeros
  mult   <- sign(ddmmss)             # -10212 -> "-010212" | 10212 -> " 010212"
  ddmmss <- sprintf("% 07d", ddmmss) # -10212 -> "-010212" | 10212 -> " 010212"

  # Extract degrees, minutes, and seconds     # "-010212"       | " 010212"
  degs <- substr(ddmmss, 2, 3) %>% as.integer # "-01"     ->  1 | " 01"     ->  1
  mins <- substr(ddmmss, 4, 5) %>% as.integer #    "02"   ->  2 |    "02"   ->  2
  secs <- substr(ddmmss, 6, 7) %>% as.integer #      "12" -> 12 |      "12" -> 12

  # Return decimal degrees with proper sign
  mult * (degs + (mins / 60) + (secs / 3600))

}

# To parse latitude or longitude stored as degrees and decimal minutes without leading zeros
# e.g. -01°02'12" is stored as -1022
ddmmm_to_decimal <- function(ddmmm) {

  # Extract the sign and left pad with zeros
  mult <- sign(ddmmm)              # -1022 -> "-01022" | 1022 -> " 01022"
  ddmmm <- sprintf("% 06d", ddmmm) # -1022 -> "-01022" | 1022 -> " 01022"

  # Extract degrees and decimal minutes           # "-01022"        | " 01022"
  degs <- substr(ddmmm, 2, 3) %>% as.integer      # "-01"    -> 1   | " 01"    -> 1
  mins <- substr(ddmmm, 4, 6) %>% as.integer / 10 #    "022" -> 2.2 |    "022" -> 2.2

  # Return decimal degrees with proper sign
  mult * (degs + (mins / 60))

}

# Helper to display decimal degrees as degrees minutes seconds
# e.g. -1.03667 -> -1°02'12"
decimal_to_ddmmss <- function(x) {

  s <- sign(x)
  x <- abs(x)
  dd <- trunc(x)
  mm <- trunc((x * 60) %% 60)
  ss <- trunc((x * 3600) %% 60)

  paste(s * dd, "°", sprintf("%02d", mm), "'", sprintf("%02d", ss), '"', sep = "")

}

# To convert latitude or longitude from the specified format to decimal degrees
parse_coord <- function(coord, fmt) {

  if (fmt == "ddmmss") {

    ddmmss_to_decimal(coord)

  } else if (fmt == "ddmm.m") {

    ddmmm_to_decimal(coord)

  } else {

    stop(glue("Unrecognized coordinate format: {fmt}"))

  }

}

# To parse meteo france observation data
parse_meteo_france_obs <- function(obs_path, meta) {

  # Load the observations
  layout <- rbindlist(meta$columns)
  dat <- fread(obs_path, col.names = layout$name, colClasses = layout$type, blank.lines.skip = TRUE)

  # Drop unneedded columns
  dat <- dat[, !"ignore"]

  # Parse latitude and longitude to decimal degrees
  dat[, latitude := parse_coord(latitude, meta$coord_format)]
  dat[, longitude := parse_coord(longitude, meta$coord_format)]

  # Parse the date
  dat[, date := dmy(date)]

  # Return the parsed data
  dat

}

# Create a directory to hold the parsed data
parsed_dir <- "1.work/meteo_obs/parsed"
dir.create(parsed_dir, recursive = TRUE, showWarnings = FALSE)

# List the directories containing raw observation data (data files are grouped by region)
region_dirs <- glue("{constants$data_dir}/meteo/meteo_france/raw") %>%
               list.files(pattern = "region\\d", full.names = TRUE)

# Parse and briefly check observations for all years
for (year in constants$model_years) {

  # Check whether the year has already been processed
  out_path <- glue("{parsed_dir}/meteo_data_{year}-parsed.fst")
  if (file.exists(out_path) && !OVERWRITE) {

    next

  }

  paste("Processing", year) %>% report

  # Load metadata corresponding to the year
  if (year < 2017) {
    meta <- read_yaml("2.scripts/data_management/meteo_france/obs_meta_pre2017.yml")
  } else {
    meta <- read_yaml("2.scripts/data_management/meteo_france/obs_meta_post2017.yml")
  }
  layout <- rbindlist(meta$columns)

  # Load and parse all data files for the year
  report("  Loading observations")
  obs <- glue("^(donnees_)?region\\d(_|Q){year}(0101_{year}1231.txt)?$") %>%
         list.files(region_dirs, pattern = ., ignore.case = TRUE, full.names = TRUE) %>%
         mclapply(parse_meteo_france_obs, meta = meta, mc.cores = ncores) %>%
         rbindlist

  # Ensure observations were loaded
  report("  Checking observations were parsed")
  stopifnot(nrow(obs) > 0)

  # Ensure that all observations fall within the data year
  report("  Checking for out-of-period observations")
  stopifnot(paste0(year, "-01-01") <= min(obs$date))
  stopifnot(paste0(year, "-12-31") >= max(obs$date))

  # Ensure that each station has only one set of observations per date
  report("  Checking for duplicate observations")
  dups <- obs[, .N, by = c("insee_id", "date")] %>%
          .[N > 1, .N]
  stopifnot(dups == 0)

  # Save the parsed data
  report("  Saving parsed observations")
  write_fst(obs, out_path, compress = 100)
  report(glue("  -> {out_path}"))

  # Cleanup
  rm(obs)

}

# Cleanup
rm(ddmmm_to_decimal, ddmmss_to_decimal, parse_coord, parse_meteo_france_obs, out_path, parsed_dir,
   region_dirs, year)

report("Done")
