library(magrittr)
library(fst)
library(sp)
library(raster)
library(rgeos)

# Set directories and load helper functions
file.path("~", "temperature-france", "helpers", "set_dirs.R") %>% source
file.path(helpers_dir, "report.R") %>% source

# Set the figures dir
figures_dir <- file.path(model_dir, "figures")

# Load France borders
france_2154 <- file.path(data_dir, "ign", "borders", "france_epsg-2154.shp") %>% shapefile
france_3035 <- file.path(data_dir, "ign", "borders", "france_epsg-3035.shp") %>% shapefile
france_4326 <- file.path(data_dir, "ign", "borders", "france_epsg-4326.shp") %>% shapefile

#########################################################
# Export gridded data as shapefiles / rasters for mapping
#########################################################
{
  # Load the MODIS 1 km grid
  grd <- file.path(grids_dir, "modis_grid.rds") %>% readRDS

  # Weather stations
  {
    # Load a list of all stations
    stations <-
      file.path(data_dir, "meteo_france", "observations", "station_list.csv") %>%
      read.csv

    # Convert to a SpatialPointsDataFrame
    coordinates(stations) <- ~ lng + lat
    proj4string(stations) <- proj4string(france_4326)

    # Crop to remove stations in Corsica and on buoys in Atlantic
    stations <- crop(stations, france_4326)

    # Save as a shapefile
    file.path(figures_dir, "meteo_stations.shp") %>% shapefile(stations, ., overwrite = TRUE)
    rm(stations)
  }

  # MODIS LST
  {
    # Load data for 2003
    aqua_lst <-
      file.path(extracts_dir, "modis", "2003", "modis_grid_aqua_lst_2003-06.fst") %>%
      read.fst

    terra_lst <-
      file.path(extracts_dir, "modis", "2003", "modis_grid_terra_lst_2003-06.fst") %>%
      read.fst

    # Extract data for June 21
    aqua_lst <- aqua_lst[aqua_lst$date == "2003-06-21", c("day_lst", "night_lst")]
    terra_lst <- terra_lst[terra_lst$date == "2003-06-21", c("day_lst", "night_lst")]

    # Combine into a single data frame
    names(aqua_lst) <- names(aqua_lst) %>% paste0("aqua_", .)
    names(terra_lst) <- names(terra_lst) %>% paste0("terra_", .)
    lst <- cbind(aqua_lst, terra_lst)
    rm(aqua_lst, terra_lst)

    # Transform to SpatialPointsDataFrame using lat lng from reference grid
    coordinates(lst) <- coordinates(grd)
    projection(lst) <- projection(grd)

    # Project to MODIS sinusoidal and convert to a SpatialPixelsDataFrame
    lst <- spTransform(lst, "+proj=sinu +lon_0=0 +x_0=0 +y_0=0 +a=6371007.181 +b=6371007.181 +units=m +no_defs")
    gridded(lst) <- TRUE

    # Convert to a raster and project back to EPSG:2154
    lst <- brick(lst)
    lst <- projectRaster(lst, crs = projection(france_2154))

    # Save as GeoTIFFs
    file.path(figures_dir, "modis_aqua_day_lst_june_21_2003.tif") %>% save_geotiff(lst$aqua_day_lst, ., overwrite = TRUE)
    file.path(figures_dir, "modis_aqua_night_lst_june_21_2003.tif") %>% save_geotiff(lst$aqua_night_lst, ., overwrite = TRUE)
    file.path(figures_dir, "modis_terra_day_lst_june_21_2003.tif") %>% save_geotiff(lst$terra_day_lst, ., overwrite = TRUE)
    file.path(figures_dir, "modis_terra_night_lst_june_21_2003.tif") %>% save_geotiff(lst$terra_night_lst, ., overwrite = TRUE)
    rm(lst)
  }

  # MODIS NDVI
  {
    # Load data for 2003
    aqua_ndvi <-
      file.path(extracts_dir, "modis", "2003", "modis_grid_aqua_ndvi_2003-06.rds") %>%
      readRDS

    terra_ndvi <-
      file.path(extracts_dir, "modis", "2003", "modis_grid_terra_ndvi_2003-06.rds") %>%
      readRDS

    # Combine into a single data frame
    ndvi <- data.frame(
      "aqua_ndvi" = aqua_ndvi$ndvi,
      "terra_ndvi" = terra_ndvi$ndvi
    )
    rm(aqua_ndvi, terra_ndvi)

    # Transform to SpatialPointsDataFrame using lat lng from reference grid
    coordinates(ndvi) <- coordinates(grd)
    projection(ndvi) <- projection(grd)

    # Project to MODIS sinusoidal and convert to a SpatialPixelsDataFrame
    ndvi <- spTransform(ndvi, "+proj=sinu +lon_0=0 +x_0=0 +y_0=0 +a=6371007.181 +b=6371007.181 +units=m +no_defs")
    gridded(ndvi) <- TRUE

    # Convert to a raster and project back to EPSG:2154
    ndvi <- brick(ndvi)
    ndvi <- projectRaster(ndvi, crs = projection(france_2154))

    # Save as GeoTIFFs
    file.path(figures_dir, "modis_aqua_ndvi_june_2003.tif") %>% save_geotiff(ndvi$aqua_ndvi, ., overwrite = TRUE)
    file.path(figures_dir, "modis_terra_ndvi_june_2003.tif") %>% save_geotiff(ndvi$terra_ndvi, ., overwrite = TRUE)
    rm(ndvi)
  }

  # SIM
  {
    # Load data for 2003
    sim <-
      file.path(extracts_dir, "sim", "2003", "modis_grid_sim_2003-06.fst") %>%
      read.fst

    # Extract data for June 21
    sim <- data.frame("temperature_max" = sim$temperature_max[sim$date == "2003-06-21"])

    # Transform to SpatialPointsDataFrame using lat lng from reference grid
    coordinates(sim) <- coordinates(grd)
    projection(sim) <- projection(grd)

    # Project to MODIS sinusoidal and convert to a SpatialPixelsDataFrame
    sim <- spTransform(sim, "+proj=sinu +lon_0=0 +x_0=0 +y_0=0 +a=6371007.181 +b=6371007.181 +units=m +no_defs")
    gridded(sim) <- TRUE

    # Convert to a raster and project back to EPSG:2154
    sim <- raster(sim)
    sim <- projectRaster(sim, crs = projection(france_2154))

    # Save as a GeoTIFF
    file.path(figures_dir, "sim_max_temp_june_21_2003.tif") %>% save_geotiff(sim, ., overwrite = TRUE)
    rm(sim)
  }

  # Elevation
  {
    # Load data
    elevation <- file.path(extracts_dir, "modis_1km_eu_dem_v11.rds") %>% readRDS

    # Drop grid ids
    elevation <- elevation["elevation"]

    # Transform to SpatialPointsDataFrame using lat lng from reference grid
    coordinates(elevation) <- coordinates(grd)
    projection(elevation) <- projection(grd)

    # Project to MODIS sinusoidal and convert to a SpatialPixelsDataFrame
    elevation <- spTransform(elevation, "+proj=sinu +lon_0=0 +x_0=0 +y_0=0 +a=6371007.181 +b=6371007.181 +units=m +no_defs")
    gridded(elevation) <- TRUE

    # Convert to a raster and project back to EPSG:2154
    elevation <- raster(elevation)
    elevation <- projectRaster(elevation, crs = projection(france_2154))

    # Save as a GeoTIFF
    file.path(figures_dir, "eu_dem_elevation.tif") %>% save_geotiff(elevation, ., overwrite = TRUE)
    rm(elevation)
  }

  # Land Cover
  {
    # Load data
    clc <- file.path(extracts_dir, "modis_1km_clc_2012.rds") %>% readRDS

    # Merge some land cover groupings
    clc_merged <- data.frame(
      "artificial" = (
        clc$urban_continuous +
        clc$urban_discontinuous +
        clc$commercial_industrial +
        clc$mines_dumps +
        clc$urban_green_leisure
      ),
      "agriculture" = (
        clc$arable_land +
        clc$permanent_crops +
        clc$mixed_agricultural
      ),
      "forest_natural" = (
        clc$forest +
        clc$grassland_pasture +
        clc$shurbland
      ),
      "bare_snow" = (
        clc$bare +
        clc$snow
      )
    )
    clc_merged$wetland <- clc$wetland
    clc_merged$water <- clc$water
    clc <- clc_merged
    rm(clc_merged)

    # For each row, take the column index of the land grouping with the largest area
    clc <- apply(clc, 1, function(x) {
      which(x == max(x))[1]
    }) %>% unlist %>% as.data.frame

    # Transform to SpatialPointsDataFrame using lat lng from reference grid
    coordinates(clc) <- coordinates(grd)
    projection(clc) <- projection(grd)

    # Project to MODIS sinusoidal and convert to a SpatialPixelsDataFrame
    clc <- spTransform(clc, "+proj=sinu +lon_0=0 +x_0=0 +y_0=0 +a=6371007.181 +b=6371007.181 +units=m +no_defs")
    gridded(clc) <- TRUE

    # Convert to a raster and project back to EPSG:2154
    clc <- raster(clc)
    clc <- projectRaster(clc, crs = projection(france_2154), method = "ngb")

    # Save as a GeoTIFF
    file.path(figures_dir, "clc_2012_groupings.tif") %>% save_geotiff(clc, ., overwrite = TRUE)
    rm(clc)
  }

  # Population
  {
    # Load data
    pop <- file.path(extracts_dir, "modis_1km_population.rds") %>% readRDS

    # Drop grid ids
    pop <- pop["population"]

    # Transform to SpatialPointsDataFrame using lat lng from reference grid
    coordinates(pop) <- coordinates(grd)
    projection(pop) <- projection(grd)

    # Project to MODIS sinusoidal and convert to a SpatialPixelsDataFrame
    pop <- spTransform(pop, "+proj=sinu +lon_0=0 +x_0=0 +y_0=0 +a=6371007.181 +b=6371007.181 +units=m +no_defs")
    gridded(pop) <- TRUE

    # Convert to a raster and project back to EPSG:2154
    pop <- raster(pop)
    pop <- projectRaster(pop, crs = projection(france_2154), method = "ngb")

    # Save as a GeoTIFF
    file.path(figures_dir, "insee_population.tif") %>% save_geotiff(pop, ., overwrite = TRUE)
    rm(pop)
  }

  rm(grd)
}

#########################################################
# Illustrate grid creation / aggregation / disaggregation
#########################################################
{
  # Create grid illustrations
  {
    # Take 3 x 3 crop from MODIS sinusoidal grid
    sinu <- file.path(grids_dir, "modis_grid_sinusoidal.tif") %>% raster
    sinu <- extent(sinu, 524, 526, 536, 538) %>% crop(sinu, .)
    values(sinu) <- c(
      NA, 1, NA,
      NA, 1, 1,
      1, 1, 1
    )

    # Convert to SpatialGrid, SpatialPolygons, and SpatialPoints
    grd_sinu <- as(sinu, "SpatialGrid")
    pixels_sinu <- as(sinu, "SpatialPixels")
    pts_sinu <- as(sinu, "SpatialPoints")

    # Project points to EPSG:2154
    pts_2154 <- spTransform(pts_sinu, proj4string(france_2154))

    # Project points to EPSG:3035, buffer 500 m square, and project the buffers to EPSG:2154
    pts_3035 <- spTransform(pts_sinu, proj4string(france_3035))
    buffers_3035 <- gBuffer(pts_3035, width = 500, capStyle = "SQUARE", byid = TRUE)
    buffers_2154 <- spTransform(buffers_3035, proj4string(france_2154))
  }

  # Create data illustrations
  {
    # Create a 1 km square in EPSG:2154
    square <- gBuffer(pts_2154[5, ], width = 500, capStyle = "SQUARE")

    # Fake elevation data
    elev <-
      spTransform(square, proj4string(france_3035)) %>%
      extent %>%
      extend(., 40) %>%
      raster(., res = c(25, 25), crs = proj4string(france_3035))
    values(elev) <- ncell(elev) %>% runif(., min = 1, max = 3) %>% round
    elev <- as(elev, "SpatialPixelsDataFrame")

    # Fake land cover data
    clc <-
      spTransform(square, proj4string(france_3035)) %>%
      extent %>%
      extend(., 40) %>%
      raster(., res = c(100, 100), crs = proj4string(france_3035))
    values(clc) <- ncell(clc) %>% runif(., min = 1, max = 3) %>% round
    clc <- as(clc, "SpatialPixelsDataFrame")

    # Fake population data
    pop <-
      spTransform(square, proj4string(france_3035)) %>%
      extent %>%
      extend(., 40) %>%
      raster(., res = c(200, 200), crs = proj4string(france_3035))
    values(pop) <- ncell(pop) %>% runif(., min = 1, max = 3) %>% round
    pop <- as(pop, "SpatialPixelsDataFrame")

    # Fake lst data
    lst <-
      spTransform(square, "+proj=sinu +lon_0=0 +x_0=0 +y_0=0 +a=6371007.181 +b=6371007.181 +units=m +no_defs") %>%
      extent %>%
      extend(., 1000) %>%
      raster(., res = c(1000, 1000), crs = "+proj=sinu +lon_0=0 +x_0=0 +y_0=0 +a=6371007.181 +b=6371007.181 +units=m +no_defs")
    values(lst) <- ncell(lst) %>% runif(., min = 1, max = 3) %>% round
    lst <- as(lst, "SpatialPixelsDataFrame")

    # Fake SIM data
    sim <-
      spTransform(square, "+proj=lcc +nadgrids=ntf_r93.gsb,null +lat_1=46.8 +lat_0=46.8 +lon_0=0 +k_0=0.99987742 +x_0=600000 +y_0=2200000 +a=6378249.2 +b=6356515 +towgs84=-168,-60,320,0,0,0,0 +pm=paris +units=m +no_defs") %>%
      extent %>%
      extend(., 3500) %>%
      raster(., res = c(8000, 8000), crs = "+proj=lcc +nadgrids=ntf_r93.gsb,null +lat_1=46.8 +lat_0=46.8 +lon_0=0 +k_0=0.99987742 +x_0=600000 +y_0=2200000 +a=6378249.2 +b=6356515 +towgs84=-168,-60,320,0,0,0,0 +pm=paris +units=m +no_defs")
    values(sim) <- 1
    sim <- as(sim, "SpatialGridDataFrame")

    # Fake disaggregated data
    disagg <- raster(pop) %>% disaggregate(., 4) %>% as(., "SpatialPixelsDataFrame")
  }

  # Plot to screen
  {
    # Reduce plot margins
    old_margins <- par("mar")
    par(mar = c(1, 1, 1, 1))

    # Plot the sinusoidal grid, pixels, and points
    plot(grd_sinu, lwd = 3)
    plot(pixels_sinu, lwd = 3)
    plot(pts_sinu, xlim = bbox(pixels_sinu)["x", ], ylim = bbox(pixels_sinu)["y", ], pch = 19, cex = 3)

    # Plot the EPSG:2154 points, EPSG:3035 points, EPSG:3035 buffers, and EPSG:2154 buffers
    plot(pts_2154, xlim = bbox(buffers_2154)["x", ], ylim = bbox(buffers_2154)["y", ])
    plot(pts_3035, xlim = bbox(buffers_3035)["x", ], ylim = bbox(buffers_3035)["y", ])
    plot(buffers_3035)
    plot(buffers_2154)

    # Plot dummy data
    plot(square, lwd = 4)

    plot(elev, what = "image", border = "white", col = RColorBrewer::brewer.pal(3, "Dark2"))
    plot(spTransform(square, proj4string(elev)), lwd = 4, add = T,)

    plot(clc, what = "image", border = "white", col = RColorBrewer::brewer.pal(3, "Dark2"))
    plot(spTransform(square, proj4string(clc)), lwd = 4, add = T,)

    plot(pop, what = "image", border = "white", col = RColorBrewer::brewer.pal(3, "Dark2"))
    plot(spTransform(square, proj4string(pop)), lwd = 4, add = T,)

    plot(lst, what = "image", border = "white", col = RColorBrewer::brewer.pal(3, "Dark2"))
    plot(spTransform(square, proj4string(lst)), lwd = 4, add = T,)

    plot(sim, what = "image", border = "white", col = RColorBrewer::brewer.pal(3, "Dark2"))
    plot(spTransform(square, proj4string(sim)), lwd = 4, add = T,)

    # Plot aggregation
    plot(elev, what = "image", border = "white", col = RColorBrewer::brewer.pal(3, "Dark2"))
    plot(elev, what = "image", border = "white", col = RColorBrewer::brewer.pal(3, "Dark2"))
    plot(spTransform(buffers_3035[5, ], proj4string(elev)), lwd = 4, add = T,)

    # Plot disaggregation
    plot(pop, what = "image", border = "white", col = RColorBrewer::brewer.pal(3, "Dark2"))
    plot(disagg, what = "image", border = "white", col = RColorBrewer::brewer.pal(3, "Dark2"))
    plot(disagg, what = "image", border = "white", col = RColorBrewer::brewer.pal(3, "Dark2"))
    plot(spTransform(buffers_3035[5, ], proj4string(pop)), lwd = 4, add = T,)

    # Restore the plot margins
    par(mar = old_margins)
  }

  # Plot to png files
  {
    png(file.path(figures_dir, "01_grid_sinu.png"), width = 400, height = 400)
    par(mar = c(1, 1, 1, 1))
    plot(grd_sinu, lwd = 3)
    dev.off()

    png(file.path(figures_dir, "02_grid_masked.png"), width = 400, height = 400)
    par(mar = c(1, 1, 1, 1))
    plot(pixels_sinu, lwd = 3)
    dev.off()

    png(file.path(figures_dir, "03_grid_points.png"), width = 400, height = 400)
    par(mar = c(1, 1, 1, 1))
    plot(pts_sinu, xlim = bbox(pixels_sinu)["x", ], ylim = bbox(pixels_sinu)["y", ], pch = 19, cex = 2)
    dev.off()

    png(file.path(figures_dir, "04_points_2154.png"), width = 400, height = 400)
    par(mar = c(1, 1, 1, 1))
    plot(pts_2154, xlim = bbox(buffers_2154)["x", ], ylim = bbox(buffers_2154)["y", ], pch = 19, cex = 2)
    dev.off()

    png(file.path(figures_dir, "05_points_3035.png"), width = 400, height = 400)
    par(mar = c(1, 1, 1, 1))
    plot(pts_3035, xlim = bbox(buffers_3035)["x", ], ylim = bbox(buffers_3035)["y", ], pch = 19, cex = 2)
    dev.off()

    png(file.path(figures_dir, "06_buffers_3035.png"), width = 400, height = 400)
    par(mar = c(1, 1, 1, 1))
    plot(buffers_3035, lwd = 3)
    dev.off()

    png(file.path(figures_dir, "07_buffers_2154.png"), width = 400, height = 400)
    par(mar = c(1, 1, 1, 1))
    plot(buffers_2154, lwd = 3)
    dev.off()

    png(file.path(figures_dir, "08_square_2154.png"), width = 400, height = 400)
    par(mar = c(1, 1, 1, 1))
    plot(square, lwd = 4)
    dev.off()

    png(file.path(figures_dir, "09_elev.png"), width = 400, height = 400)
    par(mar = c(1, 1, 1, 1))
    plot(elev, what = "image", border = "white", col = RColorBrewer::brewer.pal(3, "Dark2"))
    plot(spTransform(square, proj4string(elev)), lwd = 4, add = T,)
    dev.off()

    png(file.path(figures_dir, "10_clc.png"), width = 400, height = 400)
    par(mar = c(1, 1, 1, 1))
    plot(clc, what = "image", border = "white", col = RColorBrewer::brewer.pal(3, "Dark2"))
    plot(spTransform(square, proj4string(clc)), lwd = 4, add = T,)
    dev.off()

    png(file.path(figures_dir, "11_pop.png"), width = 400, height = 400)
    par(mar = c(1, 1, 1, 1))
    plot(pop, what = "image", border = "white", col = RColorBrewer::brewer.pal(3, "Dark2"))
    plot(spTransform(square, proj4string(pop)), lwd = 4, add = T,)
    dev.off()

    png(file.path(figures_dir, "12_lst.png"), width = 400, height = 400)
    par(mar = c(1, 1, 1, 1))
    plot(lst, what = "image", border = "white", col = RColorBrewer::brewer.pal(3, "Dark2"))
    plot(spTransform(square, proj4string(lst)), lwd = 4, add = T,)
    dev.off()

    png(file.path(figures_dir, "13_sim.png"), width = 400, height = 400)
    par(mar = c(1, 1, 1, 1))
    plot(sim, what = "image", border = "white", col = RColorBrewer::brewer.pal(3, "Dark2"))
    plot(spTransform(square, proj4string(sim)), lwd = 4, add = T,)
    dev.off()

    png(file.path(figures_dir, "14_high_res.png"), width = 400, height = 400)
    par(mar = c(1, 1, 1, 1))
    plot(elev, what = "image", border = "white", col = RColorBrewer::brewer.pal(3, "Dark2"))
    dev.off()

    png(file.path(figures_dir, "15_aggregation.png"), width = 400, height = 400)
    par(mar = c(1, 1, 1, 1))
    plot(elev, what = "image", border = "white", col = RColorBrewer::brewer.pal(3, "Dark2"))
    plot(spTransform(buffers_3035[5, ], proj4string(elev)), lwd = 4, add = T,)
    dev.off()

    png(file.path(figures_dir, "15_low_res.png"), width = 400, height = 400)
    par(mar = c(1, 1, 1, 1))
    plot(pop, what = "image", border = "white", col = RColorBrewer::brewer.pal(3, "Dark2"))
    dev.off()

    png(file.path(figures_dir, "16_disaggregation.png"), width = 400, height = 400)
    par(mar = c(1, 1, 1, 1))
    plot(disagg, what = "image", border = "white", col = RColorBrewer::brewer.pal(3, "Dark2"))
    dev.off()

    png(file.path(figures_dir, "17_reaggregation.png"), width = 400, height = 400)
    par(mar = c(1, 1, 1, 1))
    plot(disagg, what = "image", border = "white", col = RColorBrewer::brewer.pal(3, "Dark2"))
    plot(spTransform(buffers_3035[5, ], proj4string(elev)), lwd = 4, add = T,)
    dev.off()
  }
}

#####################
# Illustrate modeling
#####################
{
  # Make dummy data
  {
    # 3 x 3 grid
    r <- extent(0, 3, 0, 3) %>% raster(., nrows = 3, ncols = 3)
    r$id <- 1:9
    grd <- as(r, "SpatialPixelsDataFrame")
    rm(r)

    # Stations in cells 1, 3, and 7
    grd$ta <- c(
      1, 0, 1,
      0, 0, 0,
      1, 0, 0
    )
    stations <- SpatialPoints(coordinates(grd)[which(grd$ta == 1), ] + runif(6, min = -0.4, max = 0.4))

    # Ts for cells 1, 2, 3, 4, 6, 7
    grd$ts <- c(
      1, 1, 1,
      1, 0, 1,
      1, 0, 0
    )

    # Stage 1 = calibration in cells with Ts and Ta
    grd$calib <- 1 * and(grd$ts, grd$ta)
    # 1, 0, 1
    # 0, 0, 0
    # 1, 0, 0

    # Stage 2 = prediction in cells with Ts but not Ta
    grd$pred <- 1 * and(grd$ts, !grd$ta)
    # 0, 1, 0
    # 1, 0, 1
    # 0, 0, 0

    # Buffer around the center cell
    buff <- gBuffer(as(grd, "SpatialPoints")[5], width = 2, quadsegs = 16)

    # Stage 3 = filling in cells with neither Ts nor Ta
    grd$fill <- 1 * and(!grd$ts, !grd$ta)
    # 0, 0, 0
    # 0, 1, 0
    # 0, 1, 1
  }

  # Plot to screen
  {
    old_margins <- par("mar")
    par(mar = c(1, 1, 1, 1))

    # Plot Ts and stations
    plot(grd["ts"], what = "image", border = TRUE, col = c("transparent", grey(0.9)))
    plot(stations, bg = "#004080", pch = 21, cex = 2, lwd = 2, add = TRUE)

    # Plot calibration and stations
    plot(grd[grd$calib == 1, "calib"], what = "image", border = TRUE, col = "#008040")
    plot(stations, bg = "#004080", pch = 21, cex = 2, lwd = 2, add = TRUE)

    # Plot calibration, prediction, and stations
    plot(grd[grd$calib == 1, "calib"], what = "image", border = T, col = "#008040")
    plot(grd[grd$pred == 1, "pred"], what = "image", border = T, col = "#00C080", add = TRUE)
    plot(stations, bg = "#004080", pch = 21, cex = 2, lwd = 2, add = TRUE)

    # Plot calibration, prediction, stations, center cell borders, and center cell buffer
    plot(grd[grd$calib == 1, "calib"], what = "image", border = T, col = "#008040", xlim = bbox(buff)["x", ], ylim = bbox(buff)["y", ])
    plot(grd[grd$pred == 1, "pred"], what = "image", border = T, col = "#00C080", add = TRUE)

    plot(buff, add = T)
    plot(stations, bg = "#004080", pch = 21, cex = 2, lwd = 2, add = TRUE)

    plot(geometry(grd[7, ]), lwd = 2, add = T)


    # Plot calibration, prediction, filling, and stations
    plot(grd[grd$calib == 1, "calib"], what = "image", border = T, col = "#008040")
    plot(grd[grd$pred == 1, "pred"], what = "image", border = T, col = "#00C080", add = TRUE)
    plot(grd[grd$fill == 1, "fill"], what = "image", border = T, col = "#00F0C0", add = TRUE)
    plot(stations, bg = "#004080", pch = 21, cex = 2, lwd = 2, add = TRUE)

    par(mar = old_margins)
  }
}


library(magrittr)
library(sp)

file.path("~", "temperature-france", "helpers", "set_dirs.R") %>% source
file.path(helpers_dir, "report.R") %>% source

meteo_dir <- file.path(extracts_dir, "meteo_data")

vals <- c("min", "max", "mean")

meteo_data <- lapply(2000:2016, function(year) {
  paste("Loading", year, "data") %>% report

  annual_data <-
    paste0("meteo_data_", year, ".rds") %>%
    file.path(meteo_dir, .) %>%
    readRDS %>%
    as.data.frame

  cols <- paste0("temperature_", vals)
  annual_data[, c("insee_id", "latitude", "longitude", "date", cols)]
}) %>% do.call(rbind, .)


stats <- lapply(vals, function(val) {
  col <- paste0("temperature_", val)
  data.frame(
    "lowest" = min(meteo_data[[col]], na.rm = TRUE),
    "highest" = max(meteo_data[[col]], na.rm = TRUE),
    "average" = mean(meteo_data[[col]], na.rm = TRUE),
    row.names = val
  )
}) %>% do.call(rbind, .)

extremes <- rbind(
  meteo_data[which(meteo_data$temperature_min == stats["min", "lowest"])[1], ],
  meteo_data[which(meteo_data$temperature_max == stats["max", "lowest"])[1], ],
  meteo_data[which(meteo_data$temperature_mean == stats["mean", "lowest"])[1], ],
  meteo_data[which(meteo_data$temperature_min == stats["min", "highest"])[1], ],
  meteo_data[which(meteo_data$temperature_max == stats["max", "highest"])[1], ],
  meteo_data[which(meteo_data$temperature_mean == stats["mean", "highest"])[1], ]
)
