library(magrittr)
library(glue)
library(data.table)
library(fst)
library(sp)
library(raster)
library(rasterVis)


source("2.scripts/helpers/constants.R")
source("2.scripts/helpers/report.R")
source("2.scripts/helpers/table_to_raster.R")

# Mean and SD of daily Ta across all years ----------

meteo <- list.files("1.work/meteo_obs/clean", pattern = "meteo_data_\\d{4}-clean\\.fst",
                    full.names = T) %>%
         lapply(read_fst, as.data.table = TRUE) %>%
         rbindlist

min(meteo$temperature_min, na.rm = TRUE)
max(meteo$temperature_max, na.rm = TRUE)
mean(meteo$temperature_mean, na.rm = TRUE)
sd(meteo$temperature_mean, na.rm = TRUE)


# Mean of daily predicted Ta across all years ----------

report("Aggregating final 1 km predictions")
smry_path <- "1.work/predictions/mod3_all_pred_final_smry.fst"
if (file.exists(smry_path)) {

  smry_1km <- read_fst(smry_path, as.data.table = TRUE)

} else {

  smry_cols <- c("x", "y", "tmin", "tmean", "tmax")
  smry_1km <- lapply(constants$model_years, function(yr) {

                path <- glue("1.work/predictions/mod3_{yr}_pred_final.fst")
                report(glue("  Processing {path}"))
                cols <- c("date", "grd_1km_id", "grd_1km_x", "grd_1km_y", "tmin", "tmean", "tmax")
                preds <- read_fst(path, as.data.table = TRUE, columns = cols) %>%
                         setnames(c("grd_1km_x", "grd_1km_y"), c("x", "y"))
                smry <- preds[, lapply(.SD, mean), .SDcols = smry_cols, by = "grd_1km_id"]
                smry[, year := as.integer(yr)]
                smry

  }) %>%
  rbindlist %>%
  setkey("grd_1km_id") %>%
  .[, lapply(.SD, mean), .SDcols = smry_cols, by = "grd_1km_id"]

  # Save
  report("Saving")
  write_fst(smry_1km, smry_path, compress = 100)
  report(glue("-> {smry_path}"))

}

report("Aggregating final 200 m predictions")
smry_path <- "1.work/predictions/mod5_all_pred_final_smry.fst"
if (file.exists(smry_path)) {

  smry_200m <- read_fst(smry_path, as.data.table = TRUE)

} else {

  smry_cols <- c("x", "y", "tmin", "tmean", "tmax")
  smry_200m <- lapply(constants$model_years, function(yr) {

                 path <- glue("1.work/predictions/mod5_{yr}_pred_m4.fst")
                 report(glue("  Processing {path}"))
                 preds <- read_fst(path, as.data.table = TRUE)
                 smry <- preds[, lapply(.SD, mean), .SDcols = smry_cols, by = "grd_200m_id"] %>%
                         na.omit # drop rows with no prediction
                 smry[, year := as.integer(yr)]
                 smry

               }) %>%
               rbindlist %>%
               setkey("grd_200m_id") %>%
               .[, lapply(.SD, mean), .SDcols = smry_cols, by = "grd_200m_id"]

  # Save
  report("Saving")
  write_fst(smry_200m, smry_path, compress = 100)
  report(glue("-> {smry_path}"))

}


# Convert to rasters
report("Converting to rasters")
grd_1km_pxs <- readRDS("1.work/grids/grd_1km_pxs_sinu.rds")
rst_1km <- table_to_raster(smry_1km, grd_1km_pxs)

grd_200m_pts <- readRDS("1.work/grids/grd_200m_pts.rds")
rst_200m <- table_to_raster(smry_200m, grd_200m_pts)

# Save rasters
report("Saving rasters")

tif_path <- "1.work/predictions/mod3_all_pred_final_smry.tif"
save_geotiff(rst_1km, tif_path)
meta_path <- paste0(tif_path, ".bandnames")
names(rst_1km) %>% as.list %>% fwrite(meta_path, col.names = FALSE)
paste("  ->", tif_path) %>% report

tif_path <- "1.work/predictions/mod5_all_pred_final_smry.tif"
save_geotiff(rst_200m, tif_path)
meta_path <- paste0(tif_path, ".bandnames")
names(rst_200m) %>% as.list %>% fwrite(meta_path, col.names = FALSE)
paste("  ->", tif_path) %>% report

rm(tif_path, meta_path)



# ## FIXME WIP
#
# # Map all of France
#
#
#
#
# paris_1km <- extent(120000, 220000, 5382000, 5482000)
# paris_200m <- extent(3710000, 3810000, 2850000, 2930000)
# # paris_1km <- extent(155000, 185000, 5420000, 5450000)
# # paris_200m <- extent(3745000, 3773000, 2878000, 2906000)
#
# crop_1km <- crop(rst_1km, paris_1km)
# crop_200m <- crop(rst_200m, paris_200m)
#
#
#
#
# grd_200m_rst <- brick("1.work/grids/grd_200m_inputs.tif") %>%
#                 set_names(fread("1.work/grids/grd_200m_inputs.tif.bandnames", header = FALSE)) %>%
#                 magrittr::extract2("include") %>%
#                 crop(crop_200m)
# x <- mask(crop_200m, grd_200m_rst)
#
# mapview(x$tmin, alpha.regions = 1, na.color = NA, layer.name = "200m tmin")
#
#
# crop_200m_sinu <- projection(crop_1km) %>%
#                   projectExtent(crop_200m, .) %>%
#                   projectRaster(crop_200m, .)
#
#
# mapview(crop_1km$tmin, alpha.regions = 1, na.color = NA, layer.name = "1km tmin") +
#   mapview(crop_200m_sinu$tmin, alpha.regions = 1, na.color = "white", layer.name = "200m tmin")
#
# mapview(crop_1km$tmean, alpha.regions = 1, na.color = NA, layer.name = "1km tmean") +
#   mapview(crop_200m_sinu$tmean, alpha.regions = 1, na.color = "white", layer.name = "200m tmean")
#
# mapview(crop_1km$tmax, alpha.regions = 1, na.color = NA, layer.name = "1km tmax") +
#   mapview(crop_200m_sinu$tmax, alpha.regions = 1, na.color = "white", layer.name = "200m tmax")
#
#
#
#
#
#
#
#
#
# library(sf)
# library(ggplot2)
# library(ggmap)
# library(ggspatial)
# library(leaflet)
# library(mapview)
#
# # https://stackoverflow.com/questions/47749078/how-to-put-a-geom-sf-produced-map-on-top-of-a-ggmap-produced-raster/50844502#50844502
# fix_ggmap_bbox <- function(map) {
#
#   if (!inherits(map, "ggmap")) stop("map must be a ggmap object")
#
#   # Extract the map's bounding box and convert from EPSG:4326 to EPSG:3857 (web mercator)
#   bbox_3857 <- attr(map, "bb") %>%
#                unlist %>%
#                set_names(c("ymin", "xmin", "ymax", "xmax")) %>%
#                st_bbox(crs = 4326) %>%
#                st_as_sfc %>%
#                st_transform(crs = 3857) %>%
#                st_bbox
#
#   # Overwrite the bbox of the ggmap object with the transformed coordinates
#   attr(map, "bb")$ll.lat <- bbox_3857["ymin"]
#   attr(map, "bb")$ll.lon <- bbox_3857["xmin"]
#   attr(map, "bb")$ur.lat <- bbox_3857["ymax"]
#   attr(map, "bb")$ur.lon <- bbox_3857["xmax"]
#
#   map
#
# }
#
# france <- read_sf("1.work/borders/france_epsg-2154.shp") %>%
#           st_transform(crs = 3857)
#
# stns <- fread("1.work/meteo_obs/stations_after_cleaning.csv") %>%
#         .[order(-type, n_obs), lapply(.SD, last), by = "insee_id"] %>%
#         .[type < 4, category := "Automatic"] %>%
#         .[type > 3, category := "Manual"] %>%
#         as.data.frame %>%
#         st_as_sf(coords = c("longitude", "latitude"), crs = 4326) %>%
#         st_transform(crs = 3857)
#
# basemap <- c(-5.5, 42.2, 8.5, 51.2) %>%
#            get_stamenmap(zoom = 6, maptype = "terrain-background", crop = TRUE) %>%
#            fix_ggmap_bbox %>%
#            ggmap
#
# m <- basemap +
#        geom_sf(data = france, aes(fill = "Study area"), colour = "black", stroke = 2,
#                inherit.aes = FALSE) +
#        geom_sf(data = stns[stns$category == "Manual", ], aes(colour = category),
#                inherit.aes = FALSE, show.legend = "point", size = 1) +
#        geom_sf(data = stns[stns$category == "Automatic", ], aes(colour = category),
#                inherit.aes = FALSE, show.legend = "point", size = 1) +
#        scale_colour_viridis_d() +
#        scale_fill_manual(values = NA) +
#        guides(
#          fill = guide_legend(order = 1, title = NULL, override.aes = list(shape = NA)),
#          colour = guide_legend(order = 2, title = "Weather stations")
#        ) +
#        annotation_scale(plot_unit = "m", style = "ticks", location = "br", width_hint = 0.1) +
#        theme(
#
#          legend.position = c(0.128, 0.145),
#          legend.box.background = element_rect(colour = "black", fill = alpha("white", 0.5)),
#          legend.background = element_rect(colour = NA, fill = NA),
#          legend.key = element_rect(colour = NA, fill = NA),
#          legend.margin = margin(3, 5, 3, 5),
#          legend.spacing.y = unit(3, "points"),
#          legend.text = element_text(size = 8),
#          legend.title = element_text(size = 9),
#
#          panel.border = element_rect(colour = "black", fill = NA, size = 1)
#        )
# m
# ggsave("3.outcomes/paper/study_area_map.png", plot = m, width = 5, height = 5, units = "in")
#
#
#
#
#
#
#
#
#
#
#
#
#
# # For Ron
#
# mod4 <- read_fst('1.work/joined/mod4_2010.fst', as.data.table = TRUE) %>%
#         .[, .(date, grd_200m_id, lat, lon, tmin, tmean, tmax, pred_tmin, pred_tmean, pred_tmax,
#               res_tmin, res_tmean, res_tmax)]
#
# x <- mod4[, lapply(.SD, mean), .SDcols = c("lat", "lon", "res_tmin", "res_tmean", "res_tmax"),
#           by = "grd_200m_id"] %>%
#      as.data.frame %>%
#      st_as_sf(coords = c("lon", "lat"), crs = 4326)
#
# for (tvar in c("tmin", "tmean", "tmax")) {
#
#   for (mo in 1:12) {
#
#     report(glue("{tvar} {day}"))
#
#     day <- glue("2010-{mo}-15") %>%
#            strftime("%Y-%m-%d")
#
#     dat <- mod4[date == day, .(date, grd_200m_id, lat, lon, res_tmin, res_tmean, res_tmax)] %>%
#            as.data.frame %>%
#            st_as_sf(coords = c("lon", "lat"), crs = 4326)
#
#     m <- ggplot(dat) +
#            geom_sf(data = france, fill = NA) +
#            geom_sf(aes(colour = res_tmin)) +
#            scale_colour_gradient2(low = "#b2182b", mid = "white", high = "#2166ac") +
#            labs(title = glue("{tvar} {day}")) +
#            theme_bw()
#
#     glue("3.outcomes/model_performance/daily_resid/{tvar}_{day}.png") %>%
#       ggsave(width = 8, height = 8)
#
#   }
#
# }
#
#
#
#
