library(magrittr)
library(glue)
library(data.table)
library(fst)
library(sf)
library(ggplot2)
library(ggmap)

source("2.scripts/helpers/constants.R")
source("2.scripts/helpers/report.R")

report("")
report("Mapping model residuals")
report("---")



# Stage 1 (1 km calibration) residuals ----------

report("Loading m1 CV predictions (1km temperatures)")
m1_cv <- read_fst("1.work/models/m1_all_cv.fst", as.data.table = TRUE)
m1_cv[, res := pred_m1_cv - obs]


# Print a summary of the residuals
m1_cv[, .(rmse = sqrt(mean(res^2)), mae = mean(abs(res)), mean = mean(res), sd = sd(res),
          median = median(res), iqr = IQR(res)), by = tvar]


report("Calculating mean and sd of 1 km residuals for each station")
smry_1km <- m1_cv[, .(res = mean(res), sd = sd(res), .N), by = c("tvar", "stn_id")]
rm(m1_cv)

# Sort by residual value
# This will alter plotting order to emphasize stations with large residuals
smry_1km[, r_ord := abs(res)]
setorderv(smry, cols = c("tvar", "r_ord"))
smry_1km[, r_ord := NULL]

# Calculate each station's contribution to the total number of observations
# This will be used to set point size to deemphasize stations with few observations
smry_1km <- smry_1km[, .(n_tot = sum(N)), by = tvar] %>%
            smry_1km[., on = "tvar"]
smry_1km[, n_share := N / n_tot]
smry_1km[, n_tot := NULL]

# Set up faceting
smry_1km[, tvar := chartr("t", "T", tvar) %>% factor(levels = c("Tmin", "Tmean", "Tmax"))]

# Add geometry
report("Joining station locations")
stns <- st_read("1.work/grids/stations_gridded.shp") %>%
        st_transform(4326)

# Keep a single location for each station (some stations relocated during study period)
stns <- stns[!duplicated(stns$insee_id), ]

# Join
pts_1km <- as.data.table(stns) %>%
           .[smry_1km, .(stn_id, tvar, res, sd, N, n_share, geometry), on = c(insee_id = "stn_id")] %>%
           st_as_sf

# Save
report("Saving 1 km residuals")
st_write(pts_1km, "3.outcomes/paper/m1_resid.shp")
report("-> 3.outcomes/paper/m1_resid.shp")


report("Mapping 1 km residuals")
france <- st_read("1.work/borders/france_epsg-2154.shp")
theme_set(theme_bw())

# Mean residual per station
res_1km <- ggplot() +
           geom_sf(data = france, fill = "white") +
           geom_sf(data = pts_1km, aes(colour = res, alpha = abs(res), size = n_share)) +
           scale_color_distiller(
             type = "div",
             palette = "RdBu",
             limits = c(-3.1, 3.1),
             breaks = seq(-3, 3, by = 1),
             guide = guide_colourbar(
                       title = "Mean Residual",
                       title.position = "right",
                       title.vjust = 0.85)
                     ) +
           scale_alpha(range = c(0.5, 1), limits = c(0, 3.1), guide = FALSE) +
           scale_size(range = c(0.2, 1), guide = FALSE) +
           annotation_scale(location = "br", style = "ticks", width_hint = 0.1) +
           facet_wrap(~ tvar) +
           theme(
             axis.text = element_blank(),
             axis.ticks = element_blank(),
             legend.position = "bottom",
             legend.justification = "left",
             legend.key.width = unit(1.03, "cm"),
             legend.margin = margin(0),
             legend.text = element_text(size = 10),
             panel.grid = element_line(colour = "transparent"),
             plot.margin = margin(0)
           )
ggsave("3.outcomes/paper/res_1km_mean.jpg", res_1km, width = 6.3, height = 3, dpi = 300)
report("-> 3.outcomes/paper/res_1km_mean.jpg")

# Standard deviation of residual per station
sd_1km <- ggplot() +
          geom_sf(data = france, fill = "white") +
          geom_sf(data = pts_1km, aes(colour = sd, alpha = abs(sd), size = n_share)) +
          scale_color_distiller(
            type = "div",
            palette = "Oranges",
            direction = 1,
            guide = guide_colourbar(
              title = "Standard Deviation of Residuals",
              title.position = "right",
              title.vjust = 0.85)
          ) +
          scale_alpha(range = c(0.5, 1), guide = FALSE) +
          scale_size(range = c(0.2, 1), guide = FALSE) +
          annotation_scale(location = "br", style = "ticks", width_hint = 0.1) +
          facet_wrap(~ tvar) +
          theme(
            axis.text = element_blank(),
            axis.ticks = element_blank(),
            legend.position = "bottom",
            legend.justification = "left",
            legend.key.width = unit(1.03, "cm"),
            legend.margin = margin(0),
            legend.text = element_text(size = 10),
            panel.grid = element_line(colour = "transparent"),
            plot.margin = margin(0)
          )
ggsave("3.outcomes/paper/res_1km_sd.jpg", sd_1km, width = 6.3, height = 3, dpi = 300)
report("-> 3.outcomes/paper/res_1km_sd.jpg")

# INDIVIDUAL MAPS
# measures <- c("Tmin", "Tmean", "Tmax")
# res_maps <- lapply(measures, function(measure) {
#
#               dat <- pts_1km[pts_1km$tvar == measure, ]
#               subscript <- sub("T", "", measure)
#
#               ggplot() +
#                 geom_sf(data = france, fill = "white") +
#                 geom_sf(data = dat, aes(colour = res, alpha = abs(res), size = n_share)) +
#                 scale_color_distiller(
#                   type = "div",
#                   palette = "RdBu",
#                   limits = c(-3.1, 3.1),
#                   breaks = seq(-3, 3, by = 1),
#                   guide = guide_colourbar(title = "Residual", label.hjust = 1)
#                 ) +
#                 scale_alpha(range = c(0.5, 1), limits = c(0, 3.1), guide = FALSE) +
#                 scale_size(range = c(0.2, 2), guide = FALSE) +
#                 annotation_scale(location = "bl", style = "ticks", width_hint = 0.1) +
#                 labs(tag = bquote(T[.(subscript)])) +
#                 theme(
#                   axis.text = element_blank(),
#                   axis.ticks = element_blank(),
#                   legend.background = element_rect(fill = "transparent"),
#                   legend.justification = c("left", "bottom"),
#                   legend.key.height = unit(20, "pt"),
#                   legend.position = c(0.005, 0.1),
#                   legend.text = element_text(size = 10),
#                   legend.title = element_text(size = 12),
#                   panel.grid = element_line(colour = "transparent"),
#                   lot.margin = margin(0),
#                   plot.tag = element_text(
#                     margin = margin(t = 12, r = -38, b = -24),
#                     size = 12
#                   )
#                 )
#             })


# Stage 4 (200 m ensemble) residuals ----------

report("Loading m4 CV predictions (200 m residuals)")
m4_cv <- read_fst("1.work/models/m4_all_cv.fst", as.data.table = TRUE) %>%
         setnames("res", "m1_res")
m4_cv[, res := pred_m4_cv - m1_res]


# Print a summary of the residuals
m4_cv[, .(rmse = sqrt(mean(res^2)), mae = mean(abs(res)), mean = mean(res), sd = sd(res),
          median = median(res), iqr = IQR(res)), by = tvar]


report("Calculating mean and sd of m4 residuals")
smry_200m <- m4_cv[, .(res = mean(res), sd = sd(res), .N), by = c("tvar", "grd_200m_id")]
rm(m4_cv)

# Sort by residual value
# This will alter plotting order to emphasize locations with large residuals
smry_200m[, r_ord := abs(res)]
setorderv(smry_200m, cols = c("tvar", "r_ord"))
smry_200m[, r_ord := NULL]

# Calculate each location's contribution to the total number of observations
# This will be used to set point size to deemphasize locations with few observations
smry_200m <- smry_200m[, .(n_tot = sum(N)), by = tvar] %>%
             smry_200m[., on = "tvar"]
smry_200m[, n_share := N / n_tot]
smry_200m[, n_tot := NULL]

# Set up faceting
smry_200m[, tvar := chartr("t", "T", tvar) %>% factor(levels = c("Tmin", "Tmean", "Tmax"))]

# Add geometry
report("Joining station locations")
pts_200m <- as.data.table(stns) %>%
            .[smry_200m, .(insee_id, tvar, res, sd, N, n_share, geometry),
              on = c(grd_200m = "grd_200m_id")] %>%
            st_as_sf

# Save
report("Saving 200 m residuals")
st_write(pts_1km, "3.outcomes/paper/m4_resid.shp")
report("-> 3.outcomes/paper/m4_resid.shp")


report("Mapping 200 m residuals")

# Mean residual per station
res_200m <- ggplot() +
            geom_sf(data = france, fill = "white") +
            geom_sf(data = pts_200m, aes(colour = res, alpha = abs(res), size = n_share)) +
            scale_color_distiller(
              type = "div",
              palette = "RdBu",
              limits = c(-0.35, 0.35),
              breaks = seq(-0.3, 0.3, length.out = 7),
              guide = guide_colourbar(
                title = "Mean Residual",
                title.position = "right",
                title.vjust = 0.85)
            ) +
            scale_alpha(range = c(0.5, 1), limits = c(0, 0.35), guide = FALSE) +
            scale_size(range = c(0.2, 1), guide = FALSE) +
            annotation_scale(location = "br", style = "ticks", width_hint = 0.1) +
            facet_wrap(~ tvar) +
            theme(
              axis.text = element_blank(),
              axis.ticks = element_blank(),
              legend.position = "bottom",
              legend.justification = "left",
              legend.key.width = unit(1.03, "cm"),
              legend.margin = margin(0),
              legend.text = element_text(size = 10),
              panel.grid = element_line(colour = "transparent"),
              plot.margin = margin(0)
            )
ggsave("3.outcomes/paper/res_200m_mean.jpg", res_200m, width = 6.3, height = 3, dpi = 300)
report("-> 3.outcomes/paper/res_200m_mean.jpg")

# Standard deviation of residual per station
sd_200m <- ggplot() +
           geom_sf(data = france, fill = "white") +
           geom_sf(data = pts_200m, aes(colour = sd, alpha = abs(sd), size = n_share)) +
           scale_color_distiller(
             type = "div",
             palette = "Oranges",
             direction = 1,
             guide = guide_colourbar(
               title = "Standard Deviation of Residuals",
               title.position = "right",
               title.vjust = 0.85)
           ) +
           scale_alpha(range = c(0.5, 1), guide = FALSE) +
           scale_size(range = c(0.2, 1), guide = FALSE) +
           annotation_scale(location = "br", style = "ticks", width_hint = 0.1) +
           facet_wrap(~ tvar) +
           theme(
             axis.text = element_blank(),
             axis.ticks = element_blank(),
             legend.position = "bottom",
             legend.justification = "left",
             legend.key.width = unit(1.03, "cm"),
             legend.margin = margin(0),
             legend.text = element_text(size = 10),
             panel.grid = element_line(colour = "transparent"),
             plot.margin = margin(0)
           )
ggsave("3.outcomes/paper/res_200m_sd.jpg", sd_200m, width = 6.3, height = 3, dpi = 300)
report("-> 3.outcomes/paper/res_200m_sd.jpg")
