#
# Tables and figures for revised paper submitted to Environmental Research
#

source("init.R")


# Setup -----

# Should the script overwrite previous summaries?
FORCE <- FALSE


# Functions -----

# Count LST pixels per day over entire study period
count_lst <- function(path, force = FALSE) {

               if (file.exists(path) && !force) {

                 report("Loading daily LST counts")
                 counts <- read_fst(path, as.data.table = TRUE)
                 return(counts)

               }

               report("Counting LST presence on each day during study period")
               counts <- lapply(constants$model_years, function(year) {

                 lapply(1:12, function(month) {

                   cols <- c("date", "aqua_day_lst", "terra_day_lst", "aqua_night_lst", "terra_night_lst")
                   dt <- glue("1.work/joined/mod3_{year}-{sprintf('%02d', month)}.fst") %>%
                         read_fst(columns = cols, as.data.table = TRUE) %>%
                         .[, .(
                               ad = sum(!is.na(aqua_day_lst)),
                               td = sum(!is.na(terra_day_lst)),
                               an = sum(!is.na(aqua_night_lst)),
                               tn = sum(!is.na(terra_night_lst))
                             ), by = .(date)]
                   dt

                 }) %>%
                 rbindlist

               }) %>%
               rbindlist

               # Save
               report("Saving for reference")
               write_fst(counts, path, compress = 100)
               report(glue("-> {path}"))

               # Return
               counts

          }

# Convert a gridded dataset to a raster
grid_to_raster <- function(dt, xcol, ycol, prj, new_prj = NULL) {

  # Ensure dt is a data.table
  if (!is.data.table(dt)) {
    dt <- as.data.table(dt)
  }

  # Exclude any character-type data columns (rasters layers must be numeric)
  char_cols <- map_lgl(dt, is.character) %>%
               which %>%
               names
  if (length(char_cols) > 0) {

    warning(glue("Dropping character columns `{paste(char_cols, collapse = '` `')}`"))
    dt <- dt[, !(char_cols), with = FALSE]

  }

  # Convert to spatial points
  coords <- dt[, c(xcol, ycol), with = FALSE]
  dt <- dt[, !c(xcol, ycol), with = FALSE]
  coordinates(dt) <- coords
  proj4string(dt) <- prj
  if (!is.null(new_prj)) {
    dt <- spTransform(dt, new_prj)
  }
  gridded(dt) <- TRUE
  brick(dt)

}

# Load saved m3 importances
load_m3_importance <- function(path, force = FALSE) {

  if (file.exists(path) && !force) {

    report("Loading m3 feature importances")
    m3_imp <- read_fst(path, as.data.table = TRUE)
    return(m3_imp)

  }

  report("Getting m3 feature importances for all years")

  imps <- expand.grid(
            year = constants$model_years,
            tvar = c("tmin", "tmean", "tmax"),
            algo = c("rf", "xgb"),
            stringsAsFactors = FALSE
          ) %>%
          pmap_dfr(function(year, tvar, algo) {

            report(glue("{algo} {tvar} {year}"))
            m <- glue("~/code/P055.FR.Ta/1.work/models/m3_{tvar}_{year}_{algo}_model.rds") %>%
                 readRDS()
            n <- getTaskSize(m)
            imp <- getFeatureImportance(m)$res
            names(imp)[which(names(imp) %like% "pred_")] <- "pred_m2"
            c(algo = algo, tvar = tvar, year = year, n = n, as.list(imp))

          }) %>%
          as.data.table %>%
          setcolorder(c("algo", "tvar", "year", "n", "x", "y", "day", "pred_m2", "bt", "ndvi",
                        "elev", "pop", "artificial", "vegetation", "bare", "water", "climate")) %>%
          setkeyv(c("algo", "tvar", "year"))

  # Save
  report("Saving for reference")
  write_fst(imps, path, compress = 100)
  report(glue("-> {path}"))

  # Return
  imps

}


# 1 km Ta on select days -----

dir.create("1.work/summaries/paper", showWarnings = FALSE)
lst_counts <- count_lst("1.work/summaries/paper/lst_counts.fst", force = FORCE)

for (d in c("2003-02-18", "2009-05-29", "2012-08-10", "2015-11-01")) {

  d <- as.Date(d)

  ta_1km <- glue("1.work/predictions/mod3_{year(d)}_pred_1km.fst") %>%
            read_fst(as.data.table = TRUE, from = yday(d - 1) * 627787 + 1, to = yday(d) * 627787) %>%
            .[date == d, ]
  brk_1km <- ta_1km[, .(grd_1km_x, grd_1km_y, tmin, tmean, tmax)] %>%
             .[, trng := tmax - tmin] %>%
             grid_to_raster(xcol = "grd_1km_x", ycol = "grd_1km_y", prj = constants$epsg_2154,
                            new_prj = constants$proj_sinu)
  path <- glue("1.work/summaries/paper/m1_pred_{d}.tif")
  save_geotiff(brk_1km, path)
  report(glue("-> {path}"))

  ta_200m <- glue("1.work/predictions/mod5_{year(d)}_pred_m4.fst") %>%
             read_fst(as.data.table = TRUE, from = yday(d - 1) * 246315 + 1, to = yday(d) * 246315) %>%
             .[date == d, ]
  brk_200m <- ta_200m[, .(x, y, tmin, tmean, tmax)] %>%
              .[, trng := tmax - tmin] %>%
              grid_to_raster(xcol = "x", ycol = "y", prj = constants$epsg_3035)
  path <- glue("1.work/summaries/paper/m4_pred_{d}.tif")
  save_geotiff(brk_200m, path)
  report(glue("-> {path}"))

}


# Feature importances for m3 RF and XGB -----

m3_imp <- load_m3_importance("1.work/summaries/paper/m3_importances.fst", force = F) %>%
          .[, climate.9 := NULL] %>%
          setnames(
            c("pred_m2", "bt", "ndvi", "climate.1", "climate.2", "climate.3", "climate.4",
              "climate.5", "climate.6", "climate.7", "climate.8"),
            c("Tap_1km", "Tb", "NDVI", "Mountain", "Semi-continental", "Modified oceanic",
              "Trans. oceanic", "Oceanic", "Mod. Mediterranean", "Southwest basin",
              "Mediterranean")
          ) %>%
          melt(id.vars = c("algo", "tvar", "year", "n"), variable.name = "feature",
               value.name = "imp")
m3_imp[, tvar := factor(
                   tvar, levels = c("tmin", "tmean", "tmax"),
                   labels = c("Rtmin", "Rtmean", "Rtmax")
                 )]
m3_imp[, pct := imp / sum(imp, na.rm = TRUE), by = .(algo, tvar, year)]
m3_imp[, order := mean(pct), by = .(algo, tvar, feature)]

m3_imp[algo == "rf", ] %>%
  na.omit() %>%
  ggplot() +
    geom_boxplot(aes(x = reorder(feature, order), y = pct)) +
    coord_flip() +
    facet_grid(. ~ tvar) +
    labs(x = "Feature", y = "Relative importance (%)") +
    theme_bw()
ggsave("3.outcomes/figures/m3_rf_importance.png", width = 20, height = 10, units = "cm")

m3_imp[algo == "xgb", ] %>%
  na.omit() %>%
  ggplot() +
    geom_boxplot(aes(x = reorder(feature, order), y = pct)) +
    coord_flip() +
    facet_grid(. ~ tvar) +
    labs(x = "Feature", y = "Relative importance (%)") +
    theme_bw()
ggsave("3.outcomes/figures/m3_xgb_importance.png", width = 20, height = 10, units = "cm")
