library(magrittr)
library(parallel)
library(data.table)
library(fst)
library(sp)
library(ncdf4)
library(raster)
library(mapview)

source("2.scripts/helpers/constants.R")
source("2.scripts/helpers/report.R")

preds_as_raster <- function(dtable) {
  coordinates(dtable) <- ~ grd_1km_x + grd_1km_y
  proj4string(dtable) <- constants$epsg_2154
  dtable <- spTransform(dtable, constants$proj_sinu)
  gridded(dtable) <- TRUE
  stack(dtable)
}



# 1. Jan 1 in southwestern France:

# Load predictions from the original 2001-2013 tmean model
orig <- brick("0.raw/kloog_france_temperature/temp_kloog_1km_2009.nc")
proj4string(orig) <- "+proj=lcc +nadgrids=ntf_r93.gsb,null +a=6378249.2000 +rf=293.4660210000000 +pm=2.337229167 +lat_0=46.800000000 +lon_0=0.000000000 +k_0=0.99987742 +lat_1=46.800000000 +x_0=600000.000 +y_0=2200000.000 +units=m +no_defs"

# Load intermediate and final predictions from the new models
jan1_final <- read_fst("1.work/predictions/mod3_2009_pred_final.fst", as.data.table = TRUE, to = 627787)
jan1_mod3 <- read_fst("1.work/predictions/mod3_2009_pred_m2.fst", as.data.table = TRUE, to = 627787)

# Create raster of predictions
jan1_final_r <- preds_as_raster(jan1_final)
jan1_mod3_r <- preds_as_raster(jan1_mod3)

# Map all of France
jan1_map <- mapview(orig$X1,                                alpha.regions = 1, maxpixels = 1e7, method = "ngb", na.color = NA, legend = TRUE)
jan1_map <- mapview(jan1_final_r$tmean,     map = jan1_map, alpha.regions = 1, maxpixels = 1e7, method = "ngb", na.color = NA, legend = TRUE)
jan1_map <- mapview(jan1_final_r$tmean_src, map = jan1_map, alpha.regions = 1, maxpixels = 1e7, method = "ngb", na.color = NA, legend = TRUE)
jan1_map <- mapview(jan1_mod3_r$idw_tmean,  map = jan1_map, alpha.regions = 1, maxpixels = 1e7, method = "ngb", na.color = NA, legend = TRUE)
jan1_map

# Map southwest
jan1_ext <- extent(-151000, 270000, 4800000, 5050000)
jan1_ext_orig <- crop(jan1_final_r, jan1_ext) %>% projectExtent(crs(orig)) %>% extent

plot(orig$X1,                ext = jan1_ext_orig, axes = FALSE, box = FALSE, col = viridis::inferno(100), main = "Old model Tmean 2009-01-01")
plot(jan1_final_r$tmean,     ext = jan1_ext,      axes = FALSE, box = FALSE, col = viridis::inferno(100), main = "New model Tmean 2009-01-01")
plot(jan1_final_r$tmean_src, ext = jan1_ext,      axes = FALSE, box = FALSE, col = viridis::inferno(3),   main = "New model Tmean source 2009-01-01")
plot(jan1_mod3_r$idw_tmean,  ext = jan1_ext,      axes = FALSE, box = FALSE, col = viridis::inferno(100), main = "IDW tmean 2009-01-01")



# 2. Jan 3 in massif central / Rhone valley

# Load intermediate and final predictions from the new models
jan3_final <- read_fst("1.work/predictions/mod3_2009_pred_final.fst", as.data.table = TRUE, from = (627787 * 2) + 1, to = 627787 * 3)
jan3_mod3 <- read_fst("1.work/predictions/mod3_2009_pred_m2.fst", as.data.table = TRUE, from = (627787 * 2) + 1, to = 627787 * 3)

# Create raster of predictions
jan3_final_r <- preds_as_raster(jan3_final)
jan3_mod3_r <- preds_as_raster(jan3_mod3)

# Map all of France
jan3_map <- mapview(orig$X3,                                alpha.regions = 1, maxpixels = 1e7, method = "ngb", na.color = NA, legend = TRUE)
jan3_map <- mapview(jan3_final_r$tmean,     map = jan3_map, alpha.regions = 1, maxpixels = 1e7, method = "ngb", na.color = NA, legend = TRUE)
jan3_map <- mapview(jan3_final_r$tmean_src, map = jan3_map, alpha.regions = 1, maxpixels = 1e7, method = "ngb", na.color = NA, legend = TRUE)
jan3_map <- mapview(jan3_mod3_r$idw_tmean,  map = jan3_map, alpha.regions = 1, maxpixels = 1e7, method = "ngb", na.color = NA, legend = TRUE)
jan3_map

# Map massif central / Rhone valley
jan3_ext <- extent(181000, 435000, 4930000, 5100000)
jan3_ext_orig <- crop(jan3_final_r, jan3_ext) %>% projectExtent(crs(orig)) %>% extent

plot(orig$X3,                ext = jan3_ext_orig, axes = FALSE, box = FALSE, col = viridis::inferno(100), main = "Old model Tmean 2009-01-03")
plot(jan3_final_r$tmean,     ext = jan3_ext,      axes = FALSE, box = FALSE, col = viridis::inferno(100), main = "New model Tmean 2009-01-03")
plot(jan3_final_r$tmean_src, ext = jan3_ext,      axes = FALSE, box = FALSE, col = viridis::inferno(3),   main = "New model Tmean source 2009-01-03")

