library(magrittr)
library(data.table)
library(fst)
library(sf)
library(raster)
library(rasterVis)
library(ggplot2)

# Load helper functions
source("2.scripts/helpers/constants.R")
source("2.scripts/helpers/report.R")

# Map of study area with meteo stations ----------------------------------------

france <- st_read("1.work/borders/france_epsg-4326.shp")

meteo_cols <- c("insee_id", "station_type", "latitude", "longitude", "date", "temperature_mean",
                "temperature_min", "temperature_max")
meteo <- paste0("meteo_data_", constants$model_years, "-clean.fst") %>%
         file.path("1.work/meteo_obs/clean", .) %>%
         lapply(read_fst, as.data.table = TRUE, columns = meteo_cols) %>%
         rbindlist
names(meteo) %>% gsub("temperature_", "t", .) %>% setnames(meteo, .)
setnames(meteo, c("station_type", "latitude", "longitude"), c("type", "lat", "lon"))

# Use only 2016 meteo stations to reduce clutter
stn_pts <- meteo[year(date) == 2016, ] %>%
           .[, lapply(.SD, last), by = insee_id, .SDcols = c("lat", "lon", "type")] %>%
           .[order(type), ] %>%
           as.data.frame %>%
           st_as_sf(coords = c("lon", "lat"), crs = 4326)

# Make the map
map <- ggplot() +
       geom_sf(data = france, fill = "#dddddd", show.legend = FALSE) +
       geom_sf(data = stn_pts, show.legend = FALSE)


# Table descriptive statistics for Ta_obs and LST -----------------------------

# Meteo observations
summarize <- function(x) {
  c(
    min = min(x, na.rm = TRUE) %>% round(1),
    mean = mean(x, na.rm = TRUE) %>% round(1),
    max = max(x, na.rm = TRUE) %>% round(1),
    sd = sd(x, na.rm = TRUE) %>% round(1),
    obs = is.na(x) %>% not %>% which %>% length
  )
}

days <- uniqueN(meteo[, date])
days_per_year <- meteo[, .(days = unique(date) %>% length), by = year(date)]

tmin_stats <- meteo[, summarize(tmin) %>% as.list, by = year(date)] %>%
              .[, obs_per_day := round(obs / days_per_year$days)] %>%
              .[, year := as.character(year)]
tmin_stats <- meteo[, summarize(tmin) %>% as.list] %>%
              .[, obs_per_day := round(obs / days)] %>%
              .[, year := "Total"] %>%
              rbind(tmin_stats, .)

tmean_stats <- meteo[, summarize(tmean) %>% as.list, by = year(date)] %>%
               .[, obs_per_day := round(obs / days_per_year$days)]
tmean_stats <- meteo[, summarize(tmean) %>% as.list] %>%
               .[, obs_per_day := round(obs / days)] %>%
               .[, year := "Total"] %>%
               rbind(tmean_stats, .)

tmax_stats <- meteo[, summarize(tmax) %>% as.list, by = year(date)] %>%
              .[, obs_per_day := round(obs / days_per_year$days)]
tmax_stats <- meteo[, summarize(tmax) %>% as.list] %>%
              .[, obs_per_day := round(obs / days)] %>%
              .[, year := "Total"] %>%
              rbind(tmax_stats, .)

tmin_stats
tmean_stats
tmax_stats


# Maps of mean and sd of predictions over entire study period -----------------

tvars <- c("tmin", "tmean", "tmax")

for (tvar in tvars) {

  smrys <- lapply(constants$model_years, function(year) {
    paste0("1.work/predictions/mod3_", year, "_pred_final_", tvar, "_smry.tif") %>%
      brick %>%
      set_names(c("id", "min", "mean", "max", "sd", "n"))
  })

  means <- lapply(smrys, function(x) { x$mean }) %>% do.call(stack, .)
  sds <- lapply(smrys, function(x) { x$sd }) %>% do.call(stack, .)
  weights <- lapply(smrys, function(x) { x$n }) %>% do.call(stack, .)

  smry <- stack(
    weighted.mean(means, weights),
    weighted.mean(sds, weights)
  ) %>% set_names(c("mean", "sd"))
  paste0("1.work/predictions/mod3_all_pred_final_", tvar, "_smry.tif") %>% save_geotiff(smry, .)

}

means <- lapply(tvars, function(tvar) {
  paste0("1.work/predictions/mod3_all_pred_final_", tvar, "_smry.tif") %>% brick %>% extract2(1)
}) %>% do.call(stack, .) %>% set_names(tvars)

means_map <- levelplot(
  means,
  at = seq(-5, 20, len = 101),
  margin = FALSE,
  scales = list(draw = FALSE),
  colorkey = list(
    space = 'bottom',
    axis.line = list(col = "black")
  ),
  par.settings = magmaTheme(
    axis.line = list(col = 'transparent')
  )
)

# Save to disk
png("3.outcomes/mean_t.png", width = 1800, height = 900, res = 300)
means_map
dev.off()

sds <- lapply(tvars, function(tvar) {
  paste0("1.work/predictions/mod3_all_pred_final_", tvar, "_smry.tif") %>% brick %>% extract2(2)
}) %>% do.call(stack, .) %>% set_names(tvars)

sds_map <- levelplot(
  sds,
  at = seq(0, 10, len = 101),
  margin = FALSE,
  scales = list(draw = FALSE),
  colorkey = list(
    space = 'bottom',
    axis.line = list(col = "black")
  ),
  par.settings = viridisTheme(
    axis.line = list(col = 'transparent')
  )
)

# Save to disk
png("3.outcomes/sd_t.png", width = 1800, height = 900, res = 300)
sds_map
dev.off()
