library(data.table)
library(dplyr)
library(fst)
library(glue)
library(lubridate)
library(parallel)
library(sf)
library(stringr)
library(terra)

data_years <- 2000:2018
force <- FALSE
ncores <- 32

# Load temperature model grids
grd_1km <- fst::read_fst("1.work/grids/grd_1km.fst", as.data.table = TRUE)
grd_200m <- fst::read_fst("1.work/grids/grd_200m.fst", as.data.table = TRUE)

# Load grid rasters
grd_1km_rst <- terra::rast("1.work/grids/grd_1km_pxs_sinu.tif")
names(grd_1km_rst) <- read.csv("1.work/grids/grd_1km_pxs_sinu.tif.bandnames", header = FALSE)
grd_1km_rst <- grd_1km_rst$idx
grd_200m_rst <- terra::rast("1.work/grids/grd_200m_pxs.tif")
names(grd_200m_rst) <- read.csv("1.work/grids/grd_200m_pxs.tif.bandnames", header = FALSE)

# Add an index to the 200m raster
grd_200m[, idx := .I]
cell_idx <- terra::cellFromXY(grd_200m_rst, grd_200m[, .(x, y)])
grd_200m_rst$idx <- NA
grd_200m_rst <- grd_200m_rst$idx
grd_200m_rst[cell_idx] <- grd_200m$idx
rm(cell_idx)

# Compute area of each grid cell
grd_1km_rst$area <- terra::cellSize(grd_1km_rst)
grd_200m_rst$area <- terra::cellSize(grd_200m_rst)

# Load and merge 200m grid polygons
zones_200m <- sf::st_read("1.work/grids/grd_200m_pys.shp", quiet = TRUE) |>
  sf::st_geometry() |>
  sf::st_transform("EPSG:3035") |>
  sf::st_union()

# Compute intersection and difference of grids with each IRIS
message("Linking IRIS editions to temperature model grids")
iris_years <- list.files(
  "1.work/contours-iris", pattern = "CONTOURS-IRIS_\\d{4}", full.names = TRUE
) |>
  stringr::str_extract("\\d{4}$") |>
  as.integer()
for (iris_year in iris_years) {

  # Check output
  out_paths <- list(
    `200m` = glue::glue("1.work/grids/iris-{iris_year}_ta-grd-200m.csv"),
    `1km` = glue::glue("1.work/grids/iris-{iris_year}_ta-grd-1km.csv")
  )
  if (all(sapply(out_paths, file.exists)) && !force) {
    message(
      "Skipping existing files:\n  ", paste(out_paths, collapse = "\n  "),
      "\nUse force = TRUE to overwrite"
    )
    next
  }

  # Compute grid cells contained in each IRIS
  message("Linking IRIS ", iris_year, " to temperature model grid cells")
  iris_paths <- list.files(
    glue::glue("1.work/contours-iris/CONTOURS-IRIS_{iris_year}"),
    pattern = "(CONTOURS-)?IRIS(_D0)?(\\d\\d)?.shp", full.names = TRUE
  ) |>
    sort()
  if (length(iris_paths) == 0) {
    stop("No IRIS shapefiles found in 1.work/contours-iris/CONTOURS-IRIS_", iris_year)
  } else if (length(iris_paths) > 1) {
    result_200m <- parallel::mclapply(
      iris_paths,
      function(x) {
        iris <- sf::st_read(x, quiet = TRUE) |>
          dplyr::rename_with(tolower) |>
          sf::st_make_valid() |>
          sf::st_transform("EPSG:3035") |>
          sf::st_set_agr("constant")

        iris_200m <- sf::st_intersection(iris, zones_200m) |>
          sf::st_transform(sf::st_crs(grd_200m_rst))
        rslt <- terra::extract(grd_200m_rst, iris_200m, exact = TRUE) |>
          dplyr::filter(!is.na(idx))
        cbind(rslt, iris = iris_200m$dcomiris[rslt$ID])
      },
      mc.cores = ncores, mc.preschedule = FALSE
    )
    result_1km <- parallel::mclapply(
      iris_paths,
      function(x) {
        iris <- sf::st_read(x, quiet = TRUE) |>
          dplyr::rename_with(tolower) |>
          sf::st_make_valid() |>
          sf::st_transform("EPSG:3035") |>
          sf::st_set_agr("constant")

        iris_1km <- sf::st_difference(iris, zones_200m) |>
          sf::st_transform(sf::st_crs(grd_1km_rst))
        rslt <- terra::extract(grd_1km_rst, iris_1km, exact = TRUE) |>
          dplyr::filter(!is.na(idx))
        cbind(rslt, iris = iris_1km$dcomiris[rslt$ID])
      },
      mc.cores = ncores, mc.preschedule = FALSE
    )
  } else {
    iris <- sf::st_read(iris_paths, quiet = TRUE) |>
      dplyr::rename_with(tolower) |>
      sf::st_transform("EPSG:3035") |>
      sf::st_set_agr("constant")
    result_200m <- parallel::mclapply(
      split(iris, sort(rep(1:ncores, length.out = nrow(iris)))),
      function(x) {
        # Compute the intersection of the IRIS with the 200m grid
        x <- sf::st_make_valid(x)
        iris_200m <- sf::st_intersection(x, zones_200m) |>
          sf::st_transform(sf::st_crs(grd_200m_rst))
        rslt <- terra::extract(grd_200m_rst, iris_200m, exact = TRUE) |>
          dplyr::filter(!is.na(idx))
        cbind(rslt, iris = iris_200m$code_iris[rslt$ID])
      },
      mc.cores = ncores
    )
    result_1km <- parallel::mclapply(
      split(iris, sort(rep(1:ncores, length.out = nrow(iris)))),
      function(x) {
        # Compute the intersection of the IRIS with the 200m grid
        x <- sf::st_make_valid(x)
        iris_1km <- sf::st_difference(x, zones_200m) |>
          sf::st_transform(sf::st_crs(grd_1km_rst))
        rslt <- terra::extract(grd_1km_rst, iris_1km, exact = TRUE) |>
          dplyr::filter(!is.na(idx))
        cbind(rslt, iris = iris_1km$code_iris[rslt$ID])
      },
      mc.cores = ncores
    )
    rm(iris)
  }
  result_200m <- dplyr::bind_rows(result_200m) |>
    data.table::setDT()
  result_1km <- dplyr::bind_rows(result_1km) |>
    data.table::setDT()

  # Compute the area of each IRIS-grid cell intersection
  result_200m <- result_200m[, .(area = area * fraction), keyby = .(idx, iris)]
  result_1km <- result_1km[, .(area = area * fraction), keyby = .(idx, iris)]

  # Add cell IDs
  result_200m[, grd_200m_id := grd_200m[result_200m, .(id), on = "idx"]]
  result_1km[, grd_1km_id := grd_1km[result_1km, .(id), on = "idx"]]

  # Save IRIS-grid cells crosswalks
  out_path <- out_paths$`200m`
  tmp_path <- paste0(out_path, ".tmp")
  data.table::fwrite(result_200m[, .(grd_200m_id, iris, area)], tmp_path)
  file.rename(tmp_path, out_path)
  message("-> ", out_path)

  out_path <- out_paths$`1km`
  tmp_path <- paste0(out_path, ".tmp")
  data.table::fwrite(result_1km[, .(grd_1km_id, iris, area)], tmp_path)
  file.rename(tmp_path, out_path)
  message("-> ", out_path)

  rm(result_200m, result_1km, iris_paths)

}
rm(grd_1km_rst, grd_200m_rst, zones_200m)

# Extract daily temperature for each IRIS
iris_years <- list.files(
  "1.work/grids", pattern = "iris-\\d{4}_ta-grd-1km.csv", full.names = TRUE
) |>
  stringr::str_extract("\\d{4}") |>
  as.integer()
for (data_year in data_years) {

  # Load IRIS - grid crosswalks
  iris_year <- min(iris_years[iris_years >= data_year])
  cross_1km <- glue::glue("1.work/grids/iris-{iris_year}_ta-grd-1km.csv") |>
    data.table::fread(colClasses = c(iris = "character"), key = c("grd_1km_id", "iris"))
  cross_200m <- glue::glue("1.work/grids/iris-{iris_year}_ta-grd-200m.csv") |>
    data.table::fread(colClasses = c(iris = "character"), key = c("grd_200m_id", "iris"))

  ta_1km_path <- glue::glue("1.work/predictions/mod3_{data_year}_pred_1km.fst")
  ta_200m_path <- glue::glue("1.work/predictions/mod5_{data_year}_pred_200m.fst")
  dates_1km <- fst::fst(ta_1km_path)$date
  dates_200m <- fst::fst(ta_200m_path)$date
  mclapply(
    1:12,
    function(month) {
      month_start <- lubridate::date(glue::glue("{data_year}-{month}-01"))
      out_path <- glue::glue(
        "1.work/predictions/ta_iris-{iris_year}_{strftime(month_start, '%Y-%m')}.csv"
      )
      if (file.exists(out_path) && !force) {
        message("Skipping existing file: ", out_path, ". Use force = TRUE to overwrite")
        next
      }
      message(
        "Computing daily temperature in ", strftime(month_start, "%Y-%m"), " for IRIS ", iris_year
      )

      # Load 1 km temperature data
      idx_1km <- which(
        dates_1km >= month_start & dates_1km < month_start + lubridate::period(1, "month")
      )
      ta_1km <- fst::read_fst(
        ta_1km_path, as.data.table = TRUE, from = min(idx_1km), to = max(idx_1km)
      )
      rm(idx_1km)

      # Compute area-weighted mean of 1 km temperatures for each IRIS
      result_1km <- ta_1km[
        cross_1km, .(iris, date, area, tmin, tmean, tmax), on = "grd_1km_id", by = .EACHI
      ]
      result_1km <- result_1km[
        ,
        .(
          area = sum(area),
          tmin = round(sum(area * tmin) / sum(area), 1),
          tmean = round(sum(area * tmean) / sum(area), 1),
          tmax = round(sum(area * tmax) / sum(area), 1)
        ),
        keyby = .(iris, date)
      ]
      rm(ta_1km)

      # Load 200 m temperature data
      idx_200m <- which(
        dates_200m >= month_start & dates_200m < month_start + lubridate::period(1, "month")
      )
      ta_200m <- fst::read_fst(
        ta_200m_path, as.data.table = TRUE, from = min(idx_200m), to = max(idx_200m)
      )
      rm(idx_200m)

      # Compute area-weighted mean of 200 m temperatures for each IRIS
      result_200m <- ta_200m[
        cross_200m, .(iris, date, area, tmin, tmean, tmax), on = "grd_200m_id", by = .EACHI
      ]
      result_200m <- result_200m[
        ,
        .(
          area = sum(area),
          tmin = round(sum(area * tmin) / sum(area), 1),
          tmean = round(sum(area * tmean) / sum(area), 1),
          tmax = round(sum(area * tmax) / sum(area), 1)
        ),
        keyby = .(iris, date)
      ]
      rm(ta_200m)

      # Combine 1 km and 200 m temperatures
      result <- rbind(result_1km, result_200m)[
        ,
        .(
          tmin = round(sum(area * tmin) / sum(area), 1),
          tmean = round(sum(area * tmean) / sum(area), 1),
          tmax = round(sum(area * tmax) / sum(area), 1)
        ),
        keyby = .(iris, date)
      ]
      rm(result_1km, result_200m)

      # Save
      tmp_path <- paste0(out_path, ".tmp")
      fwrite(result, tmp_path)
      file.rename(tmp_path, out_path)
      message("-> ", out_path)
      rm(result, idx)

      return(out_path)
    },
    mc.cores = min(12, ncores), mc.preschedule = FALSE
  )

}
