# Run the model for all years

# Load packages, helper functions, and constants
source("init.R")


# Set to TRUE to overwrite existing output
OVERWRITE <- FALSE


# Create a logfile
dir.create("log", showWarnings = FALSE)
constants$logfile <- Sys.time() %>% format("log/%Y-%m-%d_%H%M%S.log")


report(glue("OAR_JOB_ID: {Sys.getenv('OAR_JOB_ID')}"))
report(glue("OVERWRITE: {OVERWRITE}"))
report(glue("LOG: {constants$logfile}"))


report("---------------")
report("Making Ta model")
report("---------------")


# Create model grids
source_or_skip("2.scripts/grid_setup/check_modis_alignment.R")   # Check alignment of MODIS NDVI and LST data
source_or_skip("2.scripts/grid_setup/create_grd_1km.R")          # Create a 1 km grid aligned to MODIS LST data
source_or_skip("2.scripts/grid_setup/buffer_grd_1km.R")          # Create buffers around the 1 km grid centroids
source_or_skip("2.scripts/grid_setup/create_grd_200m.R")         # Create a 200 m grid
source_or_skip("2.scripts/grid_setup/create_grd_crossref.R")     # Associate each 200 m grid cell with a 1 km grid cell
source_or_skip("2.scripts/grid_setup/match_stations_to_grids.R")

# Preprocess data
source("2.scripts/preprocess/parse_meteo_data.R")                # Parse Meteo France weather station data
# source("2.scripts/preprocess/explore_meteo_data.R")            # Check the station data for outliers
source_or_skip("2.scripts/preprocess/clean_meteo_data.R")        # Remove suspicious station data
source_or_skip("2.scripts/preprocess/format_climate_types.R")    # Fix coordinates and disaggregate Joly (2010) climate types
source("2.scripts/preprocess/format_corine_land_cover.R")        # Disaggregate corine land cover
source_or_skip("2.scripts/preprocess/format_insee_population.R") # Rasterize and disaggregate INSEE population

# Extract data to grids
source("2.scripts/extract/extract_1km_data.R")
source("2.scripts/extract/extract_200m_data.R")
source("2.scripts/extract/extract_modis_lst.R")
source("2.scripts/extract/extract_modis_ndvi.R")

# Create mod1 and mod3 (databases for 1 km Ta model)
source("2.scripts/extract/create_mod3.R") # Create mod3 = all 1 km grid cell-days
source("2.scripts/extract/create_mod1.R") # Create mod1 = 1 km grid cell-days with Meteo obs

# Fit 1 km models and predict 1 km Ta
source("2.scripts/model/fit_m1.R")        # m1 = 1 km Ta from LST
source("2.scripts/model/predict_m1.R")    # predict with m1
source("2.scripts/model/fit_m2.R")        # m2 = model to gapfill 1 km Ta predictions
source("2.scripts/model/predict_m2.R")    # predict with m2
source("2.scripts/model/predict_1km.R")   # combine m1 and m2 predictions -> final 1 km preds

# Create mod4 and mod5 (databases for 200 m residual model)
source("2.scripts/extract/create_mod5.R") # Create mod5 = 200 m grid cell-days
source("2.scripts/extract/create_mod4.R") # Create mod4 = 200 m grid cell-days with Meteo obs

# Fit 200 m models and predict 200 m residual of 1 km Ta
source("2.scripts/model/fit_m3.R")          # m3 = 200 m residual of 1 km Ta; RF and XGB
source("2.scripts/model/predict_m3_mod4.R") # predict with m3 at station locations
source("2.scripts/model/fit_m4.R")          # m4 = ensemble of RF and XGB
source("2.scripts/model/predict_m3_mod5.R") # predict with m3 on 200 m grid
source("2.scripts/model/predict_m4_mod5.R") # predict with m4 on 200 m grid
                                            # also combines 1 km preds and m4 -> final 200 m preds

# Evaluate performance
source_or_skip("2.scripts/model/assess_models.R") # summarize cross-validated model performance
source("2.scripts/model/plot_models.R")           # save annual summary rasters and plots


report("--------")
report("Finished")
report("--------")


# Report any warnings
w <- warnings()
if (length(w) > 0) {

  report("")
  report(glue("Warnings: {length(w)}"))
  for (i in seq_along(w)) {
    report(glue("{i}: In {w[i]} : {names(w)[i]}"))
  }

}
