# Convert Hough et al. (2020) 200 m temperature model output from annual fst to monthly netCDF

library(data.table)
library(fst)
library(here)
library(logger)
library(lubridate)
library(nctools)  # available from https://github.com/ihough/nctools
library(sf)
library(stringr)


# -------------
# Configuration
# -------------

fst_dir <- here::here("1.work/predictions")
fst_paths <- list.files(fst_dir, "mod5_\\d{4}_pred_200m\\.fst", full.names = TRUE)  # input files
out_dir <- dirname(fst_paths[1])  # output dir for netCDF files
force <- FALSE  # set to TRUE to overwrite existing files
logger::log_threshold(logger::INFO)  # logging verbosity


# ------
# Script
# ------

logger::log_layout(logger::layout_glue_colors)

# Define global attributes and coordinate reference system
global_atts = list(
  nctools::nc_att(name = "conventions", value = "CF-1.12"),
  nctools::nc_att(name = "title", value = "Modelled ambient temperature at 2m above the surface"),
  nctools::nc_att(name = "source", value = "https://doi.org/10.1016/j.envres.2020.109244")
)
crs <- nctools::nc_crs(
  "epsg3035",
  grid_mapping_name = "lambert_azimuthal_equal_area",
  crs_wkt = stringr::str_replace_all(sf::st_crs(3035)$wkt, "\\s+", " "),
  reference_ellipsoid_name = "GRS 1980",
  atts = list(
    nc_att("latitude_of_projection_origin", 52L),
    nc_att("longitude_of_projection_origin", 10L),
    nc_att("false_easting", 4321000L),
    nc_att("false_northing", 3210000L)
  )
)

# Iterate over annual fst files
for (fst_path in fst_paths) {
  logger::log_info(paste("Converting", fst_path, "from fst to netCDF"))

  # Iterate over months, loading data from fst file and writing to netCDF file
  fst_data <- fst::fst(fst_path)
  data_year <- lubridate::year(fst_data[1, "date"])
  for (m in 1:12) {
    # Remember which variables are defined so we can remove new variables after each iteration
    env_names <- names(environment())

    # Check output path
    nc_path <- paste0("ta_200m_", data_year, "-", stringr::str_pad(m, 2, pad = 0), ".nc")
    nc_path <- file.path(out_dir, nc_path)
    if (file.exists(nc_path) && !force) {
      log_warn("Skipping existing path ", nc_path, ". Set force = FALSE to overwrite.")
      next
    }

    # Load data for the month
    logger::log_info(paste("Loading data for", month.name[m], data_year, "from", fst_path))
    m_start <- lubridate::date(paste(data_year, m, 1, sep = "-"))
    m_end <- m_start + lubridate::period(1, "month")
    idx <- as.integer(fst_data$date)
    idx <- which(idx >= as.integer(m_start) & idx < as.integer(m_end))
    if (max(diff(idx)) == 1) {
      data <- fst::read_fst(fst_path, from = min(idx), to = max(idx), as.data.table = TRUE)
    } else {
      data <- data.table::as.data.table(fst_data[idx, ])
    }
    rm(idx)

    # Rename columns to match netCDF variables
    crossref <- c(grd_200m_id = "grid_id", date = "time")
    data.table::setnames(data, old = names(crossref), new = crossref)

    # Sort by x, y, and time. The goal is to produce a netCDF file with dimensions (time, y, x)
    # where x varies fastest, as recommended by CF conventions and required by COARDS conventions.
    # RNetCDF needs the data to have order (x, y, time) i.e. reversed dimensions because it writes
    # data using R arrays, which have the first dimension varying fastest.
    nc_dims <- rev(c("time", "y", "x"))
    logger::log_debug("Sorting data by [", paste(nc_dims, collapse = ", "), "]")
    data.table::setkeyv(data, nc_dims)

    # Define netCDF coordinates
    logger::log_debug("Defining netCDF coordinates")
    t_coord <- nctools::nc_coord(
      name = "time",
      values = as.integer(data$time),
      axis = "T",
      long_name = "time",
      standard_name = "time",
      units = "days since 1970-01-01"
    )
    y_coord <- nctools::nc_coord(
      "y",
      values = nctools::fill_seq(data$y), # complete grid including locations with no data
      axis = "Y",
      long_name = "y coordinate of projection",
      standard_name = "projection_y_coordinate",
      units = "m"
    )
    x_coord <- nctools::nc_coord(
      "x",
      values = nctools::fill_seq(data$x), # complete grid including locations with no data
      axis = "X",
      long_name = "x coordinate of projection",
      standard_name = "projection_x_coordinate",
      units = "m"
    )

    # Define netCDF data variables
    logger::log_debug("Defining netCDF data variables")
    tmean <- nctools::nc_var(
      "tmean",
      data = data,
      dims = c("time", "y", "x"),
      type = "NC_FLOAT",
      pack_type = "NC_SHORT",
      scale_factor = 0.01,
      long_name = "24-hour mean ambient air temperature at 2 m above surface",
      standard_name = "air_temperature",
      units = "degC",
      grid_mapping = "epsg3035",
      atts = list(
        nctools::nc_att("units_metadata", value = "temperature: on_scale"),
        nctools::nc_att("cell_methods", value = "area: mean time: mean (interval: 24 hours)"),
        nctools::nc_att("number_of_significant_digits", value = 3, type = "NC_BYTE")
      ),
      deflate = 5,
      shuffle = TRUE,
      chunking = c(time = 1L, y = NA, x = NA)
    )
    tmin <- nctools::nc_var(
      "tmin",
      data = data,
      dims = c("time", "y", "x"),
      type = "NC_FLOAT",
      pack_type = "NC_SHORT",
      scale_factor = 0.01,
      long_name = "24-hour minimum ambient air temperature at 2 m above surface",
      standard_name = "air_temperature",
      units = "degC",
      grid_mapping = "epsg3035",
      atts = list(
        nctools::nc_att("units_metadata", value = "temperature: on_scale"),
        nctools::nc_att("cell_methods", value = "area: mean time: minimum (interval: 24 hours)"),
        nctools::nc_att("number_of_significant_digits", value = 3, type = "NC_BYTE")
      ),
      deflate = 5,
      shuffle = TRUE,
      chunking = c(time = 1L, y = NA, x = NA)
    )
    tmax <- nctools::nc_var(
      "tmax",
      data = data,
      dims = c("time", "y", "x"),
      type = "NC_FLOAT",
      pack_type = "NC_SHORT",
      scale_factor = 0.01,
      long_name = "24-hour maximum ambient air temperature at 2 m above surface",
      standard_name = "air_temperature",
      units = "degC",
      grid_mapping = "epsg3035",
      atts = list(
        nctools::nc_att("cell_methods", value = "area: mean time: maximum (interval: 24 hours)"),
        nctools::nc_att("number_of_significant_digits", value = 3, type = "NC_BYTE"),
        nctools::nc_att("units_metadata", value = "temperature: on_scale")
      ),
      deflate = 5,
      shuffle = TRUE,
      chunking = c(time = 1L, y = NA, x = NA)
    )
    grid_id <- nctools::nc_var(
      "grid_id",
      data = data[time == min(time), ],
      dims = c("y", "x"),
      type = "NC_CHAR",
      grid_mapping = "epsg3035",
      long_name = "id of grid cell",
      deflate = 5,
      shuffle = TRUE
    )

    # Save memory by removing unneeded 'data' object
    rm(data)

    # Write to netCDF file
    nctools::nc_write(
      path = nc_path, force = force, log_level = logger::log_threshold(), crs = crs,
      atts = global_atts, coords = list(t_coord, y_coord, x_coord),
      vars = list(tmean, tmin, tmax, grid_id)
    )

    # Cleanup
    rm(list = setdiff(names(environment()), env_names))

  }

}
